"use strict";

// Declare global where the helper methods will be defined.
var dashboardHelper = {};

( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;


	/**
	* Creates an html <object> element embedded with the passed listing,
	* and attaches it to the container.
	*
	* @param {MFiles.ShellListing} listing -The listing to attach.
	* @param {jQuery|HTML Node} container - An element/selector to which the listing should be attached.
	* @param {string} [id] - The id of that should be added to the created <object> element.
	* @param {string} [className] - The class name(s) that should be added to the created <object> element.
	* @returns {jQuery} The newly created <object> node wrapped with jQuery.
	* @throws An exception is thrown if the container is null or resolves to no/many nodes.
	*/
	dashboardHelper.attachListing = function( listing, container, id, className ) {

		// Sanity checks.
		container = $( container );
		if( container.length != 1 )
			MFiles.ThrowError( l( "ERR_INVALID_LISTING_CONTAINER" ) );

		// The listing throws an "Invalid Window Handle" error if we attach to a non-visible container.
		// To work around this we create a temporary, visible, small and off the page container to attach to.
		// Once the listing is attached to the window, we switch it to its intended container.
		var tempContainerID = 'viewutils-temp-listing-container';
		var tempContainer = $( "#" + tempContainerID );
		if( !tempContainer.length ) {
			tempContainer = $( "<div></div>" )
				.attr( "id", tempContainerID )
				.css( {
					position: "absolute",
					left: "-5000px",
					top: "0px",
					height: "1px",
					width: "1px"
				} ).appendTo( "body" );
		}


		// Create the object element that holds the listing.
		var obj = $( "<object classid='clsid:" + MFiles.CLSID.ShellListingCtrl + "' tabindex='1'></object>" )
			.appendTo( tempContainer );

		// Set the id if one was passed.
		if( id )
			obj.prop( "id", id );

		// Set the class name(s) if any were passed.
		if( className )
			obj.addClass( className );

		// Save a reference to the listing in the element's data.
		obj.data( "listing", listing );

		// Get a handle on the underlying elem.
		var elem = obj.get( 0 );

		// Attach the listing to the activeX control.
		elem.AttachToListingWindow( listing );

		// Now that the listing is attached to the window,
		// move it to it's proper container.
		obj.appendTo( container );

		// Return the jQuery wrapped listing's 'object' element.
		return obj;
	}


	/*
	* Helper funciton to virtually select an object in the listing.
	*
	* @param {MFiles.ShellListing} listing - The listing where the object should be selected.
	* @param {MFiles.ObjVer} objVer - The object version to select. Version can be -1.
	*/
	dashboardHelper.virtuallySelectObjVer = function( vault, listing, objVer ) {

		// Find the correct version if it isn't specified.
		if( objVer.Version == -1 ) {
			objVer = vault.ObjectOperations.GetLatestObjVer( objVer.objID, true, false );
		}

		// Setup the objVer in the right structure for virtual selection.
		var objOrFileVers = new MFiles.ObjOrFileVers();
		var objOrFileVer = new MFiles.ObjOrFileVer();
		objOrFileVer.ObjVer = objVer;
		objOrFileVers.Add( -1, objOrFileVer );

		// The listing must be active for the virtual selection to work.
		listing.ActivateListing();

		// Virtually select the grouping object.
		listing.SetVirtualSelection( objOrFileVers );
	}


	/*
	* Retruns a string representation of an objID or objVer
	*
	* @param {ObjID|ObjVer} objID - The objID to create a string of.
	* @returns {string} - String representation of the passed objID.
	*/
	dashboardHelper.objIDStr = function( objID ) {
		return objID.Type + "-" + objID.ID;
	}


} )();