"use strict";

// Declare global where our methods will be defined.
var EventInterface = {};

( function() {

	/*
	* @typedef EventHandler
	* @type {object}
	*
	* @property event - The name of the event the callback is intended to handle.
	* @property callback - The callback function to handle an event.
	* @property context - The value to use as "this" when the callback function is run.
	*/


	/*
	* Binds a callback to an event.
	*
	* @param {string} name - The name of the event to listen to.
	* @param {callback} callback - The callback function to handle the event.
	* @param {object} context - The value to use as "this" when the callback function is run.
	* @returns {object} Returns the method's own 'this' object for chaining.
	*/
	EventInterface.bind = function( event, callback, context ) {

		// Add the new handler to the events structure.
		this._events = this._events || {};
		this._events[event] = this._events[event] || [];
		this._events[event].push( {
			event: event,
			callback: callback,
			context: context
		} );

		return this;
	},



	/*
	* Unbinds a callback from an event (stops listening). 
	* A handler must match all passed (non-null/undefined) arguments to be unbound.
	*
	* @param {string} [event] - The name of the event the handler is currently bound to.
	*		If not specified, the handlers can be removed from all events.
	* @param {callback} [callback] - The callback handler to unbind.
	*		If not specified, all handlers can be unbound regardless of callback.
	* @param {object} [context] - The 'this' object used by the callback.
	*		If not specified, handlers can be unbound regardless of context.
	* @returns {object} Returns the method's own 'this' object for chaining.
	*/
	EventInterface.unbind = function( event, callback, context ) {

		// Check specified parameters.
		if( !event && !callback && !context ) {

			// No event, callback or context specified, unbind all events.
			this._events = {}
			return this;
		}


		// If an event was specified but there is no handler or context defined, 
		// simply clear all callbacks of the event.
		if( event && !callback && !context ) {
			delete this._events[eventName];
			return this;
		}


		// Loop over each event.
		for( var eventName in this._events ) {

			// Skip this event if an event was specified but this doesn't match it.
			if( event && event != eventName )
				continue;

			// Create a new list of handlers.
			var keepHandlers = [];

			// Loop over each event's handlers.
			for( var i in this._events[eventName] ) {
				var handler = this._events[eventName][i];

				// Drop the callback if no handler was specified, but the context was and it matches. 
				if( !callback && context && handler.context === context )
					continue;

				// Drop the callback if no context was specified, but the handler was and it matches. 
				if( !context && callback && handler.callback === callback )
					continue;

				// Drop the callback if both context and hanlder were specified and they both match.
				if( callback && context && handler.context === context && handler.callback === callback )
					continue;

				// If we made it here we can keep the callback.
				keepHandlers.push( handler );

			}

			// Update the event handlers.
			this._events[eventName] = keepHandlers;

		}

		return this;
	},



	/*
	* Internal method for triggering events, and executing their bound callbacks.
	*
	* If an event is cancelable and a callback returns false, the method will immediately 
	* return false. No further callbacks will be triggered.
	*
	* @param {string} event - The name of the event.
	* @param {Array} params - The params to pass to the bound callbacks when triggered.
	* @param {bool} [cancelable] - Indicates whether the event is cancelable. Default: False.
	* @returns {bool} Returns false if the event is canceled, true otherwise.
	*/
	EventInterface.trigger = function( event, params, cancelable ) {

		// Check if any handlers exist for the event.
		if( this._events && this._events[event] ) {

			// Loop over the callbacks for the specified event.
			for( var i = 0; i < this._events[event].length; i++ ) {

				var handler = this._events[event][i];

				// Trigger the callback
				var retVal = handler.callback.apply( handler.context || this, params );

				// If the event is cancelable, and the handler returned "false"
				// cancel the event.
				if( cancelable && retVal === false )
					return false;
			}
		}

		// Return true if we successfully called all handlers and there were no cancellations.
		return true;
	}


} )();