"use strict";

// Declare global where our object will be exported.
var NamedValueSettings;

( function() {

	/*
	* Provides easy access to named values stored in a certain storage type and alias.
	*
	* @class
	* @param {MFiles.Vault} vault - The vault in which the named values reside.
	* @param {MFiles.MFNamedValueType} type - The type of storage used for the named values.
	* @param {string} alias - The alias under which the named values are stored.
	* @param {boolean} [autoCommit] - Indicates whether updates to settings should be committed 
	*		to the server immediately.  Default: true.
	*/
	NamedValueSettings = function( vault, type, alias, autoCommit ) {
		this.vault = vault;
		this.type = type;
		this.alias = alias;
		this.settings = [];

		this.autoCommit = ( autoCommit || autoCommit === undefined );
		this.pendingChanges = new MFiles.NamedValues(); 

		this.loadAll();
	}


	/*
	 * Retrieves the settings as NamedValues from the server.
	 * 
	 * @returns {MFiles.NamedValues}
	 */
	NamedValueSettings.prototype.getNamedValues = function() {
		return this.vault.NamedValueStorageOperations.GetNamedValues( this.type, this.alias );
	}


	/*
	 * Loads/Refreshes the cached settings.
	 */
	NamedValueSettings.prototype.loadAll = function() {

		// Make sure any pending changes are committed before we reload.
		this.commit();

		// Reset our internal settings cache.
		this.settings = [];

		// Retrieve the named values from the designated namespace.
		var values = this.getNamedValues();

		// Loop over each key=>value pair.
		// Serializing the value and storing it with the key in the settings object.
		for( var i = 1; i <= values.Names.Count; i++ ) {

			// Resolve the key.
			var key = values.Names.Item( i );

			// Resolve the value.
			var value = values.Value( key );

			// Strip the value of any possible comments, and parse it.
			value = strip( value );
			var viewSettings = JSON.parse( value );

			// Update our internal settings cache.
			this.settings[key] = viewSettings;
		}
	}


	/*
	 * Saves a setting object to a particular key in these settings.
	 * Only commits it to server if autoCommit = true, otherwise
	 * commit() should be called.
	 *
	 * @param {string} key - The key where the settings are to be stored.
	 * @param {object} setting - The value to store.
	 * @returns {object} The stored value (clone of the one passed in).;
	 */
	NamedValueSettings.prototype.save = function( key, setting ) {

		// Serialize the Settings for namedvaluestorage.
		var serializedValue = JSON.stringify( setting, null, 4 ).replace( /\n/g, "\r\n" );

		// Add the settings to the list of pending changes to be commited to the server.
		this.pendingChanges.Value( key ) = serializedValue;		

		// Commit the changes right away, if the object is configured to do so.
		if( this.autoCommit )
			this.commit();

		// Make sure the settings are up to date within the context.
		// We restore from serizalization so we don't worry about storing references
		// to objects that were made by transient dashboards.
		return this.settings[key] = JSON.parse( serializedValue );
	}

	/*
	 * Commits pending setting changes to the server.
	 *
	 */
	NamedValueSettings.prototype.commit = function() {

		// If there are no changes to commit, don't bother.
		if( this.pendingChanges.Names.Count === 0 )
			return;

		// Commit the pendingChanges.
		this.vault.NamedValueStorageOperations.SetNamedValues( this.type, this.alias, this.pendingChanges );

		// Reset the pendingChanges.
		this.pendingChanges = new MFiles.NamedValues();

	}


	/*
	 * Deletes the key and its related settings.
	 *
	 * @param {string} key - The key of the settings to delete.
	 */
	NamedValueSettings.prototype.remove = function( key ) {

		// Remove the settings for the passed key in namedvaluestorage.
		var keys = new MFiles.Strings();
		keys.Add( -1, key );
		this.vault.NamedValueStorageOperations.RemoveNamedValues( this.type, this.alias, keys );

		// Make sure the key is removed from our cached settings object.
		delete this.settings[key];
	}



} )();