"use strict";

// Declare global where our object will be exported.
var ExtendedViewGroupListing;

( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;

	/**
	 * A control to show a listing panel that contains all the grouping levels of the current folder path.
	 * Grouping level objects are made 'selectable' so their metadata can be seen.
	 *
	 * @class
	 *
	 * @param {MFiles.ShellFrame} shellFrame - The shellFrame in which the panel can be shown.
	 * @param {HTMLNode | jQuery} container - The html element in which the panel should render itself.
	 * @param {FolderPathSettingsManager} folderSettingsManager - A folder path settings manager.
	 * @param {PathInfo} pathInfo - Information about the current path.
	 * @param {number} taskPaneGroup - The id of the task pane group that should be used for task pane commands.
	 * @param {ThemeDefinition} theme - A custom theme that is active in the shellFrame.
	 */
	ExtendedViewGroupListing = function( shellFrame, container, folderSettingsManager, pathInfo, taskPaneGroup, theme ) {

		// Store passed refs and initialize object variables.
		this.shellFrame = shellFrame;
		this.vault = shellFrame.ShellUI.Vault;
		this.folderSettingsManager = folderSettingsManager;
		this.pathInfo = pathInfo;
		this.pathSettings = this.folderSettingsManager.getUserFolderPathSettings( this.pathInfo );
		this.taskPaneGroup = taskPaneGroup;
		this.theme = theme;

		this.view;
		this.hasLevels = false;
		this.hasRealObjLevels = false;
		this.hasOtherLevels = false;
		this.groupLevelListings = [];


		// Theme the application based on which version of the client is running.
		var mfApp = new MFiles.MFilesClientApplication();
		if( mfApp.GetAPIVersion().Major < 12 ) {

			// Old client detected.
			
			// Remove the 2018 theme, and manage the theme the old-fashioned way.
			$( "#css-2018" ).remove();
			this.setThemeValues();

		} else {

			// 2018 client detected.

			// Initialize new style theme support.
			ThemeUtil.addThemeSupport( shellFrame.shellUI, $.mf.dashboard, new ViewExtensionTheme() );
		}

		// Initialize expanded container.
		this.$ = $( container ).addClass( "inactive" );
		this.$.html( ExtendedViewGroupListing.TEMPLATE );

		// Render the grouping levels.
		// Automatically toggles the visiblilty based on whethere there is something to show.
		this.renderGroupingLevels();

		// Bind expando behavior.
		var self = this;
		$( ".viewext-group-listing-header, .viewext-group-listing .expando" ).click( function( evt ) {
			self.toggleExpandedState();
		} );

		// Initialize the task pane commands if applicable.
		this.initializeCommands();


		// Listen for changes to the viewSettings.
		this.folderSettingsManager.bind( "AfterUpdatePathSettings", function( pathInfo, pathSettings, common ) {

			// If personal settings were changed, update the state.
			if( !common && this.pathInfo.folderPath == pathInfo.folderPath ) {

				// Update our reference to the settings.
				this.pathSettings = pathSettings;

				// If any relevant ui state settings have changed update the ui state.
				if( pathSettings.groupingPanelVisible != this.visible ||
					pathSettings.groupingPanelExpanded != this.expanded )
					this.setUIState( pathSettings );
			}

		}, this );


		// Set the initial state of the panel.
		this.setUIState( this.pathSettings );

	}



	/*
	* Injects custom theme values into the panel's style sheets.
	* Custom theme values must be set with the ThemeEditor vault application.
	*
	*/
	ExtendedViewGroupListing.prototype.setThemeValues = function() {

		// Don't bother trying to use theme values if no custom theme is defined.
		if( !this.theme )
			return;

		// Default color values are stored in the default css file.


		// Define how the theme values map to specific css rules.
		// Rules are defined by theme sections.
		// Each rule contains a css seclector and the css properties maped to a theme value key.
		var styles = {
			"listingTheme": [
				{
					// Active Group Header.
					selector: ".viewext-group-listing-header",
					properties: {
						"color": "sortableHeader_TextColor_Active",
						"background-color": "sortableHeader_BackgroundColor_Active",
						"border-color": "sortableHeader_DividerColor_Active"
					}
				},
				{
					// Inactive Group Header.
					selector: ".inactive .viewext-group-listing-header",
					properties: {
						"color": "sortableHeader_TextColor_Inactive",
						"background-color": "sortableHeader_BackgroundColor_Inactive",
						"border-color": "sortableHeader_DividerColor_Inactive"
					}
				},
				{
					// Item.
					selector: ".viewext-group-level-object",
					properties: {
						"color": "item_TextColor_Normal"
					}
				},
				{
					// Hovered Item.
					selector: ".viewext-group-level-object.selectable:hover",
					properties: {
						"color": "item_TextColor_Hot",
						"background-color": "item_BackgroundColor_Hot",
						"border-color": "item_BackgroundColor_Hot"
					}
				},
				{
					// Selected Item.
					selector: ".viewext-group-level-object.selectable.selected",
					properties: {
						"color": "item_TextColor_Selected",
						"background-color": "item_BackgroundColor_Selected",
						"border-color": "item_BackgroundColor_Selected"
					}
				},
				{
					// Hovered Selected Item.
					selector: ".viewext-group-level-object.selectable.selected:hover",
					properties: {
						"color": "item_TextColor_HotSelected",
						"background-color": "item_BackgroundColor_HotSelected",
						"border-color": "item_BackgroundColor_HotSelected"
					}
				},
				{
					// Inactive Selected Item.
					selector: ".inactive .viewext-group-level-object.selectable.selected",
					properties: {
						"color": "item_TextColor_SelectedNoFocus",
						"background-color": "item_BackgroundColor_SelectedNoFocus",
						"border-color": "item_BackgroundColor_SelectedNoFocus"
					}
				}

			]
		};

		// Loop over the theme sections defined in styles.
		for( var section in styles ) {

			// Loop over each rule defined in the section.
			for( var i in styles[section] ) {

				// Get a handle on this rule.
				var rule = styles[section][i];

				// Loop over the properties defined for this rule.
				for( var prop in rule.properties ) {

					// Resolve the theme key defined as the rule property's value.
					var key = rule.properties[prop];

					// Resolve the value from the theme with the theme key.
					var value = this.theme[section][key];

					// If a custom value is set for this theme key, set it in our custom css.
					if( value && value != "default" ) {
						var css = {};
						css[prop] = value;
						jss.set( rule.selector, css );
					}

				}
			}
		}


	}



	/**
	 * Classes used within the panel.
	 * @const
	 *
	 */
	var CLASS = ExtendedViewGroupListing.CLASS = {
		HEADER: "viewext-group-listing-header",
		LEVELS: "viewext-group-levels",
		LEVEL: "viewext-group-level",
		LEVEL_VALUE: "group-level-value",
		LEVEL_OBJECT: "viewext-group-level-object",
		SELECTED: "selected",
		EXPANDED: "expanded",
		COLLAPSED: "collapsed"
	}

	/**
	 * The basic html structure of the panel.
	 * @const 
	 *
	 */
	ExtendedViewGroupListing.TEMPLATE = "<div class='expando'></div>" +
										"<div class='viewext-group-listing-header'>Grouping Level Objects</div>" +
										"<div class='viewext-group-levels'></div>" +
										"<div class='viewext-group-level-listings' style='display:none'></div>";




	/**
	 * Initializes TaskPane commands to show and hide the panel (if applicable).
	 *
	 */
	ExtendedViewGroupListing.prototype.initializeCommands = function() {

		// Do not create commands if there is no use for them.
		if( !this.hasLevels )
			return;

		// Create a grouping level for view extensions.


		// Create the show/hide panel commands.
		this.showPanelCmd = this.shellFrame.Commands.CreateCustomCommand( l( "CMD_SHOW_GROUP_LEVEL_PANEL" ) );
		this.hidePanelCmd = this.shellFrame.Commands.CreateCustomCommand( l( "CMD_HIDE_GROUP_LEVEL_PANEL" ) );

		// Set icons.
		this.shellFrame.Commands.SetIconFromPath( this.showPanelCmd, "images\\grouping_levels.ico" );
		this.shellFrame.Commands.SetIconFromPath( this.hidePanelCmd, "images\\grouping_levels.ico" );

		// Set the default state to hidden.
		this.shellFrame.Commands.SetCommandState( this.showPanelCmd, CommandLocation_All, CommandState_Hidden );
		this.shellFrame.Commands.SetCommandState( this.hidePanelCmd, CommandLocation_All, CommandState_Hidden );

		// Add the show/hide panel commands to a View Extensions TaskPane group.
		this.shellFrame.TaskPane.AddCustomCommandToGroup( this.showPanelCmd, this.taskPaneGroup, 1 );
		this.shellFrame.TaskPane.AddCustomCommandToGroup( this.hidePanelCmd, this.taskPaneGroup, 1 );

		// Setup the custom command handler.
		var self = this;
		this.shellFrame.Commands.Events.Register( MFiles.Event.CustomCommand, function( cmd ) {

			// See if one of our commands was triggered.
			if( self.showPanelCmd == cmd ) {

				// Show the panel if the "show" command was triggered.
				self.setVisibleState( true );

			} else if( self.hidePanelCmd == cmd ) {

				// Hide the panel if the "hide" command was triggered.
				self.setVisibleState( false );
			}
		} );

	}

	/**
	 * Set's the panels visible and expanded states based on the folder path settings.
	 *
	 * @param {ExtendedFolderPathSettings} folderPathSettings - The current folder path's settings.
	 */
	ExtendedViewGroupListing.prototype.setUIState = function( folderPathSettings ) {

		// Set internal UI values.
		this.visible = folderPathSettings.groupingPanelVisible;
		this.expanded = folderPathSettings.groupingPanelExpanded;

		// Update the ui state with the new values.
		this.setVisibleState( this.visible );

		// No need to call setExpandedState, if the panel is visible it will get called.

	}


	/**
	 * Injects the panels visible and expanded states into the passed folder path settings.
	 *
	 * @param {ExtendedFolderPathSettings} [folderPathSettings] - The current folder path's settings. 
	 *			If not specified, the values are appended to an new, empty object.
	 * @param {ExtendedFolderPathSettings}
	 */
	ExtendedViewGroupListing.prototype.getUIState = function( folderPathSettings ) {

		// Use the default object (if passed), or else create a new one.
		folderPathSettings = folderPathSettings || {};

		// Add the visibility, keep it at -1 if the panel is not relevent here.
		if( this.hasLevels )
			folderPathSettings.groupingPanelVisible = this.visible;
		else
			folderPathSettings.groupingPanelVisible = -1;

		// Add the expanded state.
		folderPathSettings.groupingPanelExpanded = this.expanded;

		// Return the object we set the state to.
		return folderPathSettings;
	}



	/*
	* Sets the panel's visible state.
	*
	* @param {boolean} visible - The visible state to set. True = visible, False = hidden.
	*/
	ExtendedViewGroupListing.prototype.setVisibleState = function( visible ) {

		// Set the visibility.
		this.visible = visible;


		// If the visibility has never been explicity set for this view, 
		// automatically show it if there are real objects in the grouping.
		if( this.visible === -1 ) {
			if( this.hasRealObjLevels )
				visible = true;
			else
				visible = false;
		}


		if( !this.hasLevels ) {

			// Ignore visibility settings, and simply hide the panel, if there are no grouping levels.
			this.$.hide();
			this.visible = -1;

		} else if( visible ) {

			// Show the panel.
			this.$.show();

			// Refresh the expanded state.
			this.setExpandedState( this.expanded );

			// Hide the show command, show the hide command
			this.shellFrame.Commands.SetCommandState( this.showPanelCmd, CommandLocation_All, CommandState_Hidden );
			this.shellFrame.Commands.SetCommandState( this.hidePanelCmd, CommandLocation_All, CommandState_Active );

		} else {

			// Hide the panel.
			this.$.hide();

			// Hide the hide command, show the show command
			this.shellFrame.Commands.SetCommandState( this.showPanelCmd, CommandLocation_All, CommandState_Active );
			this.shellFrame.Commands.SetCommandState( this.hidePanelCmd, CommandLocation_All, CommandState_Hidden );

		}

		// Update layout.
		$( window ).resize();

		// Update the folder path settings if something has changed.
		if( this.pathSettings.groupingPanelVisible !== this.visible ) {
			this.pathSettings.groupingPanelVisible = this.visible;
			this.folderSettingsManager.setUserFolderPathSettings( this.pathInfo, this.pathSettings );
		}
	}

	/*
	* Sets the panel's expanded state. 
	* Panel remains collapsed if there are no levels to show.
	*
	* @param {Number} expanded - The expanded state to set. 
	*		0 = collapsed, 1 = expanded, 2 = semi-collapsed.
	*/
	ExtendedViewGroupListing.prototype.setExpandedState = function( expanded ) {

		// Remove any of the possible expand/collapse state classes from the element.
		var elem = $( ".viewext-group-listing" ).removeClass( "expanded collapsed semi-collapsed" );

		// Keep the panel collapsed if there are no levels.
		// Don't provide semi-collapse if it's not available.
		if( !this.hasLevels )
			expanded = 0;
		else if( expanded === 2 && !( this.hasRealObjLevels && this.hasOtherLevels ) )
			expanded = 1;


		// Set the internal state value. 
		this.expanded = expanded;

		// Update the header text based on what is shown.
		var header = $( ".viewext-group-listing-header" );
		if( expanded == 2 || ( this.hasRealObjLevels && !this.hasOtherLevels ) )
			header.text( l( "HEADER_GROUPING_LEVEL_OBJECTS" ) );
		else
			header.text( l( "HEADER_GROUPING_LEVELS" ) );



		// Update the class based on the state.
		if( expanded == 2 )
			elem.addClass( "semi-collapsed" );
		else if( expanded )
			elem.addClass( "expanded" );
		else
			elem.addClass( "collapsed" );

		// Adjust the indentation of items based on what is now visible.
		this.updateGroupingItemIndentation();

		// Trigger a window resize to get the layout to update. 
		$( window ).resize();


		// Update the folder path settings if something has changed.
		if( this.pathSettings.groupingPanelExpanded !== expanded ) {
			this.pathSettings.groupingPanelExpanded = expanded;
			this.folderSettingsManager.setUserFolderPathSettings( this.pathInfo, this.pathSettings );
		}
	}


	/*
	* Toggles the panels expanded/collapsed state.  Rotates between collapsed, expanded and semi-collapsed.
	* Semi-Collapsed is skipped if not applicable.
	*
	*/
	ExtendedViewGroupListing.prototype.toggleExpandedState = function( expanded ) {

		// Increment the current expanded state.
		var expanded = this.expanded + 1;

		// If we are passed semi-collapsed, or are semi-collapsed but shouldn't be, wrap the status back to closed.
		// Semi collapse should not be available if there is only one type of grouping object (real or other).
		if( expanded > 2 || ( expanded == 2 && ( !this.hasRealObjLevels || !this.hasOtherLevels ) ) )
			expanded = 0;

		// Set the newly resolved state.
		this.setExpandedState( expanded );
	}


	/*
	* Updates how far a grouping item is indented based on the number of 
	* visible objects above it.
	*
	*/
	ExtendedViewGroupListing.prototype.updateGroupingItemIndentation = function() {
		var depth = 0;
		var pad = 18;
		this.$.find( "." + CLASS.LEVELS ).children().each( function() {
			if( $( this ).is( ":visible" ) ) {
				var indent = pad * depth++;
				$( this ).css( "margin-left", indent + "px" );
			}
		} );
	}


	/**
	 * Creates a listing item in the panel for each grouping level in the current folder path.
	 *
	 */
	ExtendedViewGroupListing.prototype.renderGroupingLevels = function() {

		// Clear the element.
		var container = this.$.find( "." + CLASS.LEVELS ).empty();

		var folderDefs = this.shellFrame.CurrentFolder;

		// Reset internal values, that will be updated during the render process.
		this.hasLevels = !!this.pathInfo.groupLevels.length;
		this.hasRealObjLevels = !!this.pathInfo.realObjLevels;
		this.hasOtherLevels = ( this.pathInfo.groupLevels.length - this.pathInfo.realObjLevels ) > 0;


		// Loop over all group levels and create an item in the grouping panel.
		for( var i = 0; i < this.pathInfo.groupLevels.length; i++ ) {
			var groupLevelInfo = this.pathInfo.groupLevels[i];

			this.createPropertyGroupLevel( container, groupLevelInfo );
		}


		// Adjust the indentation of items based on what is now visible.
		this.updateGroupingItemIndentation();

	}



	/**
	 * Renders an individual folder path grouping level as a listing item in the panel.
	 *
	 * @param {HtmlElement|jQuery} container - The html element in which the item should be rendered.
	 * @param {Number} level - The depth of the listing item the Nth item in the listing.
	 */
	ExtendedViewGroupListing.prototype.createPropertyGroupLevel = function( container, levelInfo ) {

		// Resolve the folder def.
		var folderDef = levelInfo.folderDef;

		// Calculate which level we are at for indenting.
		var index = container.children().length;

		// Create the grouping level container.
		var elem = $( "<div tabindex='0'></div>" ).addClass( CLASS.LEVEL )
			.appendTo( container );

		// Create the title element for the level.
		var title = $( "<span></span>" ).addClass( CLASS.LEVEL_VALUE ).appendTo( elem )

		// Check if the property folder has an associated value.
		if( !folderDef.PropertyFolder.isNULL() ) {

			// Determine if the property folder is segmented, in which case, calculate the display value differently.
			var segmentSize = viewUtils.getExpressionSegmentSize( this.vault, levelInfo.level.Expression );
			if( segmentSize ) {

				// Folder is segment based, calculate and set the display name.
				title.text( viewUtils.getIDSegmentDisplayName( folderDef.PropertyFolder.Value, segmentSize ) );

			} else {

				// Default. use the property folder's display value as the title.
				title.text( folderDef.PropertyFolder.DisplayValue );
			}

		} else {

			// There is no value at this level.
			// This level must point to a folder for null values.

			// Show the name of the null folder as the title (helps with context).
			title.text( levelInfo.level.NULLFolderName );

			// Add the "empty" modifier to the item.
			$( "<span></span>" ).addClass( "group-level-multiple-items" ).appendTo( elem ).text( l( "GROUP_LEVEL_EMPTY" ) );

		}

		// Append the Folder Level Name to the level.
		// $("<span class='group-level-name'></span>").appendTo(elem).text(" (" +levelInfo.displayValue + ")");  

		// If an object type was resolved setup icons and click behavior
		if( levelInfo.basedOnValueList ) {

			// Resolve this folderDef's underlying value.
			var lookup;
			var type = folderDef.PropertyFolder.DataType;
			if( type == MFDatatypeLookup || type == MFDatatypeMultiSelectLookup )
				lookup = folderDef.PropertyFolder.GetValueAsLookup();

			// Handle any real object level.
			if( levelInfo.objType.RealObjectType ) {

				// Set class and object icon.
				elem.addClass( CLASS.LEVEL_OBJECT );
				imgPath = $.mf.icon( levelInfo.valueList );
			}

			var imgPath;
			if( levelInfo.objType.RealObjectType && lookup ) {

				// Determine the objVer this folder represents.
				var objVer = new MFiles.ObjVer();
				objVer.SetIDs( levelInfo.objType.ID, lookup.Item, lookup.Version );
				var deleted = false;
				if( objVer.Version == -1 ) {
					
					// This could fail if the object has been recently deleted (#156727).
					try {

						// Try to get the latest version.
						objVer = this.vault.ObjectOperations.GetLatestObjVer( objVer.objID, true, false );

					} catch( err ) {

						// Indicate that the object is deleted.
						$( "<span></span>" )
								.addClass( "group-level-deleted" )
								.appendTo( elem )
								.text( l( "ELEMENT_DELETED" ) );
						deleted = true;
					}
				}
				elem.data( "objVer", objVer );


				// If the object isn't deleted, add additional info/behavior.
				if( !deleted ) {

					// See if there is a class specific image and if so use it instead of the object icon.
					var objectVersion = this.vault.ObjectOperations.GetObjectInfo( objVer, false, false );
					var classImg = $.mf.icon( MFBuiltInValueListClasses, objectVersion.Class );
					if( classImg )
						imgPath = classImg;

					// Apply click handler for 'real' object types, to cause virtual selection.								
					var self = this;
					elem.click( function() {
						self.selectGroupLevelObject( elem );
					} );

					// Add select behavior if this is a real object level has a valid lookup.
					elem.addClass( "selectable" );
				}

			} else if( lookup ) {

				// We verify lookup was set, as folders that represent value list items with name collisions
				// will only provide their text value, in which case we can't lookup an icon.

				// Property folder represents a value list item.
				// Resolve image path of value list item.
				imgPath = $.mf.icon( levelInfo.valueList, lookup.Item );

			} else if( !folderDef.PropertyFolder.isNULL() ) {

				// No Lookup was found. 
				// This level must point to multiple items with the same name (grouping level value).

				// Add "multiple object" modifier.
				$( "<span></span>" ).addClass( "group-level-multiple-items" ).appendTo( elem ).text( l( "GROUP_LEVEL_MULTIPLE_ITEMS" ) );

			}

			// If we found an image append it to the level.
			if( imgPath )
				$( "<img/>" ).attr( "src", imgPath ).prependTo( elem );
		}
	}


	/*
	* Selection handler for a grouping level listing item based on a real object.
	*
	* @param {jQuery} elem - The jQuery wrapped listing item element that is being selected.
	* @param {MFiles.ObjVer} objVer - The grouping level's underlying object version.
	*/
	ExtendedViewGroupListing.prototype.selectGroupLevelObject = function( elem ) {

		// Try to retrieve an existing listing.
		var objVer = elem.data( "objVer" );
		var id = dashboardHelper.objIDStr( objVer );
		var groupLevelListing = this.groupLevelListings[id];

		// Create a new listing if one wasn't resolved.
		if( !groupLevelListing ) {
			groupLevelListing = this.groupLevelListings[id] = {};

			// Make sure we have the latest objver. It could have changed since the view was loaded.
			// Once the listing is loaded however, the listeners should keep it up-to-date, or the refresh action will.
			objVer = this.vault.ObjectOperations.GetLatestObjVer( objVer.ObjID, true, true );
			elem.data( "objVer", objVer );

			// Create the additional shell listing.
			var viewPath = viewUtils.getDefaultPathForObject( this.vault, objVer, MFLatestSpecificBehaviorLatest );
			groupLevelListing.listing = this.shellFrame.CreateAdditionalListingForPath( viewPath );
			groupLevelListing.elem = dashboardHelper.attachListing( groupLevelListing.listing, ".viewext-group-level-listings" );

			// Listen for when the item is selected/deselected and update the group panel element accordingly.
			groupLevelListing.listing.Events.Register( MFiles.Event.SelectionChanged, function( items ) {
				if( items.Count == 1 ) {
					elem.addClass( CLASS.SELECTED );
				} else {
					elem.removeClass( CLASS.SELECTED );
				}
			} );

			// Listen for when the item is updated in the listing.
			groupLevelListing.listing.Events.Register( MFiles.Event.ContentChanged, function( items ) {

				if( items.Count ) {

					// Item was updated. Refresh reference and visuals.
					var item = items.ObjectVersions.Item( 1 );
					elem.data( "objVer", item.ObjVer );
					elem.find( "." + CLASS.LEVEL_VALUE ).text( item.Title );

				} else {

					// Item was deleted.					
					elem.remove();
					groupLevelListing.elem.remove();
					delete this.groupLevelListings[id]
				}

			} );

			// Listen for when the listing is activated.
			groupLevelListing.listing.Events.Register( MFiles.Event.ListingActivated, function( items ) {
				$( ".viewext-group-listing" ).removeClass( "inactive" );
			} );

			// Listen for when the listing is deactivated.
			groupLevelListing.listing.Events.Register( MFiles.Event.ListingDeactivated, function( items ) {
				$( ".viewext-group-listing" ).addClass( "inactive" );
			} );
		}

		// Make the selection change appear to happen right away. 
		// If it is wrong somehow, the event handlers will correct it, but
		// now it is taking too long for the style to update on the item clicked
		// and the UI seems really slow.It can also happen in 11.3 that the 
		// selectionChanged event doesn't fire, but this makes it look as if
		// it did.
		$( "." + CLASS.SELECTED ).removeClass( CLASS.SELECTED );
		elem.addClass( CLASS.SELECTED );
		$( ".viewext-group-listing" ).removeClass( "inactive" );

		// Select the view in the parent listing.
		// Retrying for up to 3 seconds.
		this.selectObjectVersionWithRetry( groupLevelListing.listing, objVer, 100, 30 );

	}

	/**
	 * Selects an object within a listing, retrying a set number of times, at set interval
	 * if the object is not found to allow for the listing to initially load. This accounts
	 * for new async behavior in the views, but the lack of a mechanism to know when it has
	 * initially loaded (#156757).
	 *
	 * @param {jQuery} list - The listing object which should contain the object..
	 * @param {MFiles.ObjVer} objVer - The object version to select.
	 * @param {Number} delay - The time, in milliseconds, to wait before retrying.
	 * @param {Number} retries - The number of times to retry.
	 * */
	ExtendedViewGroupListing.prototype.selectObjectVersionWithRetry = function(
		listing,
		objVer,
		delay,
		retries
	) {

		// If there are no more retries, attempt the selection without
		// the try block so the correct error is thrown.
		// JS doesn't re-throw errors nicely.
		if( retries < 1 ) {

			// Select the object.
			listing.SelectObjectVersion( objVer );
			return;
		}

		// Try to select the object.
		try {

			// Select the object.
			listing.SelectObjectVersion( objVer );

		} catch( err ) {

			// Make sure this is a not found error.
			if( err.description.indexOf( "(0x8004000B)" ) == -1 ) {
				retries = 0;
			} else {
				--retries;
			}

			// Selecting the object failed.
			// Schedule a retry.
			var self = this;
			setTimeout( function() {
				self.selectObjectVersionWithRetry( listing, objVer, delay, retries );
			}, delay );
		}
	}

	/*
	* Refreshes all group level Listings.
	* Useful for full refreshes  la F5.
	*
	* Parameters apply to the synchronous calls only and are passed directly  
	* through to ShellListing::RefreshListing()
	*
	* @param {bool} smartUpdate
	* @param {bool} refreshFromServer
	* @param {bool} forceFullRefreshFromServer
	*/
	ExtendedViewGroupListing.prototype.refreshAllListings = function(
			smartUpdate, refreshFromServer, forceFullRefreshFromServer ) {

		// Loop over group level listings.
		for( var id in this.groupLevelListings ) {

			// Refresh the listing, if one is found for this group level.
			if( this.groupLevelListings[id].listing )
				this.groupLevelListings[id].listing.RefreshListing( smartUpdate, refreshFromServer, forceFullRefreshFromServer );
		}

	}

	/*
	* Cleans up resources before being destroyed.
	* Unbinds event listeners.
	* Removes references that may potentially be external objects 
	* (e.g., created in different module/dashboard).
	*/
	ExtendedViewGroupListing.prototype.cleanUp = function() {

		// Remove listeners of custom code.
		this.folderSettingsManager.unbind( null, null, this );

		// Remove external references.
		this.folderSettingsManager = null;
		this.pathInfo = null;
	}


} )();