"use strict";

// Declare global where our object will be exported.
var ExtendedViewSubListings;

( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;


	/*
	* @typedef TabInfo
	* @type {object}
	*
	* @property {string} id - The tabs id.
	* @property {jQuery} tab - The jQuery wrapped html element representing the tab.
	* @property {jQuery} panel - The jQuery wrapped html element representing the tab panel.
	* @property {string} subListingID - The id of the sublisting this tab represents.
	* @property {SubListing[]}  subListings - The sublistings that have been loaded into the panel.
	* @property {SubListing} activeListing - The currently active listing for the tab.
	* @property {boolean} applicable - Indicates whether this tab's subLisiting defintion is applicable 
	*									to the currently selected object.
	* @property {boolean} visible - Indicates whether this tab's subLisiting defintion is visible 
	*									based on the current folder path and the currently selected object.
	*/



	/*
	* @typedef SubListing
	* @type {object}
	*
	* @property {string} id - A unique id for the SubListing and it's html container element.
	* @property {MFiles.ObjectVersion} objectVersion - The object for which this listing is applicable.
	* @property {string} path - The relative path to the listing.
	* @property {ShellListing} listing - The additional listing object created for the path.
	* @property {jQuery} elem - The jQuery wrapped DomElement which the listing is attached to.
	* @property {boolean} error - Indicates whether an error occured when last loading the listing.
	* @property {boolean} visible - Indicates whether this is the tab's active listing.
	* @property {string} subListingID - The id of the sublisting this tab represents.
	*/



	/*
	* @typedef SortedTab
	* @type {object}
	*
	* @property {TabInfo} tabInfo - The tab info of the tab.
	* @property {SubListingDefinition} conf - The tabs underlying sublisting definition.
	*/


	/**
	 * A control to show a panel that can display multiple detail listings the currently selected object
	 * in the shellFrame.
	 *
	 * @class
	 *
	 * @param {MFiles.ShellFrame} shellFrame - The shellFrame in which the panel can be shown.
	 * @param {HTMLNode | jQuery} container - The html element in which the panel should render itself.
	 * @param {SubListingManager} subListingManager - The sublisting definition manager.
	 * @param {FolderPathSettingsManager} folderSettingsManager - A folder path settings manager.
	 * @param {PathInfo} pathInfo - Information about the current path.
	 * @param {number} taskPaneGroup - The id of the task pane group that should be used for task pane commands.
	 * @param {ThemeDefinition} theme - A custom theme that is active in the shellFrame.
	 */
	ExtendedViewSubListings = function( shellFrame, container, subListingManager, folderSettingsManager, pathInfo, taskPaneGroup, theme ) {

		// Store passed refs and initialize object variables.
		this.shellFrame = shellFrame;
		this.vault = shellFrame.ShellUI.Vault;
		this.subListingManager = subListingManager;
		this.lazyLoadListings = this.subListingManager.appSettings.registrySettings.lazyLoading;
		this.folderSettingsManager = folderSettingsManager;
		this.pathInfo = pathInfo;
		this.folderSettings = this.folderSettingsManager.getUserFolderPathSettings( this.pathInfo );
		this.taskPaneGroup = taskPaneGroup;
		this.theme = theme;

		// Setup html structure.
		this.$ = $( container )
		this.$.append( ExtendedViewSubListings.TEMPLATE )
		$( ".viewext-sublistings-header" ).text( l( "HEADER_SUBLISTING_NO_OBJECT_SELECTED" ) );

		this.tabControl = this.$.find( "." + CLASS.TAB_CONTROL );
		this.tabContainer = this.$.find( "." + CLASS.TAB_CONTROL + ">ul" );
		this.tabs = {};
		this.activeTab = null;

		// Initialize task pane commands.
		this.initializeCommands();

		// Initialize tab menu. Must be setup before tabs are initialized!
		this.menu = this.initializeSubListingMenu();

		this.initializeTabs();

		this.selectedObject = null;
		this.viewListings = {};

		// Set the UI State.
		this.setUIState( this.folderSettings );

		// Listen for changes to the viewSettings.
		this.folderSettingsManager.bind( "AfterUpdatePathSettings", function( pathInfo, pathSettings, common ) {

			// Check if personal settings for the current folder path were changed.
			if( !common && this.pathInfo.folderPath == pathInfo.folderPath ) {

				// Update our reference to the current settings.
				this.folderSettings = pathSettings;

				// Check if something significant has change.
				if( pathSettings.subListingPanelVisible != this.visible ||
					pathSettings.subListingPanelExpanded != this.expanded ||
					pathSettings.subListingPanelHeight != this.height ) {

					// Update the UI State.
					this.setUIState( this.folderSettings );
				}

				// Refresh no matter what!  Tab visibility could have changed.
				this.refresh();
			}


		}, this );



		// Initialize expando behavior
		var self = this;
		this.$.find( ".expando" ).click( function( evt ) {
			self.setExpandedState( !self.expanded );
		} );

		// Initialize settings button
		this.settingButton = this.$.find( ".viewext-sublisting-settings-button" );
		this.addMenuBehavior( this.settingButton );

		// Bind to events from the shellFrame.
		this.bindToShellFrame( shellFrame );

		// Make sure any currently selected item in the primary listing is loaded right away.
		this.onSelectionChanged( shellFrame.Listing, shellFrame.Listing.CurrentSelection );

	}




	/**
	 * Classes used by the panel.
	 * @const
	 */
	var CLASS = ExtendedViewSubListings.CLASS = {
		TAB_CONTROL: "viewext-sublisting-tabs"
	};


	/**
	 * Basic HTML structure of the panel.
	 * @const {string}
	 */
	ExtendedViewSubListings.TEMPLATE = "<div class='expando'></div><div class='viewext-sublisting-settings-button'></div>" +
							"<div class='viewext-sublistings-header'>No Object Selected</div>" +
							"<div class='viewext-sublisting-tabs'>" +
								"<ul>" +
									"<li class='new-tab-tab ui-state-default' title='Add Tab' tabindex='0'>" +
										"<span class='ui-icon ui-icon-plusthick'></span>" +
									"</li>" +
								"</ul>" +
							"</div>" +
							"<div class='viewext-viewlistings' style='display:none'></div>";

	/**
	* The HTML structure of a tab.
	* @const {string}
	*/
	ExtendedViewSubListings.TAB_TEMPLATE = "<li>" +
												"<a>" +
													"<span class='tab-title'></span>" +
													"<span class='item-count'></span>" +
													"<span class='tab-menu-button ui-icon ui-icon-triangle-1-s' tabindex='0'></span>" +
												"</a>" +

											"</li>";


	/*
	* The HTML structure of the menu.
	* @const {string}
	*
	* Please note that ORDER MATTERS! Do not change the order of the menu-items 
	* without updating ExtendedViewSubListings::updateMenuOptions()
	*
	* (Refresh all only used for debugging.)
	*/
	ExtendedViewSubListings.MENU_TEMPLATE = "<ul class='ui-menu tab-menu'>" +
												"<li class='viewext-sublisting-menuitem-browse'>Browse in New Window...</li>" +
												"<li class='viewext-sublisting-menuitem-hide'>Hide</li>" +
												"<li class='viewext-sublisting-menuitem-refresh'>Refresh</li>" +
												"<li class='viewext-sublisting-menuitem-configtab'>Edit...</li>" +
												"<li class='viewext-sublisting-menuitem-configview'>Edit View...</li>" +
												"<li class='spacer spacer-1'></li>" +
												"<li class='viewext-sublisting-menuitem-create'>Create New...</li>" +
												"<li class='spacer spacer-2'></li>" +
												"<li class='viewext-sublisting-menuitem-settings'>Manage Settings</li>" +
												/* "<li class='viewext-sublisting-menuitem-refreshAll'>Refresh</li>" + */
												"<li class='spacer spacer-3'></li>" +
												"<li class='viewext-sublisting-menuitem-reset'>Reset settings</li>" +
												"<li class='viewext-sublisting-menuitem-save'>Save as Common Settings</li>" +
												"<li class='spacer spacer-4'></li>" +
												"<li class='viewext-sublisting-menuitem-export'>Export Common Listings</li>" +
												"<li class='viewext-sublisting-menuitem-import'>Import Common Listings</li>" +
											"</ul>";


	/*
	* Binds listeners to the shellFrame.
	*
	* @param {MFiles.ShellFrame} shellFrame - The shellFrame to bind to.
	*/
	ExtendedViewSubListings.prototype.bindToShellFrame = function( shellFrame ) {

		// Bind to the shellFrame's current listing.
		this.bindToListing( shellFrame.Listing );

		// Listing for new listings, and bind to them.
		var self = this;
		shellFrame.Events.Register( MFiles.Event.NewShellListing, function( listing ) {
			self.bindToListing( listing );
		} );
	}



	/*
	* Binds listeners to a ShellListing.
	*
	* @param {MFiles.ShellListing} listing - The shellListing to bind to.
	*/
	ExtendedViewSubListings.prototype.bindToListing = function( listing ) {


		// Check the justCreatedListing flag so we don't bind to listings created by this panel.
		if( !this.justCreatedListing ) {

			// Listing created outside of the panel.

			var self = this;

			// Listen for selection changes in it.
			listing.Events.Register( MFiles.Event.SelectionChanged, function( shellItems ) {
				self.onSelectionChanged( listing, shellItems );
			} );

			// Listen for content changes in it.
			listing.Events.Register( MFiles.Event.ContentChanged, function( shellItems ) {
				self.onContentChanged( listing, shellItems );
			} );


		} else {

			// Skip the listing because this panel created it.

			// Reset the justCreatedListing flag.
			this.justCreatedListing = false;
		}
	}



	/**
	 * Initializes TaskPane commands to show and hide the panel (if applicable).
	 *
	 */
	ExtendedViewSubListings.prototype.initializeCommands = function() {

		// Create the show/hide panel commands.
		this.showPanelCmd = this.shellFrame.Commands.CreateCustomCommand( l( "CMD_SHOW_SUBLISTING_PANEL" ) );
		this.hidePanelCmd = this.shellFrame.Commands.CreateCustomCommand( l( "CMD_HIDE_SUBLISTING_PANEL" ) );

		// Set icons.
		this.shellFrame.Commands.SetIconFromPath( this.showPanelCmd, "images\\details_listing.ico" );
		this.shellFrame.Commands.SetIconFromPath( this.hidePanelCmd, "images\\details_listing.ico" );

		// Set the default state to hidden.
		this.shellFrame.Commands.SetCommandState( this.showPanelCmd, CommandLocation_All, CommandState_Hidden );
		this.shellFrame.Commands.SetCommandState( this.hidePanelCmd, CommandLocation_All, CommandState_Hidden );

		// Add the show/hide panel commands to a View Extensions TaskPane group.
		this.shellFrame.TaskPane.AddCustomCommandToGroup( this.showPanelCmd, this.taskPaneGroup, 2 );
		this.shellFrame.TaskPane.AddCustomCommandToGroup( this.hidePanelCmd, this.taskPaneGroup, 2 );


		// Add a button to show the manage sublisting dialog.
		this.settingsCmd = this.shellFrame.Commands.CreateCustomCommand( l( "CMD_MANAGE_SUBLISTINGS" ) );
		this.shellFrame.Commands.SetIconFromPath( this.settingsCmd, "images\\process.ico" );
		this.shellFrame.Commands.SetCommandState( this.settingsCmd, CommandLocation_All, CommandState_Active );
		this.shellFrame.TaskPane.AddCustomCommandToGroup( this.settingsCmd, this.taskPaneGroup, 3 );


		// Setup the custom command handler.
		var self = this;
		this.shellFrame.Commands.Events.Register( MFiles.Event.CustomCommand, function( cmd ) {

			// See if one of our commands was triggered.
			if( self.showPanelCmd === cmd ) {

				// Show the panel if the "show" command was triggered.
				self.setVisibleState( true );

			} else if( self.hidePanelCmd === cmd ) {

				// Hide the panel if the "hide" command was triggered.
				self.setVisibleState( false );

			} else if( self.settingsCmd === cmd ) {

				// Show the manage sublisting dialog.
				self.showManageSubListingsDialog();
			}
		} );

	}



	/**
	 * Initializes the tab widget, and creates a tab/panel for each
	 * sublisting definition. Also intializes the new tab menu.
	 */
	ExtendedViewSubListings.prototype.initializeTabs = function() {

		// Create a self reference for use in closures.
		var self = this;

		// Load all applicable sublisting definitions.
		var subListings = this.subListingManager.getAllSubListings();

		// Get a handle on the "new tab" tab, and add the menu behavior.
		this.newTabButton = this.$.find( ".new-tab-tab" ).hide();
		this.addMenuBehavior( this.newTabButton );


		// Create the tabs.
		for( var id in subListings ) {
			this.addTab( subListings[id], true )
		}

		// Sort the tabs.
		this.sortTabs();


		// Override the keyboard navigation events in the jQuery ui tab control.
		// They never seem to work right as they can't detect when focus is in 
		// the activeX controls, and they don't account for hidden tabs.
		// Not supporting them is better than doing them flat out wrongly!
		$.widget("ui.tabs", $.ui.tabs, {
			_tabKeydown: function (event) {
				// no operation.
			}
		});


		// Bind to the tab widget's activate and create events.
		this.tabControl.tabs( {
			activate: function( event, ui ) {

				// Update the currently active tab.
				self.onTabActivated( event, ui );
			},
			create: function( event, ui ) {

				// Resolve the initially active tab when the control is created.
				self.activeTab = self.getTabInfoForPanel( ui.panel );
			}
		} );

		// Bind to tab clicks.
		this.tabControl.on( "click", "li", function() {

			// Ensure the panel is expanded if a tab was clicked.
			if( !self.expanded )
				self.setExpandedState( true );
		} );


		// Bind to sublisting events, so we can update tabs as necessary.

		// New SubListing.
		this.subListingManager.bind( "AfterCreateSubListing", function( subListing, view ) {

			// Display the sublisting view's properties if a new view was created.
			if( view ) {
				this.openViewProperties( subListing.view );

				view = self.vault.ViewOperations.GetView( subListing.view );
				self.subListingManager.tryEnsureValidGroupingLevels( view, subListing );
			}

			// Create a tab for the new sublisting.
			this.addTab( subListing );

			// Refresh the tab control and panel.
			this.sortTabs();
			this.tabControl.tabs( "refresh" );
			this.refresh();

		}, this );


		// Modified SubListing.
		this.subListingManager.bind( "AfterModifySubListing", function( subListing ) {

			// Refresh the tab control and panel.
			this.sortTabs();
			this.tabControl.tabs( "refresh" );
			this.refresh();

		}, this );

		// Deleted SubListing.
		this.subListingManager.bind( "AfterDeleteSubListing", function( subListing ) {

			// Remove the tab associated with the deleted subListing.
			this.deleteTab( subListing.id );

			// Refresh the tab control and panel.
			this.tabControl.tabs( "refresh" );
			this.refresh();

		}, this );


		// Listen for batch updates to the sublistings.
		this.subListingManager.bind( "AfterRefreshSubListings", function( changes ) {

			var id;

			// Delete the tabs for removed sublistings.
			for( id in changes.removed )
				this.deleteTab( id );

			// Add tabs for the new sublistings without refreshing the tab control yet.
			for( id in changes.added )
				this.addTab( changes.added[id], true );

			// No need to handle the modified sublistings,
			// all sublistings will be refreshed when the control is refreshed.

			// Refresh the tab control and panel.
			this.sortTabs();
			this.tabControl.tabs( "refresh" );
			this.refresh();

		}, this );

	}



	/*
	* Returns the tab information for a sublisting.
	*
	* @param {string} id - The id of the sublisting.
	* @returns {TabInfo} - The sublisting's tab information.
	*/
	ExtendedViewSubListings.prototype.getSubListingTab = function( id ) {
		return this.tabs[id + "-detail-tab"];
	}



	/*
	* Creates a new tab for the passed sublisting configuration.
	*
	* @param {SubListingDefinition} conf - The sublisting a tab should be created for.
	* @param {boolean} noRefresh - Indicates whether the tabControl should be refreshed after tab creation.
	*		Refresh should be suppressed if mulitple tabs are being created. Like on initialization.
	* @returns {TabInfo} The newly created tab.
	*/
	ExtendedViewSubListings.prototype.addTab = function( conf, noRefresh ) {


		// Create a unique id for the tab.
		var tabID = conf.id + "-detail-tab";

		// Create a tab button.
		var button = $( ExtendedViewSubListings.TAB_TEMPLATE )
			.insertBefore( this.newTabButton )
			.hide();

		// Set the tab link to it's panel.
		button.find( "a" ).attr( "href", "#" + tabID );

		// Set the tab title.
		button.find( ".tab-title" ).text( conf.name );

		// Create the tab panel.
		var panel = $( "<div></div>" ).appendTo( this.tabControl ).attr( "id", tabID );

		// Refresh the tab control if
		if( !noRefresh )
			this.tabControl.tabs( 'refresh' );


		// Keep track of the tab details.
		/* @var {TabInfo} */
		var tab = {
			id: tabID,
			tab: button,
			panel: panel,
			subListingID: conf.id,
			subListings: {}
		};
		this.tabs[tabID] = tab;

		// Add Menu behavior.
		this.addMenuBehavior( button.find( ".tab-menu-button" ) );

		return tab;
	}



	/*
	* Removes a tab from the tab panel.
	*
	* @param {string} id - The id of the subListing definition whose tab should be removed.
	*/
	ExtendedViewSubListings.prototype.deleteTab = function( id ) {

		// Resole the tab info object, and remove it from the internal collection.
		var tabInfo = this.getSubListingTab( id )
		delete this.tabs[tabInfo.id];

		// Remove corresponding DomElements.
		tabInfo.tab.remove();
		tabInfo.panel.remove();
	}



	/*
	* Event handler for when a sublisting's tabs is activated.
	*
	* @param {object} event - The jQuery event object.
	* @param {ui} - The ui object defined by the jQueryUI Tab control.
	* @param {jQuery} ui.newTab - The tab that was just activated.
	* @param {jQuery} ui.oldTab - The tab that was just deactivated.
	* @param {jQuery} ui.newPanel - The panel that was just activated.
	* @param {jQuery} ui.oldPanel - The panel that was just deactivated.
	*/
	ExtendedViewSubListings.prototype.onTabActivated = function( event, ui ) {

		// A tab has just been activated.

		// Make sure the corresponding detail panel is visible in the tab panel.
		var tabInfo = this.getTabInfoForPanel( ui.newPanel );
		this.activeTab = tabInfo;

		// Ensure that listing is loaded.
		this.loadSublisting( tabInfo, true );

		if( tabInfo && tabInfo.activeListing )
			tabInfo.activeListing.lastActivated = event.timeStamp;
	}



	/**
	 * Event handler for when a ShellListing's selection has changed.
	 * If only a single item is selected the sublistings are updated with that as the
	 * target object. If no objects, or many objects are selected, the event is ignored.
	 * If the control is not visible it is not refreshed.
	 *
	 * @param {MFiles.ShellListing} listing - The listing in which the selection was just changed.
	 * @param {MFiles.ShellItems} items - The items currently selected in the listing.
	 */
	ExtendedViewSubListings.prototype.onSelectionChanged = function( listing, items ) {

		// Skip updating if nothing or many things are selected.
		if( !items || items.ObjectVersions.Count !== 1 )
			return;

		// Update the internal references to the target object and it's listing.
		this.selectedObjectListing = listing;
		this.selectedObject = items.ObjectVersions[0];

		// Skip refreshing the control if it isn't visible.
		if( !this.visible )
			return;

		// Refresh the control in a timeout to prevent any stuck error cycles.
		var self = this;
		setTimeout( function() {
			self.refresh();
		}, 1 );
	}


	/**
	 * Event handler for when a ShellListing's content has changed.
	 * Specifically listening for any changes to the current target object.
	 *
	 * @param {MFiles.ShellListing} listing - The listing in which the content was just changed.
	 * @param {MFiles.ShellItems} items - The items currently  in the listing.
	 */
	ExtendedViewSubListings.prototype.onContentChanged = function( listing, items ) {

		// We are only listening for changes to the target object so 
		// ignore content changes in listings other than the one where
		// the selected object is.
		if( !this.selectedObject || !this.selectedObjectListing || listing !== this.selectedObjectListing ) {
			
			// HACK: In 11.3 there is a bug when a grouping level panel object is selected before anything else,
			// the selectionChanged event never fires, but the contentChanged event fires multiple times, and in
			// the last one, the new selection can be seen in the listing object.
			// See issue #141280.
			if( !this.selectedObject && !this.selectedObjectListing && listing.IsActive ) {
				this.onSelectionChanged( listing, listing.CurrentSelection );
			}
			
			return;
		}

		// Get a handle on the target object's objVer.
		var selObjVer = this.selectedObject.ObjVer;

		// Keep track of whether we find the object, and whether it has been updated.
		var found;
		var updated;

		// Loop over all objects in the listing.
		for( var i = 1; i <= items.ObjectVersions.Count; i++ ) {

			// See if this object matches the target object.
			var objVer = items.ObjectVersions.Item( i ).ObjVer;
			if( objVer.Type === selObjVer.Type && objVer.ID === selObjVer.ID ) {

				// This is the target object.
				found = true;

				// Check for changes.
				if( objVer.Version !== selObjVer.Version ) {

					this.selectedObject = items.ObjectVersions.Item( i );
					updated = true;
				}

				// Nothing left to search for.
				break;
			}
		}

		// If the object wasn't found, it shouldn't be a target object anymore.
		if( !found ) {

			// Object has disappeared.
		}


		if( updated ) {
			// Refresh the control in a timeout to prevent any stuck error cycles.
			var self = this;
			setTimeout( function() {
				self.refresh();
			}, 1 );
		}


	}


	/**
	 * Set's the panels ui state based on the folder path settings.
	 *
	 * @param {ExtendedFolderPathSettings} folderPathSettings - The current folder path's settings.
	 */
	ExtendedViewSubListings.prototype.setUIState = function( folderPathSettings ) {

		// Set the initial height (when expanded) and state of the control.
		this.height = folderPathSettings.subListingPanelHeight;
		this.visible = folderPathSettings.subListingPanelVisible;
		this.expanded = folderPathSettings.subListingPanelExpanded;

		// Set the default visibility of the control.
		this.setVisibleState( this.visible );

		// Set the expanded state of the control.
		this.setExpandedState( this.expanded );

	}



	/**
	 * Injects the panels ui state into the passed folder path settings.
	 *
	 * @param {ExtendedFolderPathSettings} [folderPathSettings] - The current folder path's settings. 
	 *			If not specified, the values are appended to an new, empty object.
	 * @param {ExtendedFolderPathSettings}
	 */
	ExtendedViewSubListings.prototype.getUIState = function( folderPathSettings ) {

		// Use the default object (if passed), or else create a new one.
		folderPathSettings = folderPathSettings || {};

		// Add the ui values to the object.
		folderPathSettings.subListingPanelHeight = this.height;
		folderPathSettings.subListingPanelVisible = this.visible;
		folderPathSettings.subListingPanelExpanded = this.expanded;

		// Return the object we set the state to.
		return folderPathSettings;
	}



	/*
	* Refreshes the panel's internal layout.
	* Calculates practical tab widths
	* Calculates panel height.
	*/
	ExtendedViewSubListings.prototype.updateLayout = function() {


		// Update the folder path settings if the height has changed.
		if( this.folderSettings.subListingPanelHeight !== this.height ) {
			this.folderSettings.subListingPanelHeight = this.height;
			this.folderSettingsManager.setUserFolderPathSettings( this.pathInfo, this.folderSettings );
		}



		// Update the tab panel height.
		if( this.visible && this.expanded ) {
			var tabPanels = $( ".viewext-sublisting-tabs>div:visible" );
			if( tabPanels.length ) {
				var tabPanelOffset = tabPanels.offset();
				var tabPanelHeight = $( window ).height() - tabPanelOffset.top;
				$( ".viewext-sublisting-tabs>div" ).each( function() {
					$( this ).height( tabPanelHeight );
				} );
			}
		}


		// Find all visible tabs.
		var visibleTabs = $.map( this.tabs, function( tabDef, i ) {
			if( tabDef.visible ) return tabDef;
		} );

		// Make sure the tabs all fit in the horizontal space.
		if( visibleTabs.length ) {

			var width = this.tabContainer.width();  // Resolve the window width.
			var newTabTabWidth = $( ".new-tab-tab" ).outerWidth();
			var availWidth = width - newTabTabWidth;  // Leave space for the add tab button.
			var maxTabWidth = ( availWidth / visibleTabs.length ) - 3; // Get the max length of each tab, adjusting for borders & margins.

			// Keep tabs within a reasonable width overall (even if there is room)
			maxTabWidth = Math.min( 250, maxTabWidth );

			// Loop over each visible tab.
			$.each( visibleTabs, function( i, tabDef ) {

				// Set the tab's max width.
				tabDef.tab.css( "max-width", maxTabWidth );

				// Decide if there is enough room to show the item count.
				var itemCount = tabDef.tab.find( ".item-count" );
				var countWidth = 0;
				if( maxTabWidth > 80 ) {
					itemCount.show();
					countWidth = itemCount.width();
				} else {
					itemCount.hide();
				}

				// Calculate the max width of the tab title.
				// Always leave room for the tab menu button, and the item count if visible.
				var maxTabTitleWidth = maxTabWidth - 12 - 16 - countWidth; // subtract padding and margin (15), menu button (16) and count text.
				tabDef.tab.find( ".tab-title" ).css( "max-width", maxTabTitleWidth );
			} );

		}
	}


	/*
	* Sets the panel's visible state.
	*
	* @param {boolean} visible - The visible state to set. True = visible, False = hidden.
	*/
	ExtendedViewSubListings.prototype.setVisibleState = function( visible ) {

		this.visible = visible;

		if( visible ) {

			// Show the panel
			this.$.show();

			// Hide the show command, show the hide command
			this.shellFrame.Commands.SetCommandState( this.showPanelCmd, CommandLocation_All, CommandState_Hidden );
			this.shellFrame.Commands.SetCommandState( this.hidePanelCmd, CommandLocation_All, CommandState_Active );

			// Refresh the panel.
			this.refresh();

		} else {

			// Hide the panel.
			this.$.hide()

			// Hide the hide command, show the show command
			this.shellFrame.Commands.SetCommandState( this.showPanelCmd, CommandLocation_All, CommandState_Active );
			this.shellFrame.Commands.SetCommandState( this.hidePanelCmd, CommandLocation_All, CommandState_Hidden );
		}

		// Update layout.
		$( window ).resize();


		// Update the folder path settings if something has changed.
		if( this.folderSettings.subListingPanelVisible !== this.visible ) {
			this.folderSettings.subListingPanelVisible = this.visible;
			this.folderSettingsManager.setUserFolderPathSettings( this.pathInfo, this.folderSettings );
		}
	}

	/*
	* Sets the panel's expanded state. 
	* Panel remains collapsed if there are no levels to show.
	*
	* @param {Number} expanded - The expanded state to set. True = expanded, False = collapsed.
	*/
	ExtendedViewSubListings.prototype.setExpandedState = function( expanded ) {

		// Update the internal state.
		this.expanded = expanded;

		// Calculate the height of the panel.
		var height;
		if( expanded ) {

			// Retreive the previous 'expanded' height.
			height = this.height;
		} else {

			// Set the height of just the header and tabs.
			height = $( ".viewext-sublistings-header" ).outerHeight() + $( ".viewext-sublisting-tabs>ul" ).outerHeight();
		}


		// Update the classes and height.
		$( ".viewext-sublistings" )
            .toggleClass( "expanded", expanded )
            .toggleClass( "collapsed", !expanded )
			.height( height );


		// Update layout.
		$( window ).resize();


		// Update the folder path settings if something has changed.
		if( this.folderSettings.subListingPanelExpanded !== this.expanded ) {
			this.folderSettings.subListingPanelExpanded = this.expanded;
			this.folderSettingsManager.setUserFolderPathSettings( this.pathInfo, this.folderSettings );
		}

	}


	/*
	* Returns all tabs in order of priority.
	*
	* @returns {SortedTab[]} - An array of the tabs and their sublisting definitions.
	*/
	ExtendedViewSubListings.prototype.getTabsInPriorityOrder = function() {

		// Create an array of objects containing the tabs and their priority.
		var tabs = [];
		for( var id in this.tabs ) {
			var tabInfo = this.tabs[id]
			var conf = this.subListingManager.getSubListing( tabInfo.subListingID );
			tabs.push( {
				tabInfo: tabInfo,
				conf: conf
			} );
		}

		// Sort the tabs by their priority
		tabs.sort( function( a, b ) {
			return b.conf.priority - a.conf.priority;
		} );

		return tabs;
	}



	/**
	 * Rearranges the tab elements into their priority sorted order.
	 */
	ExtendedViewSubListings.prototype.sortTabs = function() {

		// Get the tabs information in order.
		var tabs = this.getTabsInPriorityOrder();

		// If there is one or no tabs there is nothing to do.
		if( tabs.length <= 1 )
			return;

		// Rearrange the tab element in the order provided.

		// Place the initial tab
		tabs[0].tabInfo.tab.prependTo( this.tabContainer );

		// Place each subsequent tab after the first one.
		for( var i = 1; i < tabs.length; i++ ) {
			var current = tabs[i].tabInfo;
			var previous = tabs[i - 1].tabInfo

			current.tab.insertAfter( previous.tab );
		}
	}



	/**
	 * Determines which tabs are currently visible and applicable for the current folder path
	 * and passed object version.
	 * 
	 * @param {MFiles.ObjectVersion} objectVersion - The object version to display tabs for.
	 * @return {TabInfo[]} - The currently visible tabs.
	 */
	ExtendedViewSubListings.prototype.updateVisibleTabs = function( objectVersion ) {


		// Loop over all tabs, update their applicable / visible status, and keep a list of the visible ones.
		var self = this;
		var visibleTabs = $.map( this.tabs, function( tabDef, i ) {

			// Determine if this tab is applicable & visible for the selected item and folder path.
			var conf = self.subListingManager.getSubListing( tabDef.subListingID );
			tabDef.applicable = self.subListingManager.isObjectVersionTargetForSublisting( objectVersion, conf );
			tabDef.visible = self.subListingManager.isVisibleForFolderPathAndObjectVersion( conf, self.folderSettings, objectVersion );

			// Hide/Show the tab as appropriate.
			tabDef.tab.toggle( tabDef.visible );

			// Hide non-visible panels. The tab control will worry about showing the correct ones.
			// This ensures that no panel is shown if no tab is visible.
			if( !tabDef.visible )
				tabDef.panel.hide();

			// Add it to the list if visible.
			if( tabDef.visible )
				return tabDef;
		} );


		// Return the visible tabs.
		return visibleTabs;

	}



	/**
	 * Reloads all tabs for the currently selected object.
	 *
	 */
	ExtendedViewSubListings.prototype.refresh = function() {

		if( !this.selectedObject )
			return;

		// Set selected object.
		var parent = $( ".viewext-sublistings-header" ).empty();


		var self = this;

		// Resolve the object's icon using a class specific icon if available
		// and resorting to the object type icon.
		var icon = $.mf.icon( MFBuiltInValueListClasses, this.selectedObject.Class );
		if( !icon )
			icon = $.mf.icon( this.selectedObject.ObjVer.Type );

		// Update the details header with the selected object.
		var selObjTitle = $( "<div class='sel-obj-title'><div>" ).appendTo( parent )
			.text( this.selectedObject.Title )
			.click( function( evt ) {

				// Reselect the target object in it's original listing.
				if( self.selectedObjectListing && self.selectedObject ) {
					try {
						self.selectedObjectListing.SelectObjectVersion( self.selectedObject.ObjVer );
					} catch( e ) {
						// Couldn't find the object. Swallow the error.
						// TODO: FIX THIS!
						// Seems to happen when sub item is selected.
					}
				}

			} );

		// Set the icon.
		$( "<img />" ).attr( "src", icon ).prependTo( selObjTitle );


		// Make sure the new tab button is visible when an item is selected.
		this.newTabButton.show();

		// Update the tabs that should be visible for the selected item.
		var visibleTabs = this.updateVisibleTabs( this.selectedObject );

		// Make sure all visible tabs are propertly loaded with the selected object's listings.
		$.each( visibleTabs, function( i, tabInfo ) {
			// Load sublisting but content only for the active tab if lazy loading is enabled.
			self.loadSublisting( tabInfo, ! self.lazyLoadListings || self.activeTab === tabInfo );
		} );

		// Make sure one of the visible tabs is active.
		if( visibleTabs.length && ( !this.activeTab || !this.activeTab.visible ) ) {
			this.tabControl.tabs( "option", "active", visibleTabs[0].tab.index() );
		}

		// Make sure the active tab's panel is visible, if it should be.
		if( this.activeTab && this.activeTab.visible )
			this.activeTab.panel.show();

		// Update active tab's active listing lastActive time.
		if( this.activeTab && this.activeTab.activeListing )
			this.activeTab.activeListing.lastActivated = ( new Date() ).getTime();


		// Update the collapsed layout to ensure tabs are visible if they need to be.
		// Otherwise just refresh the layout (primarily so tab widths are correct).
		if( !this.expanded )
			this.setExpandedState( false );
		else
			$( window ).resize();

		// Do some clean up so we don't have too many old listings lying around.
		this.cleanUpListings();
	}




	/*
	* Gets the tabInfo object for a panel element.
	*
	* @param {jQuery} panel - A tab panel element.
	* @returns {TabInfo} Returns the panel's tabInfo if found, otherwise undefined.
	*/
	ExtendedViewSubListings.prototype.getTabInfoForPanel = function( panel ) {

		// Make sure the panel is a jQuery wrapped element.
		var panel = $( panel );

		// Make sure we actually have a panel.
		if( panel.length < 1 )
			return;

		// Lookup the corresponding tab info by panel id.
		var tabInfo = this.tabs[panel.attr( "id" )];

		return tabInfo;
	}


	/*
	* Gets the tabInfo object for a tab element.
	*
	* @param {jQuery} tab - A tab element.
	* @returns {TabInfo} Returns the tab's tabInfo if found, otherwise undefined.
	*/
	ExtendedViewSubListings.prototype.getTabInfoForTab = function( tab ) {

		// Make sure the tab is a jQuery wrapped element.
		var tab = $( tab );

		// Make sure we actually have a tab.
		if( tab.length < 1 )
			return;

		// Lookup the corresponding tab info by panel id.
		var id = tab.find( "a" ).attr( "href" ).replace( "#", "" );
		var tabInfo = this.tabs[id];

		return tabInfo;
	}



	/**
	 * Loads/Refreshes a SubListing for a tab and the currently selected object.
	 *
	 * @param {TabInfo} tabInfo - The tab who should have its SubListing loaded.
	 * @param {LoadContent} loadContent - True to load actual content to the listing if not already loaded.
	 */
	ExtendedViewSubListings.prototype.loadSublisting = function( tabInfo, loadContent ) {

		// Hide all sublistings in the tab's panel.
		$.each( tabInfo.subListings, function( i, subListing ) {
			subListing.visible = false;
			if( subListing.elem ) {
				subListing.elem.hide();
			}
		} );

		var alreadyLoaded = false;

		// Resolve the currently selected objVer.
		if( !this.selectedObject )
			MFiles.ThrowError( l( "ERR_NO_SELECTED_OBJECT" ) );

		// Resolve the tab's underlying sublisting def
		var conf = this.subListingManager.getSubListing( tabInfo.subListingID );

		// Resolve an object specific detail listing id.
		var objVer = this.selectedObject.ObjVer;
		var id = conf.id + "-" + dashboardHelper.objIDStr( objVer );

		// Try to find an existing sublisting.
		var subListing = tabInfo.subListings[id];

		// Determine if we need to load a listing.
		if( !subListing ) {

			// No sublisting found. Create it.
			subListing = {
				id: id,
				objectVersion: this.selectedObject
			};

			// Store also for the tab.
			if( loadContent ) {
				tabInfo.subListings[id] = subListing;
			}
			

		} else if( subListing && subListing.error ) {

			// Sublisitng has an error.  Remove the error panel,
			// and try to load the listing again.
			subListing.elem.remove();
			subListing.elem = null;
			loadContent = true;

		}
		else {
			alreadyLoaded = true;
		}

		// Create a new listing with the id, if one doesn't already exist, or a previous one exists with an error.
		if( loadContent && !alreadyLoaded ) {

			try {
				// Load the listing path for the configured view and the selected object.
				subListing.path = this.subListingManager.getSubListingPathForObject( objVer, conf );

				// Create and attach the detail listings.
				this.justCreatedListing = true;
				subListing.subListingID = conf.id;
				subListing.listing = this.shellFrame.CreateAdditionalListingForPath( subListing.path );
				subListing.elem = dashboardHelper.attachListing( subListing.listing, tabInfo.panel, id, "viewext-sublisting" );
				subListing.listing.ColumnHeadersVisible = true;

				// Set the custom theme on the listing if one is defined.
				if( this.theme && this.theme.listingTheme )
					subListing.listing.SetTheme( this.theme.listingTheme );

				// Bind to the listing's content changed event so the count automatically updates when items are added/removed.
				var self = this;
				subListing.listing.Events.Register( MFiles.Event.ContentChanged, function () {
					self.updateSublistingTab( tabInfo, subListing, true );
				} );

				subListing.error = false;

			} catch( e ) {

				// An error occured loading the subListing's panel.
				subListing.error = true;

				// Put the error information into the panel instead of the listing.
				subListing.elem = $( "<div class='viewext-sublisting error'></div>" ).attr( "id", id ).appendTo( tabInfo.panel );
				var msg = $( "<div class='message'></div>" ).appendTo( subListing.elem ).text( l( "ERR_LISTING_LOAD_FAILED", e.message ) );
				var details = $( "<a class='details'></a>" ).appendTo( subListing.elem ).text( l( "BUTTON_SHOW_ERROR_DETAILS" ) ).click( function() {
					$( this ).hide().siblings( ".stack" ).show();
				} );
				var stack = $( "<div class='stack'></div>" ).appendTo( subListing.elem ).text( e.stack ).hide();

			}

		}

		// If listing was loaded, make this the only sublisting visible in the panel.
		// Note this does not make it necessarily visible to the user.
		if( loadContent || alreadyLoaded ) {
			tabInfo.activeListing = subListing;
			subListing.elem.show();
		}

		subListing.visible = true;

		// Update the tab info.
		this.updateSublistingTab( tabInfo, conf, loadContent || alreadyLoaded );

		// Return the loaded sublisting info.
		return subListing;

	}

	/**
	 * Updates a tab based on it's currently active subListing.
	 *
	 * @param {TabInfo} tabInfo - The tab to update.
	 * @param {SubListingDefinition} conf - The tab's sublisting definition.
	 * @param {UpdateItemCount} updateItemCount - True if item count should be updated.
	 */
	ExtendedViewSubListings.prototype.updateSublistingTab = function (tabInfo, conf, updateItemCount ) {

		// Resolve the count element and remove it's child nodes if it is visible. Possibly existing
		// item count may for example refer to a previosly selected target.
		var countElem = tabInfo.tab.find( ".item-count" );

		// Don't update count if the tab or sublisting aren't visible.
		if ( !tabInfo.visible || !tabInfo.activeListing || !tabInfo.activeListing.visible ) {
			if ( tabInfo.visible ) {
				// However, clear item count if tab is visible. Needed to clear count for non-selected tabs when target is changed.
				countElem.empty();
			}
			return;
		}

		// Clear item count.
		countElem.empty();

		// Update the tab title.
		var title = tabInfo.tab.find( ".tab-title" ).text( conf.name );

		// Set the title text (hover tooltip).
		if( conf.description )
			title.attr( "title", conf.description );
		else
			title.attr( "title", conf.name );


		if( tabInfo.activeListing.error ) {

			// Set the item count on the tab button to display an error icon.
			countElem.append( "<span class='ui-icon ui-icon-alert'></span>" );

		} else if( updateItemCount === true ) {

			// Set the item count on the tab button.
			countElem.text( " (" + tabInfo.activeListing.listing.items.GetObjectVersionsCount() + ")" );
		}
	}



	/**
    * Sets up the generic tab menu structure, state and behavior as well as
	* the handlers for each option.  Each tab still needs to call this.addMenuBehavior()
	* as it is created to utilize the behavior.
	*
	* @returns {jQuery} - The menu element.
    */
	ExtendedViewSubListings.prototype.initializeSubListingMenu = function() {

		// Append to body so we don't have to deal with hidden overflow issues.
		var menu = $( ExtendedViewSubListings.MENU_TEMPLATE )
				.appendTo( "body" )
				.hide();

		// Create a reference to this for use in closures.
		var self = this;

		// Create an object to hold the menu state.
		var ms = this.tabMenuState = {
			preventMenuBlur: false,
			justOpened: false,
			menu: menu,
			target: null,
			tabTarget: null
		}

		menu.on( {

			// Bind to the mousedown event inside the menu.
			mousedown: function( event ) {

				// If the mousedown event is recieived for an option
				// we know the menu was already opened, and the user
				// is intreracting with it.  Set the flags accordingly
				// and prevent any futher event processing.
				ms.preventMenuBlur = true;
				ms.justOpened = false;
				event.stopPropagation();
				event.preventDefault();

			},

			// Bind to the mouseup event inside the menu.
			mouseup: function( event ) {

				// If the mouseup event is received for an option
				// we know something has been chose so we hide the menu 
				// and allow blur.
				$( this ).hide();
				ms.preventMenuBlur = false;
				$( ".tab-menu-button" ).blur();

				// If the menu was just opened, the mouseup should be treated as a click,
				// otherwise the mouseup event is the result of a click and nothing else should be done.
				if( ms.justOpened && $( event.target ).closest( "li" ) ) {
					setTimeout( function() {
						$( event.target ).click();
					}, 0 )
				} else {
					event.stopPropagation();
					event.preventDefault();
				}
			}

		} );



		// Bind to the menu options.


		// Menu Option: Browse to sublisting's view in external window.
		$( ".viewext-sublisting-menuitem-browse" )
			.text( l( "CMD_BROWSE_SUBLISTING_VIEW" ) )
			.click( function() {
				if( !$( this ).hasClass( "ui-state-disabled" ) ) {
					var subListing = self.subListingManager.getSubListing( ms.tabTarget.subListingID );
					self.openObjectFolderInSeparateWindow( subListing.view, self.selectedObject.ObjVer );
				}
			} );


		// Menu Option: Launch the edit sublisting dialog.
		$( ".viewext-sublisting-menuitem-configtab" )
			.text( l( "CMD_EDIT_SUBLISTING" ) )
			.click( function() {
				self.subListingManager.showEditSubListingDialog( ms.tabTarget.subListingID )
			} );


		// Menu Option: Launch the view property dialog of the sublisting's view.
		$( ".viewext-sublisting-menuitem-configview" )
			.text( l( "CMD_EDIT_SUBLISTING_VIEW" ) )
			.click( function() {
				var subListing = self.subListingManager.getSubListing( ms.tabTarget.subListingID );
				self.assertCanEditSubLisitingViewProperties( subListing );
				self.openViewProperties( subListing.view );
				var viewDef = self.vault.ViewOperations.GetView( subListing.view );
				self.subListingManager.tryEnsureValidGroupingLevels( viewDef, subListing );
				self.refreshTabListings( ms.tabTarget );
			} );


		// Menu Option: Hide the sublisting tab in this foler path.
		$( ".viewext-sublisting-menuitem-hide" )
			.text( l( "CMD_HIDE_SUBLISTING" ) )
			.click( function() {
				var subListing = self.subListingManager.getSubListing( ms.tabTarget.subListingID );
				var settings = self.folderSettingsManager.getUserFolderPathSettings( self.pathInfo );
				self.subListingManager.setSubListingVisibilityInFolderPath( settings, subListing, false );
				self.folderSettingsManager.setUserFolderPathSettings( self.pathInfo, settings );
			} );


		// Menu Option: Refresh Tab Listing.
		$( ".viewext-sublisting-menuitem-refresh" )
			.text( l( "CMD_REFRESH_SUBLISTING" ) )
			.click( function() {
				self.refreshTabListings( ms.tabTarget );
			} );


		// Menu Option: Create a new sublisting.
		$( ".viewext-sublisting-menuitem-create" )
			.text( l( "CMD_CREATE_SUBLISTING" ) )
			.click( function() {

				// Calculate base properties to target the currently selected object.
				var baseProps = {};
				if( self.selectedObject ) {
					baseProps.objType = self.selectedObject.ObjVer.Type;
					baseProps.classes = [self.selectedObject.Class];
				}

				// Show the create dialog.
				self.subListingManager.showCreateSubListingDialog( baseProps );
			} );


		// Menu Option: Manage Settings.
		$( ".viewext-sublisting-menuitem-settings" )
			.text( l( "CMD_MANAGE_SUBLISTINGS" ) )
			.click( function() {
				self.showManageSubListingsDialog();
			} );

		// Menu Option: Refresh sublistings from server.
		// Only used for debugging.
		/* 
        $( ".viewext-sublisting-menuitem-refreshAll" )
            .text( l( "CMD_REFRESH_SUBLISTING" ) )
            .click( function() {
                self.subListingManager.getAllSubListings( true );
            } );
		*/

		// Menu Option: Reset Settings.
		$( ".viewext-sublisting-menuitem-reset" )
			.text( l( "CMD_RESET_SETTINGS" ) )
			.click( function() {
				self.folderSettingsManager.promptRevertFolderPathSettings( self.pathInfo );
			} );


		// Menu Option: Save Common Settings
		$( ".viewext-sublisting-menuitem-save" )
			.text( l( "CMD_SAVE_COMMON_SETTINGS" ) )
			.click( function() {
				var settings = self.folderSettingsManager.getUserFolderPathSettings( self.pathInfo );
				self.folderSettingsManager.promptSaveCommonFolderPathSettings( self.pathInfo, settings );
			} );


		// Menu Option: Export Common Settings
		$( ".viewext-sublisting-menuitem-export" )
			.text( l( "CMD_EXPORT_SUBLISTINGS" ) )
			.click( function() {
				self.subListingManager.promptExportCommonSubListings();
			} );


		// Menu Option: Import Common Settings
		$( ".viewext-sublisting-menuitem-import" )
			.text( l( "CMD_IMPORT_SUBLISTINGS" ) )
			.click( function() {
				self.subListingManager.promptImportCommonSubListings();
			} );



		return menu;

	}


	/*
	* Binds the menu to a button.  Clicking the button will show a context menu.
	* When clicked the button is set as the menu state's target.
	* 
	* @param {jQuery} button - The click target for opening the menu.
	*/
	ExtendedViewSubListings.prototype.addMenuBehavior = function( button ) {

		var self = this;
		var ms = this.tabMenuState;
		var menu = ms.menu;


		// bind to the settings menu button
		button.on( {
			mousedown: function( event ) {

				// Hide the menu if it is already visible, and the same button was clicked (Toggle on/off).
				if( ms.target === this && menu.is( ":visible" ) ) {
					menu.hide().find( ".bgiframe" ).remove();
					return;
				}

				ms.target = this;

				// Make sure the menu is focused when it is clicked.
				var button = $( this ).focus();

				// Update the tabInfo target of the menu.
				// Answers: Which tab was the menu clicked in?
				var tab = $( event.delegateTarget ).closest( "li" );
				if( tab.find( "a" ).length )
					ms.tabTarget = self.getTabInfoForTab( tab );
				else
					ms.tabTarget = null;

				// Update the menu state for the target tab.
				self.updateMenuOptions( menu, ms.target, ms.tabTarget );

				// Use a timeout to allow other buttons to blur, 
				// which may hide the menu, before we show/update it.
				setTimeout( function() {

					// Show the options menu, and position it by the trigger button.
					// Apply the iframe hack fix so that the menu can be seen above the activex listings.
					// Default src option: javascript:flase gives an untrusted source error,
					// so we use "about:blank" instead.
					menu.show().position( {
						my: "left top",
						at: "left+4 bottom-4",
						of: button
					} ).bgiframe( {
						conditional: true,
						src: "about:blank"
					} );

				}, 20 );

				// Indicate the menu was just opened.
				ms.justOpened = true;

				// Prevent further event processing.
				event.stopPropagation();
				event.preventDefault();

			},
			blur: function( event ) {

				// If the focus on the menu trigger button is lost
				// hide the menu if we aren't prevented,
				// otherwise allow blur the next time.
				if( !ms.preventMenuBlur )
					menu.hide().find( ".bgiframe" ).remove();
				else
					ms.preventMenuBlur = false;

			}

		} ).on( "mouseup click", function( event ) {

			// Don't allow the default behavior on click or mouseup (we handle it exclusively)
			event.stopPropagation();
			event.preventDefault();
		} );


	}



	/*
	 * Updates the state of the menu's options for target tab.
	 *
	 * @param {jQuery} menu - The jQuery wrapped html node of the menu.
	 * @param {DomElement} button - The button that is launching the menu. 
	 * @param {TabInfo} tabInfo - The details of the tab related to the button. Null if not tab related.
	 */
	ExtendedViewSubListings.prototype.updateMenuOptions = function( menu, button, tabInfo ) {

		// Remove all menu options for making a tab visible.
		menu.find( ".viewext-sublisting-menuitem-tab" ).remove();

		// Hide all options by default.
		menu.children().hide()

		// Set menu as icon-less by default.
		menu.removeClass( "ui-icon-menu" );

		// Only show the "Browse View..." option if there is a valid listing.
		if( tabInfo ) {

			// Show "Tab Menu" options.

			// Show everything until spacer 1.
			menu.children( ":first" ).show().nextUntil( ".spacer-1" ).show();

			// Only show the "edit tab" option if the user has the rights.
			var subListing = this.subListingManager.getSubListing( tabInfo.subListingID );
			var editable = this.subListingManager.isSubListingEditableByCurrentUser( subListing );
			$( ".viewext-sublisting-menuitem-configtab" ).toggle( editable );

			// Only show the view property option if it is safe to browse to the view.
			var enabled = ( tabInfo.activeListing && !tabInfo.activeListing.error );
			menu.find( ".viewext-sublisting-menuitem-browse" ).toggleClass( "ui-state-disabled", !enabled );


		} else if( $( button ).is( ".new-tab-tab" ) ) {

			// Show "New Tab" menu options.

			// Set menu to show icons.
			menu.addClass( "ui-icon-menu" );

			// Show the create option.
			menu.find( ".viewext-sublisting-menuitem-create" ).show();
			var spacer = menu.find( ".spacer-1" );

			// Loop over each tab.
			var self = this;
			var tabs = this.getTabsInPriorityOrder();
			$.each( tabs, function( i, sortedTab ) {

				var tab = sortedTab.tabInfo;

				// Add an option for each applicable tab to toggle the visibility.

				// Skip non-applicable or currently visible tabs.
				if( !tab.applicable )
					return;

				// Only show the spacer now that we know there will be something above it.
				spacer.show();

				// Resolve the sublisting defintion of the current tab.
				var subListing = self.subListingManager.getSubListing( tab.subListingID );

				// Create the menu option for this tab.
				var li = $( "<li></li>" ).insertBefore( spacer )
					.text( subListing.name )
					.data( "id", subListing.id )
					.addClass( "viewext-sublisting-menuitem-tab" )
					.click( function() {

						// Tab option click handler. Toggle Visibility.
						var settings = self.folderSettingsManager.getUserFolderPathSettings( self.pathInfo );
						self.subListingManager.setSubListingVisibilityInFolderPath( settings, subListing, !tab.visible );
						self.folderSettingsManager.setUserFolderPathSettings( self.pathInfo, settings );

					} );

				// If the tab is currently visible, add the class which will show a checkmark next to it.
				if( tab.visible )
					li.addClass( "visible" );
			} );
		} else {

			// Show "Settings Menu" Options.

			// Show everything between spacer-1 and 4.
			menu.find( ".spacer-1" ).nextUntil( ".spacer-4" ).show();

			// Hide the save common settings option if the user doesn't have the rights, or it isn't a common view.
			var manageCommonSettings = this.subListingManager.viewManager && this.pathInfo.view && this.pathInfo.view.Common;
			menu.find( ".viewext-sublisting-menuitem-save" ).toggle( manageCommonSettings );

			// Show spacer 4 and everything after if the user is an admin.
			if( this.subListingManager.admin )
				menu.find( ".spacer-4" ).prev().nextAll().show();

		}

	}


	/*
	 * Launches a tab's view in a separate window.
	 *
	 * @param {Number} view - The view whose property folder to launch.
	 * @param {MFiles.ObjVer} objVer - The target object for the view's grouping level. 
	 */
	ExtendedViewSubListings.prototype.openObjectFolderInSeparateWindow = function( view, objVer ) {

		// Create a propFolders definition for the resolved object.
		var viewDef = this.vault.ViewOperations.GetView( view );
		var propFolders = this.subListingManager.getObjectPropertyFolders( objVer, viewDef.Levels );

		// Launch the view.
		var windowHandle = this.shellFrame.OuterWindow.Handle;
		this.vault.ViewOperations.ShowViewOrPropertyFolder( windowHandle, view, propFolders );
	}


	/*
	* Opens the View Properties dialog for a view.
	*
	* HACK: This method goes to great lengths to show this dialog because it can't be done with MFilesAPI.
	* Essentially we create a new listing of the view's parent view, select the view in the listing and execute
	* the "properties" built-in command.  Afterwards we try to restore the previous selection/active listing.
	*
	* @param {Number} view - The id of the view to show the dialog for.
	*/
	ExtendedViewSubListings.prototype.openViewProperties = function( view ) {

		// Resolve Active Listing, default to assuming primary listing is active.
		var activeListing = this.shellFrame.Listing;
		try {
			activeListing = this.shellFrame.ActiveListing;
			activeListing.UnselectAll();
		} catch( e ) {
			// This throws an error if no listing is active.
		}

		// Resovle view's parent listing path.
		var viewPath = this.vault.ViewOperations.GetViewLocationInClient( view, false );
		viewPath = viewUtils.getRelativeMFilesPath( this.vault, viewPath );

		// Create View parent's listing.
		var viewListing = this.viewListings[viewPath];
		if( !viewListing ) {
			viewListing = this.viewListings[viewPath] = {};
			this.justCreatedListing = true;
			viewListing.listing = this.shellFrame.CreateAdditionalListingForPath( viewPath );
			viewListing.elem = dashboardHelper.attachListing( viewListing.listing, ".viewext-viewlistings" );
		}

		// Select the view in the parent listing.
		var folderDef = new MFiles.FolderDef();
		folderDef.setView( view )

		// Select the view, refreshing the listing contents and trying again if not found the first time.
		// We wait a maximum of 5 seconds for the dialog to open, before returning an error.
		// On slower connections, it could take a while for the view to load.
		this.openViewPropertiesRetryHelper( viewListing, folderDef, activeListing, true, 100, 50 );
	}

	/**
	 * Opens the View Properties dialog for a view. Refreshing the listing and retrying
	 * as needed.  This seems to be required now that views are loaded asynchronously!?
	 * @param {MFiles.Listing} viewListing - The parent listing to select the view in.
	 * @param {MFiles.FolderDef} folderDef - The folderDef representing the view in the view listing.
	 * @param {MFiles.Listing} activeListing - The listing to return focus to when the dialog closes.
	 * @param {Boolean} allowRefresh - Indicates if the viewListing should be refreshed before a retry.
	 * @param {Number} delay - The delay between each retry.
	 * @param {Number} retries - The number of retries.
	 */
	ExtendedViewSubListings.prototype.openViewPropertiesRetryHelper = function(
		viewListing,
		folderDef,
		activeListing,
		allowRefresh,
		delay,
		retries
	) {
		// Check if there are any retries left.
		if( retries < 1 ) {

			// No retries left, call directly (no try block)
			// so any errors will be shown to the user.
			viewListing.listing.SelectFolder( folderDef );

		} else {

			// Select the view, refreshing the listing contents and trying again if not found the first time.
			try {

				// Try to select the folder.
				viewListing.listing.SelectFolder( folderDef );

			} catch( e ) {

				// Refresh the listing if applicable.
				if( allowRefresh ) {
					viewListing.listing.RefreshListing( false, true, true );
				}

				// Schedule a retry.
				var self = this;
				setTimeout( function() {
					self.openViewPropertiesRetryHelper( viewListing, folderDef, activeListing, false, --retries );
				}, delay )
				return;
			}
		}

		// Trigger the properties command for the selected listing.
		this.shellFrame.Commands.ExecuteCommand( BuiltinCommand_Properties, null );

		// Restore active listing.
		activeListing.ActivateListing();
	}

	/*
	* Launches a manage subListings dialog.
	*
	*/
	ExtendedViewSubListings.prototype.showManageSubListingsDialog = function() {

		var self = this;

		// Show the manage subListings dashboard in a popup.
		this.shellFrame.ShowPopupDashboard( "choose_sublistings", false, {

			// Pass our manager objects to the dashboard.
			subListingManager: self.subListingManager,
			folderSettingsManager: self.folderSettingsManager,
			pathInfo: self.pathInfo,

			// Pass along our method for opening a sublistings view properties.
			// Customized so we can refresh the listings after the view
			// has been edited.
			openViewProperties: function( subListing ) {

				// Open the view properties dialog. (Modal)
				self.assertCanEditSubLisitingViewProperties( subListing );
				self.openViewProperties( subListing.view );

				// Make sure the view is still in good shape.
				var viewDef = self.vault.ViewOperations.GetView( subListing.view );
				self.subListingManager.tryEnsureValidGroupingLevels( viewDef, subListing );

				// Once the dialog is closed, refresh the related subListings.
				var tabInfo = self.getSubListingTab( subListing.id );
				if( tabInfo )
					self.refreshTabListings( tabInfo );
			}
		} );
	}


	/*
	* Refreshes all subListings.
	* Useful for full refreshes  la F5
	* Refreshes the active listings synchronously.
	* Inactive listings are refreshed synchronously if 
	* refreshFromServer = false,otherwise they are refreshed asynchronously.
	* 
	* Parameters apply to the synchronous calls only and are passed directly  
	* through to ShellListing::RefreshListing()
	*
	* @param {bool} smartUpdate
	* @param {bool} refreshFromServer
	* @param {bool} forceFullRefreshFromServer
	*/
	ExtendedViewSubListings.prototype.refreshAllListings = function( smartUpdate, refreshFromServer, forceFullRefreshFromServer ) {

		// Loop over tabs.
		for( var tabID in this.tabs ) {

			// Loop over tab's listings.
			for( var listID in this.tabs[tabID].subListings ) {
				var subListing = this.tabs[tabID].subListings[listID];

				// Skip subListings that have errors, or that are missing listings.
				if( subListing.error || !subListing.listing )
					continue;

				// Handle the active listings synchronously, if we aren't refreshing from the server
				// we can also refresh the inactive ones asyncrhonously, otherwise we do those asynchronously.
				if( subListing.visible || ( !refreshFromServer && !forceFullRefreshFromServer ) )
					subListing.listing.RefreshListing( smartUpdate, refreshFromServer, forceFullRefreshFromServer );
				else
					subListing.listing.RefreshListingAsync();
			}
		}

	}


	/*
	* Refreshes a tab's subListings.
	* Useful after the sublisting's view has been edited.
	* Refreshes the active listing synchronously and all others asynchronously.
	*
	* @param {TabInfo} tabInfo - The tab whose subListing's should be refreshed.
	*/
	ExtendedViewSubListings.prototype.refreshTabListings = function( tabInfo ) {

		// Handle the active listing specially.
		if( tabInfo.activeListing ) {

			// Validate the listings path, if it is not valid it will be set as an error and be reloaded.
			this.validatePath( tabInfo.activeListing );

			// Check if the current subListing is in an error state.
			if( tabInfo.activeListing.error ) {

				// SubListing had an error, try to reload the whole object.
				this.loadSublisting( tabInfo, true );

			} else {

				// Listing is present, refresh it synchronously.
				tabInfo.activeListing.listing.RefreshListing( true, true, true );
			}
		}


		// Loop over all the sublistings.
		for( var id in tabInfo.subListings ) {
			var subListing = tabInfo.subListings[id];

			// Skip the active tab, we handled that separately.
			if( tabInfo.activeListing && tabInfo.activeListing === subListing )
				continue;

			// Validate the listings path, if it is not valid it will be set as an error and be reloaded
			// the next time it should be visible.
			this.validatePath( tabInfo.activeListing );

			// Skip errors, they will try to reload the listing when they become active.
			if( subListing.error )
				continue;

			// load listing if not already done.
			if( subListing.listing === undefined) {
				subListing = loadSublisting( tabInfo, true );
			}

			// Refresh all non-active listings in the background.
			subListing.listing.RefreshListingAsync();
		}
	}


	/*
	* Checks that there is still a valid view path for the listing.
	* If there is no valid view path for the listing anymore, it updates the panel with an error.
	* If there is a valid view path and it's changes, it updates the listing object accordingly.
	* If the path wasn't valid before (i.e., sublisting.error = true) the validity is not checked,
	* and false is immediately returned.
	*
	* @param {SubListing} sublisting - The sublisitng whose path should be checked.
	* @returns {boolean} True if still valid, false if it cannot be resolved, or has changed.
	*/
	ExtendedViewSubListings.prototype.validatePath = function( sublisting ) {

		// If the listing already has an error, don't bother with it.
		if( sublisting === undefined || sublisting.error )
			return false;

		// Resolve the tab's underlying sublisting def
		var def = this.subListingManager.getSubListing( sublisting.subListingID );

		try {

			// Try to resolve the path for the sublisting and the target object.
			var path = this.subListingManager.getSubListingPathForObject( sublisting.objectVersion.objVer, def );

			// If a path was resolved, make sure it matches the previous path.
			if( path != sublisting.path )
				sublisting.error = true;

		} catch( e ) {

			// If an error occured resolving the path, mark the sublisting invalid.
			sublisting.error = true;
		}


		return !sublisting.error;
	}


	/*
	* Removes old listings from the DOM.  This keeps memory low and prevents the
	* process from having too many GDI objects - when that happens explorer goes boom.
	*/
	ExtendedViewSubListings.prototype.cleanUpListings = function() {

		var MAX_LISTINGS = 10;

		// Find all non-active listings.
		var nonActiveListings = [];

		// Loop over tabs.
		for( var tabID in this.tabs ) {

			// Loop over tab's listings.
			for( var listID in this.tabs[tabID].subListings ) {

				// Check the listing visibility.
				if( !this.tabs[tabID].visible || !this.tabs[tabID].subListings[listID].visible ) {

					// The tab is not visible or the sublisting is not visible,
					// add the sublisting and it's tab referenec to the non-active list.
					nonActiveListings.push( {
						tabInfo: this.tabs[tabID],
						subListing: this.tabs[tabID].subListings[listID]
					} );
				}
			}
		}


		// If the non-active listings haven't exceeded the maximum
		// don't bother trying to clean up.
		if( nonActiveListings.length < MAX_LISTINGS )
			return;

		// Sort the non-active listings in order they were 
		// last activated (seen by the user).
		nonActiveListings.sort( function( a, b ) {
			return b.subListing.lastActivated - a.subListing.lastActivated;
		} );

		// Loop over all listings above the max threshhold and dispose of them.
		for( var i = MAX_LISTINGS; i < nonActiveListings.length; i++ ) {

			// Non-active listing object (contains tabInfo and sublisting references).
			var l = nonActiveListings[i];

			// Destroy the handle on the lisitng object (trying to promote early gc).
			l.subListing.listing = null;

			// Empty out the listing element, 
			// remove it from the DOM and destroy the reference.
			// This might be a little over the top but it ensures these are released if something
			// happens to still be holding a reference to the sublisting.
			l.subListing.elem.empty();
			l.subListing.elem.remove();
			l.subListing.elem = null;
			delete l.tabInfo.subListings[l.subListing.id];

			// Also remove possible direct reference to the sublisting in tabInfo. 
			if( l.tabInfo.activeListing && l.tabInfo.activeListing.id === l.subListing.id )
				l.tabInfo.activeListing = null;
		}
	}

	/*
	* Cleans up resources before being destroyed.
	* Unbinds event listeners.
	* Removes references that may potentially be external objects 
	* (e.g., created in different module/dashboard).
	*/
	ExtendedViewSubListings.prototype.cleanUp = function() {

		// Remove listeners of custom code.
		this.subListingManager.unbind( null, null, this );
		this.folderSettingsManager.unbind( null, null, this );


		// Not sure if any of this helps. But it doesn't hurt.
		for( var tabID in this.tabs ) {
			var tabInfo = this.tabs[tabID];
			tabInfo.panel.show();

			for( var subID in tabInfo.subListings ) {
				var subListing = tabInfo.subListings[subID];
				subListing.elem.show();
				subListing.listing = null;
				subListing.elem.remove();
			}
			tabInfo.panel.remove();
		}


		// Remove external references.
		this.subListingManager = null;
		this.folderSettingsManager = null;
		this.pathInfo = null;
		this.folderSettings = null;
	}

	/*
	* Makes sure the sublisting is in order before editing the view.
	*
	* @param {SubListing} subListing - The sublisitng that should be checked before editing it's view.
	* @throws exception if the sublisting's object type has been deleted.
	*/
	ExtendedViewSubListings.prototype.assertCanEditSubLisitingViewProperties = function( subListing ) {

		// Make sure that the sublisting's target object type exists.
		// If we don't do this, and show the properties, then cancelling
		// will lead to an error as we try to ensure the view is good, and 
		// saving will throw a core error, because the grouping level object
		// cannot be found.
		try {

			// Simply try to reslove the object type definition.
			var ot = this.vault.ObjectTypeOperations.GetObjectType( subListing.objType );

		} catch( e ) {

			// If there was an error, it means the Object type was not found, and we assume it
			// is deleted.  Throw a nice informative error.
			throw Error( l( "ERR_OPEN_VIEW_TARGET_OBJECT_DELETED" ) );
		}

	}

} )();