"use strict";

// Declare global where our object will be exported.
var FolderPathSettingsManager;

( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;

	/*
	* Internal helper method to shallow copy an object into another.
	* Also shallow copies array's at the first level 
	* (i.e, if base.property is an array, a copy of it will be returned in target).
	*
	* @param {object} target - The object to copy values to.
	* @param {object} base - The object copy values from.
	* @returns {object} The target object for chaining.
	*/
	function extend( target, base ) {

		// Loop over all properties in the base object.
		for( var p in base ) {

			// If the target object doesn't define that value, add it
			// to the target with the same key. Json converts undefined to null, so we need to check for it as well.
			if( target[p] === undefined || target[p] === null )

				// Target property is not defined.

				if( Object.prototype.toString.call( base[p] ) === '[object Array]' ) {

					// Property is array, clone it.
					target[p] = base[p].slice();

				} else {

					// Property is not array, just copy it.
					target[p] = base[p];
				}
		}

		// Return the (possibly modified) target object.
		return target;
	}


	/*
	* Manages the settings specific to a folder path (various levels of views).
	* @class
	*/
	FolderPathSettingsManager = function( shellUI ) {

		// Initialize members.
		this.shellUI = shellUI;
		this.vault = shellUI.vault;
		this.userID = this.vault.SessionInfo.UserID;
		this.viewManager = this.vault.SessionInfo.CheckVaultAccess( MFVaultAccessManageCommonViews );

		// Load Common settings. 
		this.common = new NamedValueSettings(
				this.vault,
				MFFolderConfiguration,
				FolderPathSettingsManager.ALIAS );

		// Load personal settings.
		// Note that we won't be auto-commiting changes.
		this.user = new NamedValueSettings(
				this.vault,
				MFUserDefinedValue,
				FolderPathSettingsManager.ALIAS + "." + this.userID,
				false );


		this.bindCommitOnShellFrameStop();

	}



	/*
	* The base alias where the folder path settings are stored in NamedValueStorage
	* @const {string}
	*/
	FolderPathSettingsManager.ALIAS = "M-Files.ViewExtensions.ViewSettings";


	/*
	* @typedef ExtendedViewSettings
	* @type {object}
	*
	* @property {Number} viewID - The id of the view which the settings are for.
	* @property {string} viewGUID - The guid of the view which the settings are for (informational).
	* @property {ExtendedFolderPathSettings[]} levels - The level specific settings for each grouping level of the view.
	* @property {string} path - The path of the view which the settings are for (informational).
	*/

	/*
	* @const {ExtendedViewSettings}
	*/
	FolderPathSettingsManager.DEFAULT_VIEW_SETTINGS = {
		viewID: 0,
		viewGuid: undefined,
		levels: {},
		path: ""
	};


	/*
	* ExtendedFolderPathSettings
	*/

	/*
	* @const {ExtendedFolderPathSettings}
	* @type {object}
	*
	* @property {Boolean} groupingPanelVisible - Indicates whether the grouping panel is visible. 
	*		(A value of -1 indicates the user has not explicity selected a state, so the system will automatically guess.)
	* @property {Boolean} groupingPanelExpanded - Indicates whether the grouping panel is expanded.
	* @property {Boolean} subListingPanelVisible - Indicates whether the sublisting panel is visible.
	* @property {Boolean} subListingPanelExpanded - Indicates whether the sublisting panel is expanded.
	* @property {Number} subListingPanelHeight - Indicates the sublisting panel height in pixels.
	* @property {String[]} subListingWhiteList - Array of non-default sublisting id's which are visible on this level.
	* @property {String[]} subListingBlackList - Array of default sublisting id's which are hidden on this level.
	* @property {Number}  lastForcedReset - The epoch time in ticks when the settings were forced to reset. For common settings, this indicates
	*		when they were last saved with the forced option, for personal views it is the time inherited from the common view when
	*		being reset.
	*/
	FolderPathSettingsManager.DEFAULT_VIEW_LEVEL_SETTINGS = {
		groupingPanelVisible: -1,
		groupingPanelExpanded: 2,
		subListingPanelVisible: false,
		subListingPanelExpanded: true,
		subListingPanelHeight: 300,
		subListingWhiteList: [],
		subListingBlackList: [],
		lastForcedReset: 0,
		lastSaved: 0
	}


	// Extend the event interface
	FolderPathSettingsManager.prototype.bind = EventInterface.bind;
	FolderPathSettingsManager.prototype.unbind = EventInterface.unbind;
	FolderPathSettingsManager.prototype.trigger = EventInterface.trigger;



	/*
	* Helper method for getting a settings context based on a common boolean parameter.
	*
	* @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	* @returns {NamedValueSettings} - The settings of the requested context.
	*/
	FolderPathSettingsManager.prototype.getContext = function( common ) {

		if( common )
			return this.common;
		else
			return this.user;
	}



	/*
	* Binds to NormalShellFrame Stop events to commit setting changes.
	* We do this to reduce server traffic (only saving when they navigate away from the location).
	*
	*/
	FolderPathSettingsManager.prototype.bindCommitOnShellFrameStop = function() {


		// Create self reference for use in callbacks.
		var self = this;

		// Listen for new normal shell frames.
		this.shellUI.Events.Register( MFiles.Event.NewNormalShellFrame, function( shellFrame ) {

			// Bind to NormalShellFrame Stop events.
			shellFrame.Events.Register( MFiles.Event.Stop, function() {

				// New normal shell frame stopped.
				try {

					// Commit any pending changes to the settings.
					self.user.commit();

				} catch( e ) {

					// TODO: Determine if it would be better to bind to start or other event.
					// Swallow errors. 
					// Happens if the vault is no longer available, e.g. Log out, nothing can be done.
				}

			} );

			// We also need to bind to the new location change event, as it now fires instead of
			// the shellframe starting/stopping.
			// We leave the other listener in place for running in legacy clients, but also for
			// when users close the window or navigate to a location outside of m-files.

			// Check if the new location change event exists.
			if( MFiles.Event.ViewLocationChanged ) {

				// Bind to the new location change event,
				shellFrame.Events.Register( MFiles.Event.ViewLocationChanged, function() {

					// New normal shell frame stopped.
					try {

						// Commit any pending changes to the settings.
						self.user.commit();

					} catch( e ) { }
				} );
			}
		} );
	}



	/**
	 * Returns all the settings for the view of the passed folderPath.
	 * @private
	 *
	 * @param {FolderPathInfo} folderPath - The folder path for which to retrieve settings.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 * @returns {ExtendedViewSettings} settings - The settings for the folderPath. Null if not found.
	 */
	FolderPathSettingsManager.prototype.getViewSettings = function( folderPath, common ) {

		// TODO: Handle special folders. Root, search, etc...

		// Throw error on null value or root view.
		if( !folderPath || !folderPath.view )
			MFiles.ThrowError( l( "ERR_NO_FOLDERPATH" ) );

		// Try to resolve settings for this view.
		var viewSettings = this.getContext( common ).settings[folderPath.view.ID];

		return viewSettings;
	}



	/**
	 * Gets a default settings object.
	 *
	 * @returns {ExtendedFolderPathSettings} Default settings object.
	 */
	FolderPathSettingsManager.prototype.getDefaultFolderPathSettings = function() {
		return extend( {}, FolderPathSettingsManager.DEFAULT_VIEW_LEVEL_SETTINGS );
	}



	/**
	 * Gets the settings for a paticular folder path.
	 * @private
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to retrieve settings.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 * @returns {ExtendedFolderPathSettings} settings - The settings for the folderPath. Nuill if not found.
	 */
	FolderPathSettingsManager.prototype.getFolderPathSettings = function( folderPath, common ) {

		// Try to resolve settins for this view.
		var viewSettings = this.getViewSettings( folderPath, common )
		if( viewSettings ) {

			// Settings resolved for the view.


			// Try to resolve setting for the current folder path level.
			var level = ( folderPath.base ) ? -1 : folderPath.depth;

			var levelSettings = viewSettings.levels[level];
			if( levelSettings ) {

				// Level settings resolved.

				// Ensure the level settings has all known properties, and return.
				return extend( levelSettings, FolderPathSettingsManager.DEFAULT_VIEW_LEVEL_SETTINGS );
			}
		}

		return null;
	}



	/**
	 * Persists settings for a specific context and folder path.
	 * @private
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to store the settings.
	 * @param {ExtendedFolderPathSettings} pathSettings - The settings for the folderPath.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 * @return {ExtendedFolderPathSettings} The stored settings object. Will not be the same settings object passed in.
	 */
	FolderPathSettingsManager.prototype.setFolderPathSettings = function( folderPath, pathSettings, common ) {

		// Fail gracefully if we are no longer logged in.
		if( !this.vault.LoggedIn )
			return null;

		// Trigger the cancelable 'BeforeUpdatePathSettings' event.
		if( !this.trigger( "BeforeUpdatePathSettings", [folderPath, pathSettings, common], true ) )
			return null;

		// Resolve or create a settings object for the entire view.
		var viewSettings = this.getViewSettings( folderPath, common );
		if( !viewSettings ) {
			viewSettings = extend( {}, FolderPathSettingsManager.DEFAULT_VIEW_SETTINGS );
		}

		// Update view setting parameters.
		viewSettings.viewID = folderPath.view.ID;
		viewSettings.viewGuid = folderPath.view.GUID;

		try {

			// Add path to settings so they can be interpreted in NVS and exports.
			// Use try/catch because this fails for views with wildcards in their path 
			// (i.e. when view setting: Advanced->Appearance = "In all folders of this level")
			// and there is currently no API method to detect those or create a nice path description with the wildcards.
			var fullViewPath = this.vault.ViewOperations.GetViewLocationInClient( folderPath.view.ID, true );
			viewSettings.path = viewUtils.getRelativeMFilesPath( this.vault, fullViewPath );

		} catch( e ) {
			
			// Swallow error. 
			// The path is an informational property only, so it is safe if an error occurred resolving it.			
		}

		// Set the settings object for the folderPath level.
		// Ensuring all necessary keys are specified.
		var level = ( folderPath.base ) ? -1 : folderPath.depth;
		viewSettings.levels[level] = extend( pathSettings, FolderPathSettingsManager.DEFAULT_VIEW_LEVEL_SETTINGS );

		// Set the save time of the level settings.
		viewSettings.levels[level].lastSaved = ( new Date() ).getTime();

		// Save updated viewSettings in namedvaluestorage.
		viewSettings = this.getContext( common ).save( viewSettings.viewID, viewSettings );
		pathSettings = viewSettings.levels[level];

		// Trigger the 'AfterUpdatePathSettings' event.
		this.trigger( "AfterUpdatePathSettings", [folderPath, pathSettings, common], true )

		// Return the persisted path settings.
		return pathSettings;

	}


	/**
	 * Removes settings for a specific context and folder path.
	 * @private
	 *
	 * @param {FolderPathInfo} folderPath - The path of the settings to be removed.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 */
	FolderPathSettingsManager.prototype.removeFolderPathSettings = function( folderPath, common ) {

		var viewSettings = this.getViewSettings( folderPath, common );

		// Nothing to remove if no view settings were found, be done.
		if( !viewSettings )
			return;

		// Set the settings object for the folderPath level.
		// Ensuring all necessary keys are specified.
		var level = ( folderPath.base ) ? -1 : folderPath.depth;

		// If no level is defined, nothing to remove, be done.
		if( !viewSettings.levels || !viewSettings.levels[level] )
			return;

		// Remove the folderPath level's settings.
		delete viewSettings.levels[level];

		// Save the updated viewSettings in namedvaluestorage.
		viewSettings = this.getContext( common ).save( viewSettings.viewID, viewSettings );
	}


	/**
	 * Returns the common settings for the specified folderPath.
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to retrieve common settings.
	 * @returns {ExtendedFolderPathSettings} settings - The common settings for the folderPath.
	 */
	FolderPathSettingsManager.prototype.getCommonFolderPathSettings = function( folderPath ) {

		// Resolve the common settings for this folder.
		var settings = this.getFolderPathSettings( folderPath, true );

		// If no settings were found use the default settings.
		if( !settings )
			settings = this.getDefaultFolderPathSettings();

		return settings;
	}



	/**
	 * Persists common settings for the specified folderPath. Also updates the user's personal settings
	 * if the forceReset option is true, so any subsequent personal changes don't get inadvertantly reverted 
	 * to the now common settings. 
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to store the common settings.
	 * @param {ExtendedFolderPathSettings} settings - The common settings to save.
	 * @param {bool} forceReset - Indicates whether all users' settings should be reset.
	 * @return {ExtendedFolderPathSettings} The stored settings object. Will not be the same settings object passed in. 
	 *		Null if cancelled.
	 */
	FolderPathSettingsManager.prototype.setCommonFolderPathSettings = function( folderPath, settings, forceReset ) {

		// Handle forced reset, if requested.
		if( forceReset ) {

			// Update the reset time in the settings.
			settings.lastForcedReset = ( new Date() ).getTime();

			// Save the settings as personal.
			var userSettings = this.setFolderPathSettings( folderPath, settings, false )

			// Return early if the user setting saving was cancelled.
			if( userSettings === null )
				return null;
		} else {

			// We need to ensure that the previous forced reset time is maintained (even if they are setting things back to default).
			var commonSettings = this.getCommonFolderPathSettings( folderPath );
			settings.lastForcedReset = commonSettings.lastForcedReset;

		}


		// Save the common settings and return the results.
		return this.setFolderPathSettings( folderPath, settings, true );
	}



	/**
	 * Returns the user settings for the specified folderPath.
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to retrieve the user specific settings.
	 * @returns {ExtendedFolderPathSettings} settings - The user specific settings for the folderPath.
	 */
	FolderPathSettingsManager.prototype.getUserFolderPathSettings = function( folderPath ) {

		// Resolve the user and common settings for this folder.
		var userSettings = this.getFolderPathSettings( folderPath, false );
		var commonSettings = this.getCommonFolderPathSettings( folderPath );

		// Check whether the common settings should be used. If any of the following are true:
		// 1. If no user settings exist
		// 2. If a common lastForcedReset is specified is greater than the user settings' last saved time.
		if( !userSettings || commonSettings.lastForcedReset > userSettings.lastSaved ) {

			// Reset the user's settings to the common ones.
			userSettings = this.setUserFolderPathSettings( folderPath, commonSettings );
			// this.user.commit();
		}

		// Return a clone of the settings.
		return extend( {}, userSettings );
	}



	/**
	 * Persists user settings for the specified folderPath.
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to store the user specific settings.
	 * @param {ExtendedViewSettings} settings - The user specific settings for the folderPath.
	 * @return {ExtendedFolderPathSettings} The stored settings object. Will not be the same settings object passed in.
	 */
	FolderPathSettingsManager.prototype.setUserFolderPathSettings = function( folderPath, settings ) {

		return this.setFolderPathSettings( folderPath, settings, false );
	}


	/**
	 * Prompts a user to confirm saving common settings for the specified folderPath,
	 * and provides the option to reset all user settings.
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to store the settings.
	 * @param {ExtendedViewSettings} settings - The common settings to save for the folderPath.
	 * @return {ExtendedFolderPathSettings} The stored settings object. Will not be the same settings object passed in. 
	 *		Null if cancelled by user.
	 */
	FolderPathSettingsManager.prototype.promptSaveCommonFolderPathSettings = function( folderPath, settings ) {

		// Make sure the user has common view rights.
		if( !this.viewManager )
			MFiles.ThrowError( l( "ERR_PERMISSION_DENIED" ) );

		// Prompt to delete sublisting and it's corresponding view.
		var prompt = {
			caption: l( "WINCAPTION_SAVE_COMMON_VIEWEXT_SETTINGS" ),
			message: l( "MSG_CONFIRM_SAVE_COMMON_VIEWEXT_SETTINGS" ),
			button1_title: l( "BUTTON_OK" ),
			button2_title: l( "BUTTON_CANCEL" ),
			defaultButton: 2,
			cancelButton: 2,
			checkbox_title: l( "CHECKBOX_LABEL_REMOVE_PERSONAL_SETTINGS" ),
			checkbox_checked: false
		};
		var result = this.shellUI.ShowMessage( prompt );

		// Return right away if the user cancelled the operation.
		if( result == 2 )
			return null;

		// User confirmed.

		// Save the common settings.
		return this.setCommonFolderPathSettings( folderPath, settings, prompt.checkbox_checked );

	}


	/**
	 * Prompts a user to confirm resetting the settings for the specified folderPath,
	 * and provides options to of which settings to revert to (as well as cancel).
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to revert the settings.
	 * @return {ExtendedFolderPathSettings} The stored settings object. Null if cancelled by user.
	 */
	FolderPathSettingsManager.prototype.promptRevertFolderPathSettings = function( folderPath ) {

		// Prompt to reset settings, and provide reset options.
		// Disable the common setting option if the folderPath is in a private view.
		var result = this.shellUI.ShowMessage( {
			caption: l( "WINCAPTION_RESET_VIEWEXT_SETTINGS" ),
			message: l( "MSG_RESET_VIEWEXT_SETTINGS" ),
			button1_title: l( "BUTTON_REVERT_TO_COMMON" ),
			button1_enabled: ( folderPath.view && folderPath.view.Common ),
			button2_title: l( "BUTTON_REVERT_TO_DEFAULT" ),
			button3_title: l( "BUTTON_CANCEL" ),
			defaultButton: 3,
			cancelButton: 3
		} );

		// Return right away if the user cancelled the operation.
		switch( result ) {

			case 1: // Revert to common settings
				return this.revertToCommonFolderPathSettings( folderPath );

			case 2: // Revert to default settings
				return this.revertToDefaultFolderPathSettings( folderPath );

			case 3: // Cancel
			default: // ???
				return null;
		}

	}


	/**
	 * Resets a user's folder path settings back to the common ones.
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to reset the settings.
	 * @return {ExtendedFolderPathSettings} The stored settings object. Null if cancelled through event callback.
	 */
	FolderPathSettingsManager.prototype.revertToCommonFolderPathSettings = function( folderPath ) {

		// Trigger the cancelable 'BeforeRevertToCommonPathSettings' event.
		if( !this.trigger( "BeforeRevertToCommonPathSettings", [folderPath], true ) )
			return null;

		// Load the common settings.
		var commonSettings = this.getCommonFolderPathSettings( folderPath );

		// Save the common settings as the user's settings.
		var pathSettings = this.setUserFolderPathSettings( folderPath, commonSettings);

		// Commit the changes right away.
		this.user.commit();

		// Trigger the 'AfterUpdatePathSettings' event to indicate the settings may have changed.
		this.trigger( "AfterUpdatePathSettings", [folderPath, pathSettings, false], true )

		return pathSettings;
	}


	/**
	 * Resets a user's folder path settings back to the default ones.
	 *
	 * @param {FolderPathInfo} folderPath - The path for which to reset the settings.
	 * @return {ExtendedFolderPathSettings} The stored settings object. Null if cancelled through event callback.
	 */
	FolderPathSettingsManager.prototype.revertToDefaultFolderPathSettings = function( folderPath ) {

		// Trigger the cancelable 'BeforeRevertToCommonPathSettings' event.
		if( !this.trigger( "BeforeRevertToDefaultPathSettings", [folderPath], true ) )
			return null;

		// Resolve default settings.
		var defaultSettings = this.getDefaultFolderPathSettings();

		// Save the default settings as the user's settings. 
		// We do this instead of removing setting so they override any common settings.
		var pathSettings = this.setUserFolderPathSettings( folderPath, defaultSettings );

		// Make sure the changes are commited to the server.
		this.user.commit();

		// Trigger the 'AfterUpdatePathSettings' event to indicate the settings may have changed.
		this.trigger( "AfterUpdatePathSettings", [folderPath, pathSettings, false], true )

		return pathSettings;
	}

} )();