"use strict";

// Declare global where our object will be exported.
var ExtendedViewListing;

( function() {

	/**
	 * Classes used by the control.
	 * @const
	 */
	var CLASS = {
		PANEL_TOP: "viewext-panel-top",
		GROUP_LISTING: "viewext-group-listing",
		MAIN_LISTING: "viewext-main-listing",
		RESIZE_BAR: "resize-bar",
		PANEL_BOTTOM: "viewext-panel-bottom",
		SUB_LISTINGS: "viewext-sublistings"
	}

	/**
	* Helper method to resolve a cached theme from the ThemeEditor app.
	* @param {MFiles.Vault} vault
	* @return {object|null} The theme, if resolved. 
	*/
	function getCachedTheme( vault ) {		

		var theme = null;

		// Registry Keys and Value Names. DO NOT LOCALIZE.
		var REG_KEY = "ThemeEditor\\%VAULT%";
		var REG_VAL_ENABLED = "Enabled";
		var REG_VAL_CACHED_THEME = "CachedTheme";

		var key = REG_KEY.replace( "%VAULT%", vault.Name );

		// Check if the theme editor is disabled, and if so, return null.
		var enabled = MFiles.ReadFromRegistry( true, key, REG_VAL_ENABLED );
		if ( enabled === 0 && enabled === "0" )
			return null;

		// Resolve the theme file path.
		var themePath = MFiles.ReadFromRegistry( true, key, REG_VAL_CACHED_THEME );
		if( !themePath )
			return null;		

		try {

			// Read theme file and deserialize it.
			var fso = new ActiveXObject( 'Scripting.FileSystemObject' );
			var stream = fso.OpenTextFile( themePath, 1 ); // 1 = ForReading
			var content = stream.ReadAll();
			theme = JSON.parse( strip( content ) );

		} catch( e ) {

			// Swallow error. Issues with themeing aren't serious enough to disable behavior.
			// MFiles.ReportException( e );

		} finally {

			if( stream )
				stream.Close();
		}

		return theme;
	}



	/*
	 * Controls the view extension listing dashboard.
	 * @class
	 */
	ExtendedViewListing = function( dashboard, container ) {

		// Initialize members.
		this.shellFrame = dashboard.Parent;
		this.shellUI = this.shellFrame.ShellUI;
		this.vault = this.shellUI.Vault;


		// Use the persistent managers passed from the shellUI module.
		this.folderPathSettingsManager = dashboard.CustomData.folderPathSettingsManager;
		this.appSettingsManager = dashboard.CustomData.appSettingsManager;
		this.subListingManager = dashboard.CustomData.subListingManager;
		this.pathInfo = dashboard.CustomData.pathInfo;


		// Load any custom cached theme that might be set by the ThemeEditor application.
		this.theme = getCachedTheme( this.vault );

		/*
		// Use Dashboard created managers for easy debugging.
		this.folderPathSettingsManager = new FolderPathSettingsManager( this.shellUI );
		this.appSettingsManager = new ViewExtensionSettingsManager( this.shellUI );
		this.subListingManager = new SubListingManager( this.shellUI, this.appSettingsManager );
		this.pathInfo = viewUtils.getFolderPathInfo( this.vault, this.shellFrame.CurrentPath, this.shellFrame.CurrentFolder );
		*/

		this.taskPaneGroup = dashboard.CustomData.taskPaneGroup;

		// Initialize structure in the passed container
		// And then bind members to each component element.
		this.$ = $( container ).html( ExtendedViewListing.TEMPLATE );


		// Initialize group level panel.
		this.groupListingContainer = this.$.find( "." + CLASS.GROUP_LISTING );
		this.groupListingControl = new ExtendedViewGroupListing(
				this.shellFrame,
				this.groupListingContainer,
				this.folderPathSettingsManager,
				this.pathInfo,
				this.taskPaneGroup,
				this.theme );

		// Initialize Sublistings panel.
		this.subLisitingsContainer = this.$.find( "." + CLASS.SUB_LISTINGS );
		this.sublistingsControl = new ExtendedViewSubListings(
				this.shellFrame,
				this.subLisitingsContainer,
				this.subListingManager,
				this.folderPathSettingsManager,
				this.pathInfo,
				this.taskPaneGroup,
				this.theme );

		// Initialize Resize bar.
		this.initResizeBar();

		// Initialize main listing
		dashboardHelper.attachListing( this.shellFrame.Listing, this.$.find( "." + CLASS.MAIN_LISTING ) );

		// Create a closure bound reference to ourselves.
		var self = this;

		// Update layout once now, and whenever window is resized.
		$( window ).resize( function() {
			self.updateLayout();
		} )
		this.updateLayout();


		// Listen for the dashboard to close.
		dashboard.Events.Register( MFiles.Event.Stop, function() {
			self.cleanUp();
		} );


		// Bind to key events and route them appropriately.
		$( window ).keydown( function( evt ) {

			// Check if F5 key was pressed.
			if( evt.keyCode == 116 ) {

				// Any time F5 is pressed we automatically refresh all sublistings.
				self.sublistingsControl.refreshAllListings( false, true, !!evt.shiftKey );

				// If the primary lising isn't the active listing, we need to manually update it 
				// and the grouping level listing, if the primary listing is active they will 
				// automatically be updated when we delegate to the built-in refresh commands.
				if( self.shellFrame.ActiveListing !== self.shellFrame.Listing ) {
					self.shellFrame.Listing.RefreshListing( false, true, !!evt.shiftKey );
					self.groupListingControl.refreshAllListings( false, true, !!evt.shiftKey );
				}

				// Delegate the F5 key press to the built-in refresh or full-refresh commands
				// based on whether the shift key was also pressed.
				var cmd = ( evt.shiftKey ) ? MFiles.BuiltInCommand.RefreshFull : MFiles.BuiltInCommand.Refresh;
				self.shellFrame.Commands.ExecuteCommand( cmd, null );
			}

		} );


		// Listen for certain built-in commands to respond to.
		this.shellFrame.Commands.Events.Register( MFiles.Event.BuiltInCommand, function( cmd ) {

			// Check if a refresh command was triggered.
			if( cmd === MFiles.BuiltInCommand.Refresh || cmd === MFiles.BuiltInCommand.RefreshFull ) {

				// Refresh the group level listings if the primary listing is active (being refreshed).
				if( self.shellFrame.ActiveListing === self.shellFrame.Listing )
					self.groupListingControl.refreshAllListings( false, true, ( cmd === MFiles.BuiltInCommand.RefreshFull ) );
			}
		} );



		// Auto-launch the manage sublistings dialog, if the relevent flag from the module is set.
		if( dashboard.CustomData.launchSettingsDialog ) {
			this.sublistingsControl.showManageSubListingsDialog();
		}
	}

	/**
	 * Basic HTML structure of the control.
	 * @const {string}
	 */
	ExtendedViewListing.TEMPLATE = "<div class='viewext-panel-top'>" +
										"<div class='viewext-group-listing'></div>" +
										"<div class='viewext-main-listing'></div>" +
									"</div>" +
									"<div class='viewext-panel-bottom'>" +
										"<div class='viewext-sublistings'></div>" +
									"</div>";

	/**
	 * Adds drag/resize behavior to the bottom panel's resize bar.
	 *
	 */
	ExtendedViewListing.prototype.initResizeBar = function() {

		var self = this;

		// Initialize the jQuery UI draggable interaction on the resize bar.
		var resizeBar = $( "<div class='resize-bar'></div>" )
			.prependTo( this.sublistingsControl.$ )
			.draggable( {
				axis: "y",
				iframeFix: true,
				helper: "clone",
				appendTo: "body",
				scroll: false,
				start: function( event, ui ) {

					// Resize bar dragging started.

					// Apply the iframe hack fix so that the resize bar,
					// can be seen above the activex listings.
					// Default src option: javascript:false gives an untrusted source error,
					// so we use "about:blank" instead.
					ui.helper.bgiframe( {
						conditional: true,
						src: "about:blank"
					} );

					// Bind to the iframe's mouse up event to 
					// pass on to the actual helper, otherwise
					// the dragging will never stop.
					var iframe = $( "iframe" ).get( 0 )
					var doc = iframe.contentDocument || iframe.contentWindow.document;
					doc.addEventListener( "mouseup", function() {
						ui.helper.trigger( "mouseup" );
					} );

				},
				stop: function( event, ui ) {

					// Resize bar dragging finished.

					// Resize detail section and then have
					// the whole layout adjust to it.
					var winHeight = $( window ).height();
					self.sublistingsControl.height = winHeight - ui.offset.top;
					$( ".viewext-sublistings" ).height( self.sublistingsControl.height );

					self.updateLayout();
				}
			} );

	}

	/**
	 * Updates the layout by calculating and enforcing the size of the panels.
	 * Primarily used to fill (but not overflow) the available vertical space.
	 *
	 */
	ExtendedViewListing.prototype.updateLayout = function() {

		var minHeight = 150;

		// Resolve current window height.
		var winHeight = $( window ).height();

		// Resolve the grouping level panel height (not resizable/scrollable).
		var groupObjects = $( ".viewext-group-listing" )
		var groupObjHeight = ( groupObjects.is( ":visible" ) ) ? groupObjects.height() : 0;

		// Resolve the current sublisting panel height.
		var otherListing = $( ".viewext-sublistings" );
		var otherListingsHeight = ( otherListing.is( ":visible" ) ) ? otherListing.height() : 0;

		// Calculate main listing height.
		var mainHeight = winHeight - otherListingsHeight;
		var defaultListingHeight = mainHeight - groupObjHeight;


		// Do a bunch of extra calculations if the bottom panel is visible.
		if( otherListingsHeight > 0 && this.sublistingsControl.expanded ) {

			// Recalculate max heights if available space is limited.
			var wiggleRoom = winHeight - groupObjHeight
			if( wiggleRoom < minHeight * 2 )
				minHeight = wiggleRoom / 2;

			// Resize otherListing panel as necessary to enforce minumum sizes.
			if( defaultListingHeight < minHeight ) {
				var resize = minHeight - defaultListingHeight;
				otherListingsHeight -= resize;
				defaultListingHeight = minHeight;
				$( ".viewext-sublistings" ).height( otherListingsHeight );
			} else if( otherListingsHeight < minHeight ) {
				var resize = minHeight - otherListingsHeight;
				defaultListingHeight -= resize;
				otherListingsHeight = minHeight
				$( ".viewext-sublistings" ).height( otherListingsHeight );
			}

			// Contstrain the movement of the resize bar so the panel doesn't get too small/large.
			// An array defining a bounding box in the form [ x1, y1, x2, y2 ].
			var maxTop = groupObjHeight + minHeight // Make sure the main listing is always 100px high.
			var maxBottom = winHeight - minHeight // Make sure the sublisting panel is always 100px high.
			$( ".resize-bar:not(.ui-draggable-dragging)" ).draggable(
					"option", "containment", [0, maxTop, 10000, maxBottom] );
		}


		// Set main listing height.
		$( ".viewext-main-listing" ).height( defaultListingHeight );


		// Delegate to sub controls.
		this.sublistingsControl.updateLayout();
	}


	/*
	* Cleans up resources before being destroyed.
	* Removes references that may potentially be external objects 
	* (e.g., created in different module/dashboard).
	* Delegates cleanup call to sub-controls.
	*
	*/
	ExtendedViewListing.prototype.cleanUp = function() {

		this.groupListingControl.cleanUp();
		this.sublistingsControl.cleanUp();

		this.appSettingsManager = null;
		this.subListingManager = null;
		this.folderSettingsManager = null;
		this.pathInfo = null;
		this.folderSettings = null;

	}

} )();