"use strict";

var ViewExtensionLocalization;

( function() {

	/* @const {string} The registry key path to the current m-files software language setting. */
	var REG_MFILES_LANG_KEY_FORMAT = "HKLM\\SOFTWARE\\Motive\\M-Files\\%version%\\Language";

	/* @const {string} The application's default language identifier */
	var DEFAULT_LANGUAGE = "eng";

	/* @const {string} The path (relative to the application root) where the alternative language strings are stored */
	var LANGUAGE_FILE_PATH = "translations\\";


	/*
	* Default localization function.
	* All localization should be stored on the constructor.
	*
	* @param {string} id - A string id.
	*
	*/
	function localizer( id ) {

		// Find the string defined by the string id in the current language.
		var raw;
		if( localizer[localizer.currentLanguage] )
			raw = localizer[localizer.currentLanguage][id];

		// If no string was found, simply replace it with the id value.
		if( !raw )
			return id;

		// Replace placeholders with subsequent function parameters.		
		var args = arguments;
		var argIndex = 0;
		var str = raw.replace( /%s/gi, function( match, offset, str ) {

			// Increment which arguments this placeholder matches.
			// First placeholder value is "1" to skip the first id arg.
			argIndex++;

			// Make sure a parameter was passed for this placeholder.
			if( argIndex >= args.length )
				throw Error( "Localization string argument missing.\n\nReference: " + id + "\n\n" + str );

			// Return the parameter as the replacement text.
			return args[argIndex];
		} );

		// Make sure all passed parameters were used as placeholders.
		if( argIndex + 1 < args.length )
			throw Error( "Too many localization string arguments passed.\n\nReference: " + id + "\n\n" + raw );

		return str;

	}

	/* 
	* Initializes the localizer method by setting and loading the necessary language and its strings.
	* Needs to be called only once per environment instance or whenever the current language may have changed.
	* Should only be called once the global MFiles object is ready, as that is used to get the client version
	* and lookup the client's current language.
	*
	*/
	localizer.init = function() {

		// Load the default language.
		localizer.loadLanguage( DEFAULT_LANGUAGE );

		// Resovle the m-file version running.
		var mf = new MFiles.MFilesClientApplication();

		// Process as Web or Desktop?
		if( MFiles.CurrentApplicationPlatform === MFExtApplicationPlatformWeb ) {
			// Load from the GetClientLanguage function.
			localizer.currentLanguage = mf.GetClientLanguage() || DEFAULT_LANGUAGE;
		} else {
			// Load from the Client Machine Registry.
			localizer.version = mf.GetClientVersion().Display;

			// Resolve the language registry key for this m-files client version.
			localizer.key = REG_MFILES_LANG_KEY_FORMAT.replace( "%version%", localizer.version );

			try {

				// Read the language registry value, falling back to the default language if nothing is specified.
				var shell = new ActiveXObject( "WScript.Shell" );
				localizer.currentLanguage = shell.RegRead( localizer.key ) || DEFAULT_LANGUAGE;

			} catch( e ) {

				// Fall back to default language if there was an error reading the language registry value.
				localizer.currentLanguage = DEFAULT_LANGUAGE;
			}
		}

		// Load the current language, falling back to the default if no translation could be loaded.
		if( !localizer.loadLanguage( localizer.currentLanguage ) )
			localizer.currentLanguage = DEFAULT_LANGUAGE;

		// Resolve
		return localizer;
	}


	/*
	* Attempts to set the current language to the one passed.  The current language
	* is set to the default one if the passed lanugage strings could not be loaded.
	*
	* @param {string} lang - The language identifier (should match the one used by m-files)
	* @returns {boolean} True if the current language was set to the desired language, false if set to the default.
	*/
	localizer.setCurrentLanguage = function( lang ) {

		// Attempt to load the language strings.
		if( localizer.loadLanguage( lang ) ) {

			// Language strings successfully loaded.

			// Set the current language.
			localizer.currentLanguage = lang;
			return true;

		} else {

			// Language strings could not be loaded.

			// Use default language.
			localizer.currentLanguage = DEFAULT_LANGUAGE;
			return false;
		}
	}


	/*
	* Attempts to load the strings of a particular language from the language file path,
	* if not already loaded or if the reload option is specified.
	*
	* @param {string} lang - The language identifier (should match the one used by m-files)
	* @param {boolean} reload - Indicates if the language should be re-loaded if it is loaded already.
	* @returns {boolean} True if the language's strings are now loaded, false if they could not be loaded.
	*/
	localizer.loadLanguage = function( lang, reload ) {

		// Don't bother loading if it is already loaded.
		if( !reload && localizer[lang] )
			return true;

		// Language hasn't been loaded, try to load it now.

		// Resolve the file location.
		var path = MFiles.ApplicationPath + LANGUAGE_FILE_PATH + lang + ".json";

		// See if a file exists for the language.
		var fso = new ActiveXObject( 'Scripting.FileSystemObject' );
		if( fso.FileExists( path ) ) {

			// Read in the selected file contents.			
			var stream = fso.OpenTextFile( path, 1 );  // 1 = ForReading
			var content = stream.ReadAll();
			stream.Close();

			// Parse the file contents into a json object and store it in the localizer.
			localizer[lang] = JSON.parse( strip( content ) );

			return true;
		}

		return false;
	}



	/* 
	* Copies all the default language strings to a file in the language file directory for translation.
	* Utility method.
	*
	* @param {string} lang - The language for which a translation file should be created (affects the file name)
	* @returns {string} The path to the newly created file.
	*/
	localizer.createTranslationFile = function( lang ) {

		// Serialize the default language strings.
		var content = JSON.stringify( localizer[DEFAULT_LANGUAGE], null, 4 ).replace( /\n/g, "\r\n" );

		// Resolve the output file path.
		var path = MFiles.ApplicationPath + LANGUAGE_FILE_PATH + lang + ".json";

		// Write the content to file.
		var fso = new ActiveXObject( 'Scripting.FileSystemObject' );
		var stream = fso.OpenTextFile( path, 2, true );  // 2 = ForWriting, Create = true
		stream.Write( content );
		stream.Close();


		// Return the new file path.
		return path;
	}

	// Define global localizer object if not already done.
	if( !ViewExtensionLocalization ) {
		ViewExtensionLocalization = localizer;
	}

	// Set this file's string as the current language.
	var lang = ViewExtensionLocalization.currentLanguage = "eng";

	// Initialize this language's strings.
	ViewExtensionLocalization[DEFAULT_LANGUAGE] = {

		ERR_NOT_FOUND: "Not Found.",
		ERR_PERMISSION_DENIED: "Permission denied.",

		BUTTON_CANCEL: "Cancel",
		BUTTON_OK: "OK",


		SUBLISTING_DEFAULT_NAME: "New Detail Listing",
		WINCAPTION_DELETE_SUBLISTING: "Delete",
		CONFIRM_DELETE_SUBLISTING: "Are you sure you want to delete the detail listing \"%s\" and its corresponding view?",
		BUTTON_DELETE_SUBLISTING: "Delete",
		VIEW_NAME_COMMON_SUBLISTING_VIEW: "Detail Listings",
		VIEW_NAME_PERSONAL_SUBLISTING_VIEW: "Personal Detail Listings",

		WINCAPTION_IMPORT_SUBLISTINGS: "Import",
		CONFIRM_IMPORT_SUBLISTINGS: "To import: %s\n\nTo update: %s\n\nTo discard: %s",
		BUTTON_IMPORT_SUBLISTINGS: "Import",
		ERR_IMPORT_INVALID_FILE: "Invalid import file.",
		ERR_IMPORT_NO_SUBLISTINGS: "There are no common listings to export.",
		ERR_SUBLISTING_VIEW_NOT_FOUND: "The view for the detail listing \"%s\" was not found.",
		ERR_TARGET_OBJECT_NOT_FOUND: "The target object for the detail listing \"%s\" was not found.",
		ERR_TARGET_CLASS_NOT_FOUND: "One or more of the target object classes for the detail listing \"%s\" was not found.",
		FILE_FILTER_JSON: "JSON files (*.json)|*.json|Text files (*.txt)|*.txt|All files (*.*)|*.*",

		WINCAPTION_CREATE_SUBLISTING: "Create Detail Listing",
		WINCAPTION_EDIT_SUBLISTING: "Edit Detail Listing",

		SUBLISTING_STATUS_LABEL_VISIBLE: "Visible",
		SUBLISTING_STATUS_VISIBLE: "Visible",
		SUBLISTING_STATUS_HIDDEN: "Hidden",
		SUBLISTING_TYPE_COMMON: "Common",
		SUBLISTING_TYPE_PERSONAL: "Personal",

		SUBLISTING_LABEL_NAME: "Name",
		SUBLISTING_LABEL_DESCRIPTION: "Description",
		SUBLISTING_LABEL_TYPE: "Type",
		SUBLISTING_LABEL_OBJTYPE: "Object Type",
		SUBLISTING_LABEL_CLASSES: "Classes",
		SUBLISTING_LABEL_PRIORITY: "Priority",
		SUBLISTING_LABEL_DEFAULT_VISIBILITY: "Default",

		SUBLISTING_CAPTION_TARGET_OBJECTS: "Target Objects",
		SUBLISTING_CAPTION_TAB: "Tab",

		SUBLISTING_OPTION_COMMON: "Common to all users",
		SUBLISTING_OPTION_ALL_CLASSES: "All classes",
		SUBLISTING_OPTION_SHOW_BY_DEFAULT: "Show by default",
		ELEMENT_DELETED: "(deleted)",
		ELEMENT_HIDDEN: "(hidden)",

		ERR_SUBLISTING_NAME_REQUIRED: "You must specify a name for the detail listing.",
		ERR_SUBLISTING_CLASS_REQUIRED: "You must select at least one class.",
		WARN_OBJTYPE_CANNOT_BE_GROUPED: "A detail listing cannot be created for the currently selected object. The object type or at least one of the properties that can refer to it must be configured to be used as a grouping level.",
		WARN_NO_GROUPABLE_OBJTYPES: "Detail listings cannot be created because no object types (or properties referring to them) have been configured to be used as grouping levels.",

		SUBLISTINGS_TOGGLE_VISIBILITY_INSTRUCTIONS: "Use the \"Visible\" checkbox to control the visibility of a detail listing in this location. " +
												"Use the \"Reset\" button to revert all the listings back to their default visibility in this location.",

		CMD_RESET_SUBLISTING_VISIBILITY: "Reset",
		CMD_CREATE_SUBLISTING: "Create...",
		CMD_EDIT_SUBLISTING: "Edit...",
		CMD_EDIT_SUBLISTING_VIEW: "Edit View...",
		CMD_DELETE_SUBLISTING: "Delete",
		CMD_HIDE_SUBLISTING: "Hide",
		CMD_IMPORT_SUBLISTINGS: "Import Common Listings",
		CMD_EXPORT_SUBLISTINGS: "Export Common Listings",
		CMD_SAVE_COMMON_SETTINGS: "Save as Common Settings",
		CMD_RESET_SETTINGS: "Reset Settings",
		CMD_BROWSE_SUBLISTING_VIEW: "Browse in New Window...",
		CMD_REFRESH_SUBLISTING: "Refresh",
		CMD_MANAGE_SUBLISTINGS: "Manage Settings",

		ERR_NO_FOLDERPATH: "No folder path or view.",

		// These have been copied from m-files core dialogs.
		WINCAPTION_SAVE_COMMON_VIEWEXT_SETTINGS: "Save As Common View Extension Settings",
		MSG_CONFIRM_SAVE_COMMON_VIEWEXT_SETTINGS: "Your current view extension settings in this view will be saved as the common default settings for all users.",
		CHECKBOX_LABEL_REMOVE_PERSONAL_SETTINGS: "Remove all users' custom settings in this view?",

		// These have been copied from m-files core dialogs.
		WINCAPTION_RESET_VIEWEXT_SETTINGS: "Reset View Extension Settings",
		MSG_RESET_VIEWEXT_SETTINGS: "You can reset the view extension settings of this location to the common settings that have been specified by the administrator, or to the default settings of the application.",
		BUTTON_REVERT_TO_COMMON: "Reset to Common Settings",
		BUTTON_REVERT_TO_DEFAULT: "Reset to Default Settings",

		TASKPANEGROUP_VIEW_EXTENSIONS: "View Extensions",


		HEADER_SUBLISTING_NO_OBJECT_SELECTED: "No Object Selected",
		CMD_SHOW_SUBLISTING_PANEL: "Show Detail Listings",
		CMD_HIDE_SUBLISTING_PANEL: "Hide Detail Listings",
		ERR_NO_SELECTED_OBJECT: "No selected object.",
		ERR_LISTING_LOAD_FAILED: "The listing could not be loaded.\n\nError: %s",
		BUTTON_SHOW_ERROR_DETAILS: "Show Details",


		HEADER_GROUPING_LEVELS: "Grouping Levels",
		HEADER_GROUPING_LEVEL_OBJECTS: "Grouping Level Objects",
		CMD_SHOW_GROUP_LEVEL_PANEL: "Show Grouping Levels",
		CMD_HIDE_GROUP_LEVEL_PANEL: "Hide Grouping Levels",

		GROUP_LEVEL_MULTIPLE_ITEMS: "(multiple items)",
		GROUP_LEVEL_EMPTY: "(empty)",

		ERR_NO_OBJECT_PATH: "No path could be found for the object version.",
		ERR_NO_GROUPING_LEVEL: "The view \"%s\" (ID: %s) does not have a grouping level.",
		ERR_TOO_MANY_GROUPING_LEVELS: "The view \"%s\" (ID: %s) has more than one grouping level.",
		ERR_GROUPING_LEVEL_MISMATCH: "The grouping level does not match the object type of the target object.",

		// Should never be seen (internal)
		ERR_INVALID_GROUPING_LEVELS: "Invalid grouping level configuration.",

		ERROR_PROPERTY_FOLDER_NOT_FOUND: "A property folder could not be found for the detail listing.",
		WARN_SHOW_EMPTY_FOLDERS_DISABLE: "Possible cause: The option \"Show empty folders\" is not enabled in the grouping level of the view.",
		WARN_STRICT_GROUPING_LEVEL_FILTER: "Possible cause: The grouping level filter for the view is too strict.",
		WARN_SHARED_NAME: "Possible cause: The property folder exists for multiple objects that have the same name.",

		// Developer Errors (only useful during development - should never be seen by end users).
		ERR_INVALID_LISTING_CONTAINER: "The container must be a single node or a jQuery selector that resolves to a single node.",
		ERR_DEV_ORIG_SUBLISITNG_EDITIED: "Original sublisting edited. Changes cannot be detected.",

		WINCAPTION_MANAGE_SUBLISTINGS: "Manage Detail Listings",
		DEFAULT_SUBLISTING_VIEW_NAME: "%s by %s",
		ERR_OPEN_VIEW_TARGET_OBJECT_DELETED: "The view cannot be edited because the target object type for the detail listing has been deleted."
	};

} )();