"use strict";

// Declare global where our object will be exported.
var ViewExtensionSettingsManager;

( function() {

	/*
	* Manages the view extension application's global settings.
	* 
	* @class
	* @param {MFiles.ShellUI} shellUI - The shellUI in which the application is running.
	*/
	ViewExtensionSettingsManager = function( shellUI ) {

		// Initialize members.
		this.shellUI = shellUI;
		this.vault = shellUI.vault;
		this.userID = this.vault.SessionInfo.UserID;
		this.viewManager = this.vault.SessionInfo.CheckVaultAccess( MFVaultAccessManageCommonViews );

		// Load settings.
		this.common = new NamedValueSettings( this.vault, MFFolderConfiguration, ViewExtensionSettingsManager.ALIAS );
		this.user = new NamedValueSettings( this.vault, MFUserDefinedValue, ViewExtensionSettingsManager.ALIAS + "." + this.userID );
		this.registrySettings = this.getRegistrySettings();

	}

	/*
	* The base alias where the view extension settings are stored in NamedValueStorage
	* @const {string}
	*/
	ViewExtensionSettingsManager.ALIAS = "M-Files.ViewExtensions";

	/*
	* Loads settings from the registry.
	* @returns {object} - Settings read from the registry.
	*/
	ViewExtensionSettingsManager.prototype.getRegistrySettings = function() {
		
		/* 
			Settings in are deployed from named value storage. For lazy loading the setting is as follows:
			
			Path in registry:
				HKEY_CURRENT_USER\SOFTWARE\Motive\M-Files\<m-files-version>\Client\Common\<vault-name>\ViewExtension
			
			Example for named value storage configuration:			
				Namespace: M-Files.Core.Client.Settings
				Key: ViewExtension
				Value:
					{ 
						"Common": { 
							"ViewExtension": { 
								"LazyLoadDetailListings" : true
							} 
						} 
					} 
		*/		
		
		// Initialize the default settings (lazy loading of detail listing enabled => better performance).
		var settings = { lazyLoading : true };
		try {

			// Try to read value from registry.
			var registryPath = 'Client\\Common\\' + this.vault.Name + '\\ViewExtension';
			var regValue = MFiles.ReadFromRegistry( true, registryPath, "LazyLoadDetailListings" );
			settings.lazyLoading = regValue || regValue === undefined ? true : false;

		} catch ( e ) {
			// No-op, keep initialized value.
		}
		
		return settings;
	}
	
	/*
	* Helper method for getting a settings context based on a common boolean parameter.
	*
	* @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	* @returns {NamedValueSettings} - The settings of the requested context.
	*/
	ViewExtensionSettingsManager.prototype.getContext = function( common ) {

		if( common )
			return this.common;
		else
			return this.user;
	}


	/*
	 * Returns the key's related settings.
	 *
	 * @param {string} key - The key of the settings to find.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 * @returns {object} The object/value found at the key. Returns undefined if no value was found.
	 */
	ViewExtensionSettingsManager.prototype.value = function( key, common ) {
		return this.getContext( common ).settings[key];
	}


	/*
	 * Saves a setting object to a particular key in these settings.
	 *
	 * @param {string} key - The key where the settings are to be stored.
	 * @param {object} setting - The value to store.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 * @return {object} The stored value (clone of the one passed in).;
	 */
	ViewExtensionSettingsManager.prototype.save = function( key, setting, common ) {
		return this.getContext( common ).save( key, setting );

	}


	/*
	 * Deletes the key and its related settings.
	 *
	 * @param {string} key - The key of the settings to delete.
	 * @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	 */
	ViewExtensionSettingsManager.prototype.remove = function( key, common ) {
		this.getContext( common ).remove( key );
	}

} )();