"use strict";

// Declare global where our object will be exported.
var SubListingManager;


( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;


	/*
	* Internal helper method to shallow copy an object into another.
	*
	* @param {object} target - The object to copy values to.
	* @param {object} base - The object copy values from.
	* @returns {object} The target object for chaining.
	*/
	function extend( target, base ) {

		// Loop over all properties in the base object.
		for( var p in base ) {

			// If the target object doesn't define that value, add it
			// to the target with the same key.
			if( target[p] === undefined )
				target[p] = base[p];
		}

		// Return the (possibly modified) target object.
		return target;
	}


	/*
    * Converts arrays passed from dashboards to local arrays
    * to avoid weird behavior.
    *
    * @param {array} arr - The array that may have been created by dashboard code.
    * @returns {array} A locally created array that contains all values of the passed array
    *       with stored with numeric keys. Indexes, length and order are not necessarily preserved.
    */
	function castToLocalArray( arr ) {

		// Create our local array.
		var localArr = [];

		// Loop over all members of the passed array.
		for( var i in arr ) {

			// Try to convert the key to an integer.
			var index = parseInt( i );

			// If the key was successfully converted to an integer,
			// add it to the local array.
			if( !isNaN( index ) && arr[i] !== undefined && arr[i] !== null )
				localArr.push( arr[i] );
		}

		// Return the local array.
		return localArr;
	}


	/*
	* Internal helper method to keep the branch logic readable.
	* Determines if a value is in an array.
	*
	* @param {string} value - The value to search for in the array.
	* @param {Array} arr - The array to search.
	* @returns {number} The first position of the value in the array, or -1 if not found.
	*/
	function inArray( value, arr ) {
		for( var i = 0; i < arr.length; i++ ) {
			if( arr[i] === value )
				return i;
		}
		return -1;
	}


	/*
	* Internal helper method to keep the branch logic readable.
	* Adds a value to an array, if it isn't there already.
	*
	* @param {string} value - The value to add to the array.
	* @param {Array} list - The array the value should be added to.
	* @returns {boolean} True if the value was added, false otherwise.
	*/
	function addToList( value, list ) {
		if( inArray( value, list ) === -1 ) {
			list.push( value );
			return true;
		}
		return false;
	}

	/*
	* Internal helper method to keep the branch logic readable.
	* Removes a value from an array, if found in it.
	*
	* @param {string} value - The value to remove from the array.
	* @param {Array} list - The array the value should be removed from.
	* @returns {boolean} True if the value was removed, false otherwise.
	*/
	function removeFromList( value, list ) {
		var index = inArray( value, list );
		if( index !== -1 ) {
			list.splice( index, 1 );
			return true;
		}
		return false;
	}


	/*
	* Manages all sublisting operations.
	*
	* @class
	* @param {MFiles.ShellUI} The shellUI in which the manager is running.
	* @param {ViewExtensionSettingsManager} appSettings - A application's settings manager.
	*/
	SubListingManager = function( shellUI, appSettings ) {
		this.shellUI = shellUI;
		this.vault = shellUI.vault;
		this.appSettings = appSettings;
		this.userID = this.vault.SessionInfo.UserID;
		this.viewManager = this.vault.SessionInfo.CheckVaultAccess( MFVaultAccessManageCommonViews );
		this.admin = this.isAdmin();

		// Load settings.
		this.common = new NamedValueSettings(
				this.vault, MFFolderConfiguration, SubListingManager.SUBLISTING_ALIAS );

		this.user = new NamedValueSettings(
				this.vault, MFUserDefinedValue, SubListingManager.SUBLISTING_ALIAS + "." + this.userID );

		// Convenience member for looping over all settings contexts.
		this.contexts = {
			"common": this.common,
			"user": this.user
		}

	}


	/*
	* The base alias where the sublistings are stored in NamedValueStorage
	* @const {string}
	*/
	SubListingManager.SUBLISTING_ALIAS = "M-Files.ViewExtensions.SubListings";


	/*
	* @typedef SubListingDefinition
	* @type {object}
	*
	* @property {string} id - The unique id of the sublisting.
	* @property {string} name - The title of the sublisting (appears in the tab).
	* @property {string} description - A description of the sublisting.
	* @property {Number} objType - The id of the object type for which this sublisting is targeted.
	* @property {Number[]} classes - The ids of the classes for which this sublisting is targeted (empty = any class).
	* @property {Number} view - The id of the view which the sublisting is bound to.
	* @property {Boolean} isCommon - Indicates whether the sublisting is common or user specific.
	* @property {Boolean} isDefault - Indicates whether the sublisting should be displayed everywhere by default. 
	* @property {Number} priority - The sort priority, affecting where this sublisting is displayed amongst others.
	*/
	SubListingManager.DEFAULT_SUBLISTING = {
		id: undefined,
		name: l( "SUBLISTING_DEFAULT_NAME" ),
		description: "",
		objType: -1,
		classes: [],
		view: undefined,
		isCommon: false,
		isDefault: true,
		priority: 50,
		lastModified: 0
	};



	// Extend the event interface
	SubListingManager.prototype.bind = EventInterface.bind;
	SubListingManager.prototype.unbind = EventInterface.unbind;
	SubListingManager.prototype.trigger = EventInterface.trigger;


	/*
	* Helper method for getting a settings context based on a common boolean parameter.
	*
	* @param {bool} [common] - Indicates the settings context (True=Common, False = User). Default: False (User).
	* @returns {NamedValueSettings} - The settings of the requested context.
	*/
	SubListingManager.prototype.getContext = function( common ) {

		if( common )
			return this.common;
		else
			return this.user;
	}


	/*
	* Gets all sublistings, both common and personal.
	*
	* @param {boolean} refreshFromServer - Indicates that the view access and sublistings should be refreshed from server.
	* @returns {SubListingDefinition[]} - Associative array (javacsript object) of sublistings indexed by their id.
	*/
	SubListingManager.prototype.getAllSubListings = function( refreshFromServer ) {

		// Get a list of all the views this user has access to.
		// We will hide all sublistings whose views aren't in this list.
		// We cache the views unless a refresh has been requested.
		if( refreshFromServer || !this.allViewsByID ) {
			this.allViewsByID = {};
			var views = this.vault.ViewOperations.GetViews( MFViewCategoryNormal, true, -1 );
			for( var i = 1; i <= views.Count; i++ ) {
				var view = views.Item( i );
				this.allViewsByID[view.ID] = view;
			}
		}


		// If we are to refresh the sublistings, we get a handle
		// on the old list so we can send events for the changes 
		// detected after the refresh.
		if( refreshFromServer )
			var oldSublistings = this.getAllSubListings();



		// Merge all sublistings with accessible views into a new exportable object.
		var sublistings = {};

		// Loop over the contexts (common/user specific).
		for( var c in this.contexts ) {

			// Reload the context's sublistings if it was requested.
			if( refreshFromServer )
				this.contexts[c].loadAll();

			// Loop over the sublistings in this context.
			for( var id in this.contexts[c].settings ) {

				// Get a handle on the sublisting.
				var subListing = this.contexts[c].settings[id];

				// Add it to our exportable object if the user is an admin or the sublisting's
				// view is available to the current user.
				if( subListing && ( this.admin || this.allViewsByID[subListing.view] ) ) {
					sublistings[id] = subListing;
				}
			}
		}


		// Trigger events if the sublistings were reloaded.
		if( refreshFromServer ) {			

			// Create an object that describes what has changed after the refresh.
			// Use id indexed objects instead of arrays to avoid strange array 
			// issues across js engines.
			var changes = {
				removed: {},
				added: {},
				modified: {}
			}

			// Loop over the old sublistings list.
			for( var oldID in oldSublistings ) {

				// Check if this sublisting has been removed.
				if( !sublistings[oldID] ) {

					// This sublisting was deleted.
					changes.removed[oldID] = oldSublistings[oldID];
				}
			}

			// Loop over the new sublistings.			
			for( var newID in sublistings ) {

				// Check if this sublisting is new or has been modified.
				if( !oldSublistings[newID] ) {

					// This sublisting was just added.
					changes.added[newID] = sublistings[newID];

				} else if ( oldSublistings[newID].lastModified != sublistings[newID].lastModified ) {

					// The sublisting was been modified.
					changes.modified[newID] = sublistings[newID];
				}				
			}

			// Trigger the "AfterRefreshSubListings" with the changes.
			this.trigger( "AfterRefreshSubListings", [ changes ] );

		}

		// Return the exportable list.
		return sublistings;
	}



	/*
	* Generates a unique id as a GUID.
	* @returns {string} - A unique id string.
	*/
	SubListingManager.prototype.generateUUID = function() {

		var d = new Date().getTime();
		var uuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace( /[xy]/g, function( c ) {
			var r = ( d + Math.random() * 16 ) % 16 | 0;
			d = Math.floor( d / 16 );
			return ( c == 'x' ? r : ( r & 0x3 | 0x8 ) ).toString( 16 );
		} );
		return uuid;
	}


	/*
	* Gets a sublisting definition by id.
	*
	* @param {string} - The id of the sublisting to return.
	* @returns {SubListingDefinition} - The sublisting definition.
	* @throws An error if the sublisting is not found.
	*/
	SubListingManager.prototype.getSubListing = function( id ) {

		for( var c in this.contexts ) {
			var subListing = this.contexts[c].settings[id];
			if( subListing )
				return subListing;
		}

		MFiles.ThrowError( l( "ERR_NOT_FOUND" ) );
	}

	/*
	* Creates a copy of a sublisting definition for editing.
	*
	* @param {string} [id] - Optional. The id of the sublisting to copy. If no id is specified a generic definition is returned.
	* @returns {SubListingDefinition} - The sublisting definition for editing.
	* @throws An error if the id is specified, but the sublisting is not found.
	*/
	SubListingManager.prototype.getSubListingForEditing = function( id ) {

		// Create subListing object.
		var subListing = {};

		// Copy in existing subListing properties.
		if( id )
			subListing = extend( subListing, this.getSubListing( id ) );

		// Make sure the default sublisting name uses the current language.
		SubListingManager.DEFAULT_SUBLISTING.name = l( "SUBLISTING_DEFAULT_NAME" );

		// Fill in any missing properties.
		subListing = extend( subListing, SubListingManager.DEFAULT_SUBLISTING )

		return subListing;
	}

	/*
	* Creates a new sublisting from a sublisting defintiion.
	* Generates a unique id and view.
	*
	* @param {SubListingDefinition} - The sublisting definition to create.
	* @returns {SubListingDefinition} - The created sublisting.
	*/
	SubListingManager.prototype.createSubListing = function( subListing ) {

		// Create a unique id for the sublisting, if one isn't already specified.
		if( !subListing.id )
			subListing.id = this.generateUUID();

		// Trigger the cancelable 'BeforeCreateSubListing' event.
		if( !this.trigger( "BeforeCreateSubListing", [subListing], true ) )
			return null;

		// Create a dedicated view for the sublisting, if one isn't already specified.
		if( !subListing.view ) {
			var view = this.createViewForSubListing( subListing );
			subListing.view = view.ID;
		}

		// Set last modified time.
		subListing.lastModified = (new Date()).getTime();

		// Save the sublisting. Note: The save methods return a different object.
		var context = this.getContext( subListing.isCommon );
		subListing = context.save( subListing.id, subListing );

		// Trigger the 'AfterCreateSubListing' event.
		this.trigger( "AfterCreateSubListing", [subListing, view] );

		return subListing;
	}


	/*
	* Updates a sublisting from a sublisting defintiion.
	*
	* @param {SubListingDefinition} subListing - The sublisting definition to update.
	* @returns {SubListingDefinition} - The updated sublisting.
	*/
	SubListingManager.prototype.updateSubListing = function( subListing ) {

		// Ensure we can resolve a previous version of the subListing,
		// and that the defintion itself was not edited in place.
		// This is crucial for detecting and reactiing to context changes.
		var oldSubListing = this.getSubListing( subListing.id );
		if( subListing === oldSubListing )
			MFiles.ThrowError( l( "ERR_DEV_ORIG_SUBLISITNG_EDITIED" ) );

		// Trigger the cancelable 'BeforeModifySubListing' event.
		if( !this.trigger( "BeforeModifySubListing", [subListing, oldSubListing], true ) )
			return null;

		// Set last modified time.
		subListing.lastModified = (new Date()).getTime();

		// Simply save the sublisting. Note: The save methods return a different object.
		var context = this.getContext( subListing.isCommon );
		subListing = context.save( subListing.id, subListing );


		// Check if the subListing context has changed.
		if( oldSubListing.isCommon != subListing.isCommon ) {

			// Context has changed!

			// Switch the underlying view's context and parent view accordingly.
			var view = this.vault.ViewOperations.GetView( subListing.view );
			var parentView = this.getParentSubListingView( subListing.isCommon )
			if( view.Common != subListing.isCommon || view.Parent != parentView.ID ) {
				view.Common = subListing.isCommon;
				view.Parent = parentView.ID;
				view = this.vault.ViewOperations.UpdateView( view );
			}

			// Delete the old sublisting definition from the other context.
			this.getContext( oldSubListing.isCommon ).remove( oldSubListing.id );
		}


		// Trigger the 'AfterModifySubListing' event.
		this.trigger( "AfterModifySubListing", [subListing] );

		return subListing;

	}


	/*
	* Deletes a sublisting defintiion.
	*
	* @param {string} id - The id of the sublisting definition to delete.
	* @param {boolean} confirm - Indicates whether a confirmation prompt should be shown to the user. Default: false.
	* @returns {boolean} - True if the sublisting was deleted, false otherwise.
	* @throws An error if the sublisting is not found.
	*/
	SubListingManager.prototype.deleteSubListing = function( id, confirm ) {

		// Resolve the sublisting.
		var subListing = this.getSubListing( id );

		// Check permissions.
		if( subListing.isCommon && !this.viewManager )
			MFiles.ThrowError( l( "ERR_PERMISSION_DENIED" ) );

		// Trigger the cancelable 'BeforeDeleteSubListing' event.
		if( !this.trigger( "BeforeDeleteSubListing", [subListing], true ) )
			return false;

		// Confirm the deletion, if desired.
		if( confirm && !this.promptConfirmDeleteSubListing( subListing ) )
			return false;

		// Delete the sublisting.
		this.getContext( subListing.isCommon ).remove( id );

		// Delete the corresponding view.
		try {
			this.vault.ViewOperations.RemoveView( subListing.view );
		} catch( e ) {
			// This doesn't leave anything in a critical/unstable state.
			// The user may have already deleted the view.
		}


		// Trigger the 'AfterDeleteSubListing' event.
		this.trigger( "AfterDeleteSubListing", [subListing] );

		return true;
	}

	/*
	* Prompts the user with a delete sublisting confirmation dialog.
	*
	* @param {SubListingDefinition} subListing - The sublisting definition being deleted.
	* @returns {boolean} - True if confirmed, false otherwise.
	*/
	SubListingManager.prototype.promptConfirmDeleteSubListing = function( subListing ) {

		// Prompt to delete sublisting and it's corresponding view.
		var res = this.shellUI.ShowMessage( {
			caption: l( "WINCAPTION_DELETE_SUBLISTING" ),
			message: l( "CONFIRM_DELETE_SUBLISTING", subListing.name ),
			icon: "question",
			button1_title: l( "BUTTON_DELETE_SUBLISTING" ),
			button2_title: l( "BUTTON_CANCEL" ),
			defaultButton: 2,
			cancelButton: 2
		} );

		// Return true to continue, false if cancelled.
		return ( res != 2 );
	}


	/*
	* Creates a view for the sublisting.
	*
	* @param {SubListingDefinition} subListing - The sublisting for which the view should be created.
	* @returns {MFiles.View} - The created view.
	*/
	SubListingManager.prototype.createViewForSubListing = function( subListing ) {

		// Get or create a parent view.
		var parentView = this.getParentSubListingView( subListing.isCommon );
		var objType = this.vault.ObjectTypeOperations.GetObjectType( subListing.objType );
		var name = l( "DEFAULT_SUBLISTING_VIEW_NAME", subListing.name, objType.NameSingular);

		// Define the subListing view.
		var viewDef = new MFiles.View();
		viewDef.Name = name;
		viewDef.Visible = true;
		viewDef.Common = subListing.isCommon;
		viewDef.ViewType = MFViewTypeNormal;
		viewDef.HasParent = true;
		viewDef.Parent = parentView.ID;

		// Define the default grouping levels for the view based on the subListing's object type.
		var propsOrObjs = new MFiles.PropertyDefOrObjectTypes();
		propsOrObjs.Add( -1, viewUtils.getBestObjTypeGroupingDef( this.vault, objType ) );
		viewDef.Levels = this.createViewGroupingLevels( propsOrObjs );

		// Create the view, updating the name until we don't get a name already exists error.
		var i = 0;
		var MAX_RENAME_ATTEMPTS = 100;
		var MAX_VIEW_NAME_LENGTH = 99;
		while( i < MAX_RENAME_ATTEMPTS ) {

			try {

				// Try to create the view. And end the loop if successful.
				viewDef = this.vault.ViewOperations.AddView( viewDef );
				break;

			} catch( e ) {

				// View creation failed;

				if( !e.message.indexOf( "0x8004007E" ) !== -1 ) {

					// Name conflict, make the name unique.

					// Create a name suffix with an incremented index.
					var suffix = " (" + ++i + ")";

					// Make sure the name and the index won't overflow the view name length.					
					if( name.length + suffix.length > MAX_VIEW_NAME_LENGTH )
						name = name.substr( 0, MAX_VIEW_NAME_LENGTH - suffix.length );

					// Update the name.
					viewDef.Name = name + suffix;

				} else {

					// Unknown error. Rethrow it.
					throw e;
				}
			}
		}

		// Add the view to our internal cache.
		this.allViewsByID[viewDef.ID] = viewDef;

		return viewDef;

	}

	/*
	* Checks if a view's grouping levels are as they should be. 
	*
	* @param {MFiles.View} view - The current view definition.
	* @param {SubListingDefinition} subListing - The sublisting for which the view exists.
	* @returns {boolean} - True if the view's grouping levels are valid.
	*/
	SubListingManager.prototype.validateViewGroupingLevels = function( view, subListing ) {

		// Verify there are at least 3 grouping levels in the view.
		if( view.Levels.Count < 3 ) {
			return false;
		}

		// Get a handle on each level's expression.
		var expr1 = view.Levels.Item( 1 ).Expression;
		var expr2 = view.Levels.Item( 2 ).Expression;
		var expr3 = view.Levels.Item( 3 ).Expression;

		// Verify  the first expression is ID Segment based, and the second is ID based.
		// Both should have at least 1 indirection level and they should have the same
		// number of indirection levels. The 3rd level should have one less indirection level.
		if( expr1.Type !== MFExpressionTypeObjectIDSegment ||
			expr1.IndirectionLevels.Count < 1 ||
			expr2.Type !== MFExpressionTypeStatusValue ||
			expr2.DataStatusValueType !== MFStatusTypeObjectID ||
			expr1.IndirectionLevels.Count !== expr2.IndirectionLevels.Count ||
			expr1.IndirectionLevels.Count - 1 !== expr3.IndirectionLevels.Count )
			return false;

		// Turn third level's expression into the final indirection level for comparison with the others.
		var finalL3 = new MFiles.PropertyDefOrObjectType();
		if( expr3.Type === MFExpressionTypePropertyValue ) {

			// This is a propertyDef folder
			finalL3.PropertyDef = true;
			finalL3.ID = expr3.DataPropertyValuePropertyDef;

		} else if( expr3.Type === MFExpressionTypeTypedValue && expr3.DataTypedValueDatatype === MFDatatypeLookup ) {

			// This is a Object Type based folder that doesn't points to the target object type.
			finalL3.PropertyDef = false;
			finalL3.ID = expr3.DataTypedValueValueList;

		} else {

			// Invalid expression for third grouping level.
			return false;
		}

		// Loop over each indirection level depth.
		for( var i = 1; i <= expr1.IndirectionLevels.Count; i++ ) {

			// Get a handle on all grouping levels' indirection level at this depth.
			var l1 = expr1.IndirectionLevels.Item( i );
			var l2 = expr2.IndirectionLevels.Item( i );
			var l3 = ( expr3.IndirectionLevels.Count < i ) ? finalL3 : expr3.IndirectionLevels.Item( i );

			// Verify that this indirection level in all expressions are the same.	
			if( l1.ID !== l2.ID || l1.PropertyDef !== l2.PropertyDef ||
				l1.ID !== l3.ID || l1.PropertyDef !== l3.PropertyDef ) {

				// Levels don't match.
				return false;

			} else if( i === expr1.IndirectionLevels.Count ) {

				// This is the last indirection level defined. 

				// Verify this level points to the sublisting's
				// target object type. Only check one expression, 
				// we know the others are exactly the same at this point.
				if( l1.PropertyDef ) {

					// PropertyDef based expression.  

					// Lookup propertydef and verify its value list points to sublisting's Object type.
					var pd = this.vault.PropertyDefOperations.GetPropertyDef( l1.ID );
					if( !pd.BasedOnValueList || pd.ValueList !== subListing.objType )
						return false;

				} else if( l1.ID !== subListing.objType ) {

					// ObjType based expression points to wrong Object type.
					return false;
				}
			}
		}

		// If no issues were found, we can be done.
		return true;

	}


	/*
	* Tries to fix a view's grouping levels.
	* 
	* The first grouping level with an expression that points directly  
	* to the target object type (via object or property), starting from  
	* the deepest level, will be used as the template for all the levels.
	*
	* @param {MFiles.View} view - The current view definition.
	* @param {SubListingDefinition} subListing - The sublisting for which the view exists.
	* @returns {MFiles.View} - The new view definition, or null if it could not be updated..
	*/
	SubListingManager.prototype.fixViewGroupingLevels = function( view, subListing ) {

		// Don't try anything if the user doesn't have rights to update the view.
		if( view.Common && !this.viewManager )
			return null;

		var newLevels = null;

		// Loop over the existing grouping levels in the view backwards.
		for( var i = view.Levels.Count; i > 0 ; i-- ) {

			// Get a handle on this level's expression.
			var expr = view.Levels.Item( i ).Expression;

			var finalLevel = null;

			// Check expression type.
			if( expr.Type === MFExpressionTypePropertyValue ) {

				// This is a propertyDef folder.

				// See if the property def points to the target object type.
				var pd = this.vault.PropertyDefOperations.GetPropertyDef( expr.DataPropertyValuePropertyDef );
				if( pd.BasedOnValueList && pd.ValueList === subListing.objType ) {

					// This PropertyDef based folder's PropertyDef points to the target object type.

					// Create the indirection level template for our levels.
					finalLevel = new MFiles.PropertyDefOrObjectType();
					finalLevel.PropertyDef = true;
					finalLevel.ID = pd.ID;

				}
			} else if( expr.Type === MFExpressionTypeTypedValue &&
				expr.DataTypedValueDatatype === MFDatatypeLookup &&
				expr.DataTypedValueValueList === subListing.objType ) {

				// This is a Object Type based folder that points to the target object type.

				// Create the indirection level template for our levels.
				finalLevel = new MFiles.PropertyDefOrObjectType();
				finalLevel.PropertyDef = false;
				finalLevel.ID = expr.DataTypedValueValueList;
			}

			// Check if a final level was defined.
			if( finalLevel ) {

				// A final level of indirection was defined.
				// The current expression can be used as a template.

				// Create the grouping levels for the view.
				var propDefsOrObjTypes = expr.IndirectionLevels.Clone();
				propDefsOrObjTypes.Add( -1, finalLevel );
				newLevels = this.createViewGroupingLevels( propDefsOrObjTypes );
				break;
			}

		}

		// Check if new levels have been created.
		if( newLevels ) {

			// New grouping levels have been generated.

			// Let's update the view.
			view.Levels = newLevels;
			return this.vault.ViewOperations.UpdateView( view );

		} else {

			// TODO: Figure out if we should leave the levels broken,
			// or revert to default levels for the object type.

			// Don't know how to fix.  Leave broken.
			return null;
		}

	}

	/*
	* Checks a view's grouping levels are as expected, and if not tries to set them right.
	*
	* @param {MFiles.View} view - The current view definition.
	* @param {SubListingDefinition} subListing - The sublisting for which the view exists.
	* @returns {MFiles.View} - The resultant view definition (will have been updated if the levels were changed).
	*/
	SubListingManager.prototype.tryEnsureValidGroupingLevels = function( view, subListing ) {

		// Check if the view grouping levels are valid.
		if( !this.validateViewGroupingLevels( view, subListing ) ) {

			// Something is wrong with the view grouping levels.

			// Try to fix the view if we can.
			var updatedView = this.fixViewGroupingLevels( view, subListing );

			// Update the view reference if it was fixed.
			if( updatedView )
				view = updatedView;
		}

		return view;

	}


	/*
	* Creates grouping levels for a sublisting view based on property/object expression (supporting indirection levels).
	*
	* The propsOrObjs should be the normal indirection levels + the expression's 
	* DataPropertyValuePropertyDef or DataTypedValueValueList added at the end.
	*
	* @param {MFiles.PropertyDefOrObjectTypes} propDefsOrObjTypes - The expression levels that resolve to the target grouping object.
	* @returns {MFiles.ExpressionExs} - The grouping levels for the sublisting view.
	*/
	SubListingManager.prototype.createViewGroupingLevels = function( propDefsOrObjTypes ) {

		// Initialize return value.
		var levels = new MFiles.ExpressionExs();

		var idSegmentRangeSize = 500;

		var noOp = new MFiles.DataFunctionCall();
		noOp.SetDataNoOp();

		// Create the first grouping level based on grouping object's ID segment.
		// This is important so we don't run into property folder overflow issues when
		// there are too many objects that could be targets.
		var level1 = new MFiles.ExpressionEx();
		level1.ShowEmptyFolders = true;
		level1.Expression.SetObjectIDSegmentExpression( idSegmentRangeSize );
		level1.Expression.IndirectionLevels = propDefsOrObjTypes.Clone();
		levels.Add( -1, level1 );

		// Create the second grouping level based on grouping object's ID.
		// This is important so we don't get property folders for more than
		// one object (if there are name collisions).
		var level2 = new MFiles.ExpressionEx();
		level2.ShowEmptyFolders = true;
		level2.Expression.SetStatusValueExpression( MFStatusTypeObjectID, noOp );
		level2.Expression.IndirectionLevels = propDefsOrObjTypes.Clone();
		levels.Add( -1, level2 );

		// Create the final grouping level on the object itself.
		// This is important for:
		//	1. Auto-filling properties based on the view.
		//	2. Getting the object to appear in the grouping level listing.
		//  3. Clearly seeing what object is referenced in the path.
		var level3 = new MFiles.ExpressionEx();
		level3.ShowEmptyFolders = true;

		// Set the primary expression with the last propDefsOrObjType specified.
		var objRef = propDefsOrObjTypes.Item( propDefsOrObjTypes.Count )
		if( objRef.PropertyDef )
			level3.Expression.SetPropertyValueExpression( objRef.ID, MFParentChildBehaviorNone, noOp );
		else
			level3.Expression.SetTypedValueExpression( MFDatatypeLookup, objRef.ID, MFParentChildBehaviorNone, noOp );

		// Set the indirection levels with the rest of the propDefsOrObjType
		// (removing the one we used for the primary expression).
		level3.Expression.IndirectionLevels = propDefsOrObjTypes.Clone();
		level3.Expression.IndirectionLevels.Remove( propDefsOrObjTypes.Count );
		levels.Add( -1, level3 );

		return levels;
	}


	/**
	* Resolves a target object's sublisting folder path.
	*
	* @param {MFiles.ObjVer} objVer - The objVer of the object to find the grouping folder path of.
	* @param {SubListingDefinition} subListing - The sublisting for which the view exists.
	*/
	SubListingManager.prototype.getSubListingPathForObject = function( objVer, subListing ) {

		// Get a handle on the view's definition.
		var view = this.vault.ViewOperations.GetView( subListing.view );

		// Sanity check. Make sure the view has a grouping level.
		if( view.Levels.Count === 0 )
			throw Error( l( "ERR_NO_GROUPING_LEVEL", view.Name, viewID ) );

		// Try to fix any minor issues with the grouping levels or massage an unseen edit.
		this.tryEnsureValidGroupingLevels( view, subListing );

		// Look up the property folders for the target object in this view.
		var propFolders = this.getObjectPropertyFolders( objVer, view.Levels );

		try {

			// Look up the relative path for the view and property folder and return it.
			var path = this.vault.ViewOperations.FindPropertyFolderLocationInView( view.ID, propFolders );

		} catch( e ) {

			var errHelpers = [];

			// The view could not be found.  Try to resolve the reason to include in the error message.

			// Loop over the levels.
			for( var i = 1; i <= view.Levels.Count; i++ ) {

				// Make sure the grouping levels have the "Show Empty Folder" option enabled.
				if( !view.Levels.Item( i ).ShowEmptyFolders ) {
					errHelpers.push( l( "WARN_SHOW_EMPTY_FOLDERS_DISABLE" ) );
				}

				// Make sure the grouping levels have no conditions.
				if( view.Levels.Item( 1 ).Conditions.Count > 0 ) {
					errHelpers.push( l( "WARN_STRICT_GROUPING_LEVEL_FILTER" ) );
				}

			}

			// Add extra info to the error message.
			if( errHelpers.length )
				throw new Error( l( "ERROR_PROPERTY_FOLDER_NOT_FOUND" ) + "\n\n" + errHelpers.join( "\n" ) );
			else
				throw e;
		}

		// Massage the absolute path to be a relative one, and return it.
		path = viewUtils.getRelativeMFilesPath( this.vault, path );
		return path;
	}


	/**
	* Creates a TypedValues collection to be used as a PropertyFolders for a
	* target object in a sublisting view.
	*
	* @param {MFiles.ObjVer} objVer - The object the property folder should point to.
	* @param {MFiles.ExpressionExs} levels - The grouping levels defined by the view.
	* @returns {MFiles.TypedValues} - The PropertyFolders.
	*/
	SubListingManager.prototype.getObjectPropertyFolders = function( objVer, levels ) {

		var propFolders = new MFiles.TypedValues();

		if( levels.Count === 1 ) {

			// If there is only one level this is the legacy style,
			// where the grouping level points directly to the object.
			var folder = new MFiles.TypedValue();
			folder.SetValue( MFDatatypeLookup, objVer.ID );
			propFolders.add( -1, folder );

		} else if( levels.Count >= 3 &&
			levels.Item( 1 ).Expression.Type === MFExpressionTypeObjectIDSegment &&
			levels.Item( 2 ).Expression.Type === MFExpressionTypeStatusValue &&
			levels.Item( 2 ).Expression.DataStatusValueType === MFStatusTypeObjectID ) {

			// There are at least 3 levels.
			// The first points to the target object's id segment.
			// The second points to the target object's id.

			// The third level should point to the target object itself.
			// Verify the third level points directly to our target object type.
			var expr3 = levels.Item( 3 ).Expression;
			if( expr3.Type === MFExpressionTypePropertyValue ) {

				// This is a propertyDef folder.

				// Ensure the property def points to the target object type.
				var pd = this.vault.PropertyDefOperations.GetPropertyDef( expr3.DataPropertyValuePropertyDef );
				if( !pd.BasedOnValueList || pd.ValueList !== objVer.Type )
					throw Error( l( "ERR_GROUPING_LEVEL_MISMATCH" ) );

			} else if( expr3.Type !== MFExpressionTypeTypedValue ||
				expr3.DataTypedValueDatatype !== MFDatatypeLookup ||
				expr3.DataTypedValueValueList !== objVer.Type ) {

				// This is a Object Type based folder that doesn't point to the target object type.
				throw Error( l( "ERR_GROUPING_LEVEL_MISMATCH" ) );
			}

			// Create ID segment property folder value.
			var segmentFolder = new MFiles.TypedValue();
			var segment = Math.floor( objVer.ID / levels.Item( 1 ).Expression.DataObjectIDSegmentSegmentSize );
			segmentFolder.SetValue( MFDatatypeInteger, segment );
			propFolders.add( -1, segmentFolder );

			// Create ID property folder value.
			var idFolder = new MFiles.TypedValue();
			idFolder.SetValue( MFDatatypeInteger, objVer.ID );
			propFolders.add( -1, idFolder );

			// Create object lookup property folder value.
			var objFolder = new MFiles.TypedValue();
			objFolder.SetValue( MFDatatypeLookup, objVer.ID );
			propFolders.add( -1, objFolder );

		} else {

			// Grouping level configuration not recognized.
			MFiles.ThrowError( l( "ERR_INVALID_GROUPING_LEVELS" ) );

		}

		return propFolders;

	}


	/*
	* Creates a top level view to contain all sublisting views of a certain type (commmon or user specific).
	* The views are created at the root level and hidden by default.
	*
	* @param {boolean} common - Indicates if the view to create is a common view. Default: false (user specific).
	* @returns {MFiles.View} - The created view.
	*/
	SubListingManager.prototype.getParentSubListingView = function( common ) {

		var key = "SubListingParentView";
		var parentView = this.appSettings.value( key, common );
		var viewDef;

		// If a parent view is defined, lookup the definition.
		if( parentView ) {

			// Try to get the view by id.
			if( parentView.id ) {
				try {
					viewDef = this.vault.ViewOperations.GetView( parentView.id );
				} catch( e ) { }
			}

			// If we still don't have a view, try to get it by guid.
			if( !viewDef && parentView.guid ) {
				try {
					parentView.id = this.vault.ViewOperations.GetViewByGUID( parentView.guid );
					viewDef = this.vault.ViewOperations.GetView( parentView.id );
				} catch( e ) { }
			}
		}

		// If we did not find a view definition, 
		// check if a parent view exists that simply isn't explicitly set (due to import).
		if( !viewDef ) {

			// Resolve the name of the parent view.
			var viewName = l(( common ) ? "VIEW_NAME_COMMON_SUBLISTING_VIEW" : "VIEW_NAME_PERSONAL_SUBLISTING_VIEW" );

			// Loop over all views.
			for( var id in this.allViewsByID ) {
				var vd = this.allViewsByID[id];

				// If this is a top level view with the same name, use it.
				if( vd.HasParent === false && vd.Name === viewName ) {
					viewDef = vd;
					this.appSettings.save( key, { id: viewDef.ID, guid: viewDef.GUID }, common );
				}
			}
		}


		// If we still did not find a view definition, we must create one.
		if( !viewDef ) {

			// Define the parent view.
			viewDef = new MFiles.View();
			viewDef.Name = viewName;
			viewDef.Visible = false;
			viewDef.Common = common;
			viewDef.ViewType = MFViewTypeFilterOnly;

			// Create the parent view.
			viewDef = this.vault.ViewOperations.AddView( viewDef );

			// Add the view to our internal cache.
			this.allViewsByID[viewDef.ID] = viewDef;

			// Save the reference to this view for future use.
			this.appSettings.save( key, { id: viewDef.ID, guid: viewDef.GUID }, common );
		}


		return viewDef;
	}


	/*
	* Shows a dialog to configure and save a new sublisting.
	*
	* @param {object} [properties] - Optional. An object defining base subListing properties to load in the editor.
	* @returns {SubListingDefinition} The newly created sublisting or null if canceled.
	*/
	SubListingManager.prototype.showCreateSubListingDialog = function( properties ) {

		// Get a default subListing defintion for editing.
		var subListing = this.getSubListingForEditing();

		// Copy the passed properties to the default sublisting.
		if( properties ) {

			// Loop over each passed property only copying
			// if the property exists in the sublisting.
			for( var name in properties ) {
				if( subListing.hasOwnProperty( name ) ) {

					if( name === "classes" && properties[name] ) {

						subListing[name] = castToLocalArray( properties[name] );

					} else {

						// Copy the property to the new sublisting.
						subListing[name] = properties[name];
					}
				}

			}

			// Make sure no id is specified.
			subListing.id = undefined;
		}

		// Define data to pass to dashboard.
		var data = {
			subListing: subListing,
			retVal: -1
		};

		// Launch modal edit dashboard.
		this.shellUI.ShowPopupDashboard( "edit_sublisting", true, data );

		// The dialog was canceled. Return null.
		if( data.retVal !== 0 )
			return null;

		// Make sure the classes array is a localArray.
		data.subListing.classes = castToLocalArray( data.subListing.classes );

		// Create or Update the sublisting returned from the dialog as appropriate.
		return this.createSubListing( data.subListing );
	}


	/*
	* Shows a dialog to configure and save a new or existing sublisting.
	*
	* @param {string|} id - The id of the sublisting to show in the dialog.
	* @returns {SubListingDefinition} The modified sublisting or null if canceled.
	*/
	SubListingManager.prototype.showEditSubListingDialog = function( id ) {

		// Define data to pass to dashboard.
		var data = {
			subListing: this.getSubListingForEditing( id ),
			retVal: -1
		};

		// Launch modal edit dashboard.
		this.shellUI.ShowPopupDashboard( "edit_sublisting", true, data );

		// The dialog was canceled. Return null.
		if( data.retVal !== 0 )
			return null;

		// Make sure the classes array is a localArray.
		data.subListing.classes = castToLocalArray( data.subListing.classes );

		// Create or Update the sublisting returned from the dialog as appropriate.
		return this.updateSubListing( data.subListing );



	}



	/*
	* Determines if a sublisting can be shown within a specific folder path.
	*
	* @param {ExtendedFolderPathSettings} folderPathsettings - The settings of the folder path to update.
	* @param {SubListingDefinition} subListing - The sublisting to determine the visibility of.
	* @returns {boolean} True if the subListing is visible, false otherwise.
	*/
	SubListingManager.prototype.isVisibleInFolderPath = function( folderPathsettings, subListing ) {

		// Branch logic by whether the subListing is shown by default.
		if( subListing.isDefault ) {

			// SubListing is shown by default, return true if it is not blacklisted.
			return inArray( subListing.id, folderPathsettings.subListingBlackList ) === -1;
		} else {

			// SubListing is not shown by default, return true if it is in the white list.
			return inArray( subListing.id, folderPathsettings.subListingWhiteList ) !== -1;
		}
	}



	/*
	* Explicity sets a sublistings visibility for a folder path.
	*
	* @param {ExtendedFolderPathSettings} folderPathsettings - The settings of the folder path to update.
	* @param {SubListingDefinition} subListing - The sublisting to set the visibility for.
	* @param {boolean} visible - Indicates the visibility for the sublisting to set.
	* @returns {boolean} True if the folderPathsettings object was modified.
	*/
	SubListingManager.prototype.setSubListingVisibilityInFolderPath = function( folderPathsettings, subListing, visible ) {


		// Branch logic by desired visibility, and whether the listing is shown by default.
		if( visible ) {
			if( subListing.isDefault ) {

				// SubListing is shown by default, and should be visible.
				// Remove it from the  black list (if it is there).
				return removeFromList( subListing.id, folderPathsettings.subListingBlackList );

			} else {

				// SubListing is not shown by default, but should be visible.
				// Add it to the white list (if it is there).
				return addToList( subListing.id, folderPathsettings.subListingWhiteList );
			}
		} else {
			if( subListing.isDefault ) {

				// SubListing is shown by default, but should not be visible.
				// Add it to the black list.
				return addToList( subListing.id, folderPathsettings.subListingBlackList );

			} else {

				// SubListing is not shown by default, and should not be visible.
				// Remove it from the  black list (if it is there).
				return removeFromList( subListing.id, folderPathsettings.subListingWhiteList );
			}
		}
	}


	/*
	* Resets sublistings visibility for a folder path to it's default.
	* Also clears out deleted or invsible sublistings.
	*
	* @param {ExtendedFolderPathSettings} folderPathsettings - The settings of the folder path to update.
	* @param {SubListingDefinition[]} A list of sublistings whose visibility will have changed.
	*/
	SubListingManager.prototype.resetSubListingVisibilityInFolderPath = function( folderPathsettings ) {

		// The changed sublistings will be any in the white list or blacklist.
		var changedIDs = folderPathsettings.subListingBlackList.concat( folderPathsettings.subListingWhiteList );

		// Reset the whitelist and the blacklist.
		folderPathsettings.subListingWhiteList = [];
		folderPathsettings.subListingBlackList = [];


		// Initialize return value.
		var changedSubListings = [];

		// Loop over the id's of the sublistings whose values have changed.
		for( var i = 0; i < changedIDs.length; i++ ) {

			// Loop over the sublisting contexts.
			for( var c in this.contexts ) {

				// Try to resolve a sublisting from the current context with the changed sublisting id.
				var subListing = this.contexts[c].settings[changedIDs[i]];

				// If a sublisting was resolved, add it to the list,
				// and move on to the next id.
				if( subListing ) {
					changedSubListings.push( subListing );
					break;
				}
			}
		}

		// Return the sublisitings whose visibility will have changed.
		return changedSubListings;
	}


	/*
	* Determines if a sublistings is editable by the current vault session's user.
	*
	* @param {SubListingDefinition} subListing - The sublisting to determine the visibility of.
	* @returns {boolean} True if the subListing is editable, false otherwise.
	*/
	SubListingManager.prototype.isSubListingEditableByCurrentUser = function( subListing ) {

		// Assume the sublisting is editable if it is not common, or the user is a common view manager.
		return ( !subListing.isCommon || this.viewManager );
	}


	/*
	* Determines if a object version is applicable to a given sublisting.
	*
	* @param {MFiles.ObjectVersion} objectVersion - The object version.
	* @param {SubListingDefinition} subListing - The sublisting..
	* @returns {boolean} True if the object version is a target, false otherwise.
	*/
	SubListingManager.prototype.isObjectVersionTargetForSublisting = function( objectVersion, subListing ) {

		// Object must match exactly.
		var objMatch = ( subListing.objType == objectVersion.ObjVer.Type );


		// If classes aren't defined it matches, or if the object's class is in the defined list it matches.
		var classMatch = ( !subListing.classes ||
							!subListing.classes.length ||
							inArray( objectVersion.Class, subListing.classes ) !== -1 );

		// Object is a target if both object and class match.
		return ( objMatch && classMatch );

	}



	/*
	* Determines if a sublisting should be shown for a particular object in a particular folder.
	*
	* @param {SubListingDefinition} subListing - The sublisting.
	* @param {ExtendedFolderPathSettings} folderPathsettings - The folder path.
	* @param {MFiles.ObjectVersion} objectVersion - The object version.
	* @returns {boolean} True if the sublisting should be shown, false otherwise.
	*/
	SubListingManager.prototype.isVisibleForFolderPathAndObjectVersion = function( subListing, folderPathsettings, objectVersion ) {

		return ( this.isVisibleInFolderPath( folderPathsettings, subListing ) &&
				 this.isObjectVersionTargetForSublisting( objectVersion, subListing ) );

	}


	// Returns true if current user has full access to the vault.
	SubListingManager.prototype.isAdmin = function() {

		var adminRights = [
			//MFVaultAccessNone
			MFiles.MFVaultAccess.CreateDocs,
			MFiles.MFVaultAccess.SeeAllDocs,
			MFiles.MFVaultAccess.UndeleteDocs,
			MFiles.MFVaultAccess.DestroyDocs,
			MFiles.MFVaultAccess.ForceUndoCheckout,
			MFiles.MFVaultAccess.ChangeDocSecurity,
			MFiles.MFVaultAccess.ChangeMetaDataStructure,
			MFiles.MFVaultAccess.ManageUserAccounts,
			MFiles.MFVaultAccess.ChangeFullControlRole,
			MFiles.MFVaultAccess.VerifyVault,
			MFiles.MFVaultAccess.EditAllDocs,
			MFiles.MFVaultAccess.ExportContent,
			MFiles.MFVaultAccess.ImportContent,
			MFiles.MFVaultAccess.ManageVLItemsFromClient,
			MFiles.MFVaultAccess.ManageExternalLocations,
			MFiles.MFVaultAccess.LicenseAllowsModifications,
			MFiles.MFVaultAccess.ManageEvents,
			MFiles.MFVaultAccess.ManageTraditionalFolders,
			MFiles.MFVaultAccess.ChangeObjectStates,
			MFiles.MFVaultAccess.ChangeAssignments,
			MFiles.MFVaultAccess.ManageCommonViews,
			MFiles.MFVaultAccess.ManageCommonUISettings,
			MFiles.MFVaultAccess.ManageCommonNotificationRules,
			MFiles.MFVaultAccess.SeeDeletedDocs,
			MFiles.MFVaultAccess.MateriakizeViews
		];

		// Check vault access for each item.
		for( var i = 0; i < adminRights.length; i++ ) {
			if( !this.vault.SessionInfo.CheckVaultAccess( adminRights[i] ) )
				return false;
		}
		return true;

	}


	SubListingManager.prototype.promptExportCommonSubListings = function() {

		// Ensure user is an admin.
		if( !this.admin )
			MFiles.ThrowError( l( "ERR_PERMISSION_DENIED" ) );

		// Create an array of cloned common sublistings replacing element id's with guids.
		var subListings = [];
		for( var id in this.common.settings ) {

			// Get a handle on a clone of this subListing definition.
			var subListing = extend( {}, this.common.settings[id] );

			// Attempt to update view reference.
			try {

				// Resolve the sublisting's view, and update the id reference with it's guid.
				var view = this.vault.ViewOperations.GetView( subListing.view )
				subListing.view = view.GUID;

			} catch( e ) {

				// Sublisting view not found.
				MFiles.ThrowError( l( "ERR_SUBLISTING_VIEW_NOT_FOUND", subListing.name ) );
			}


			// Attempt to update the objType reference.
			try {

				// Replace objType id with its guid.
				subListing.objType = this.vault.ObjectTypeOperations.GetObjectType( subListing.objType ).GUID;

			} catch( e ) {

				// Sublisting target object type not found.
				MFiles.ThrowError( l( "ERR_TARGET_OBJECT_NOT_FOUND", subListing.name ) );
			}

			// Ensure we have a useable array of classes.
			subListing.classes = castToLocalArray( subListing.classes );

			// Loop over each target class.
			var classGuids = [];
			for( var i = 0; i < subListing.classes.length; i++ ) {

				// Simply ignore deleted classes
				try {

					// Lookup the class definition and add the class' guid to our list.
					var c = this.vault.ValueListItemOperations.GetValueListItemByID(
									MFBuiltInValueListClasses, subListing.classes[i] );
					classGuids.push( c.ItemGUID );

				} catch( e ) {

					// Sublisting target class was not found.
					MFiles.ThrowError( l( "ERR_TARGET_CLASS_NOT_FOUND", subListing.name ) );
				}
			}

			// Replace the sublistings class id array with our class guid array.
			subListing.classes = classGuids;

			// Add the subListing definition to our list to export.
			subListings.push( subListing );

		}



		// Ensure there are common sublistings to export.
		if( !subListings.length )
			MFiles.ThrowError( l( "ERR_IMPORT_NO_SUBLISTINGS" ) );


		// Show a Save file dialog.
		var launcher = MFiles.CreateObjectCLR( "redist\\CommonDialogs\\CommonDialogs.dll", "CommonDialogs.CommonDialogLauncher" );
		var dialog = launcher.GetSaveFileDialog()
		dialog.Filter = l( "FILE_FILTER_JSON" );
		dialog.DefaultExt = "json";
		dialog.FilterIndex = 1;
		dialog.RestoreDirectory = true;
		var retVal = dialog.ShowDialog();

		// Return null the dialog did not have a successful return value.
		if( retVal != 1 ) return null;


		// Serialize the sublistings.
		var content = JSON.stringify( subListings, null, 4 ).replace( /\n/g, "\r\n" );


		// Write the content to file.
		var fso = new ActiveXObject( 'Scripting.FileSystemObject' );
		var stream = fso.OpenTextFile( dialog.FileName, 2, true );  // 2 = ForWriting, Create = true
		stream.Write( content );
		stream.Close();

	}


	SubListingManager.prototype.promptImportCommonSubListings = function() {

		// Ensure user is an admin.
		if( !this.admin )
			MFiles.ThrowError( l( "ERR_PERMISSION_DENIED" ) );

		// Initialize an open file dialog.
		var launcher = MFiles.CreateObjectCLR( "redist\\CommonDialogs\\CommonDialogs.dll", "CommonDialogs.CommonDialogLauncher" );
		var dialog = launcher.GetOpenFileDialog()
		dialog.Filter = l( "FILE_FILTER_JSON" );
		dialog.DefaultExt = "json";
		dialog.FilterIndex = 1;
		dialog.RestoreDirectory = true;
		var retVal = dialog.ShowDialog();

		// Return null the dialog did not have a successful return value.
		if( retVal != 1 ) return null;


		// Read in the selected file contents.
		var fso = new ActiveXObject( 'Scripting.FileSystemObject' );
		var stream = fso.OpenTextFile( dialog.FileName, 1 );  // 1 = ForReading
		var content = stream.ReadAll();
		stream.Close();

		// Release references immediately in hopes of early gc.
		stream = null;
		fso = null;
		dialog = null;
		launcher = null;

		// Parse the contents.
		var subListings = JSON.parse( content );

		// Sanity check. Ensure the parsed contents is an array.
		if( Object.prototype.toString.call( subListings ) !== '[object Array]' )
			MFiles.ThrowError( l( "ERR_IMPORT_INVALID_FILE" ) );


		// Loop over all sublistings, determining whether we can import each.
		var toImport = [];
		var toUpdate = [];
		var toDiscard = [];
		for( var i = 0; i < subListings.length; i++ ) {

			var subListing = subListings[i];

			// Sanity check. Ensure sublisting is a sublisting with key values.
			if( !subListing ||
				Object.prototype.toString.call( subListing ) !== '[object Object]' ||
				!subListing.view ||
				!subListing.objType ) {

				// Unrecognized data.
				MFiles.ThrowError( l( "ERR_IMPORT_INVALID_FILE" ) );
			}

			// Try to convet all guid references to id references.
			try {

				// Lookup view id.
				subListing.view = this.vault.ViewOperations.GetViewIDByGUID( subListing.view );

				// Lookup objType id.
				subListing.objType = this.vault.ObjectTypeOperations.GetObjectTypeIDByGUID( subListing.objType );

				// Lookup class ids.
				for( var j = 0; j < subListing.classes.length; j++ ) {
					subListing.classes[j] = this.vault.ClassOperations.GetObjectClassIDByGUID( subListing.classes[j] );
				}

			} catch( e ) {

				// If anything fails put this sublisting on the discard pile.
				toDiscard.push( subListing );
				continue;
			}

			// If we made it here we can import the current subListing.

			// Check if we should update an existing subListing or create a new one.
			if( this.common.settings[subListing.id] )
				toUpdate.push( subListing );
			else
				toImport.push( subListing );
		}

		// Prompt to confirm import.
		var res = this.shellUI.ShowMessage( {
			caption: l( "WINCAPTION_IMPORT_SUBLISTINGS" ),
			message: l( "CONFIRM_IMPORT_SUBLISTINGS", toImport.length, toUpdate.length, toDiscard.length ),
			icon: "question",
			button1_title: l( "BUTTON_IMPORT_SUBLISTINGS" ),
			button1_enabled: toImport.length > 0 || toUpdate.length > 0,
			button2_title: l( "BUTTON_CANCEL" ),
			defaultButton: 2,
			cancelButton: 2
		} );

		// Quit now if they cancelled.
		if( res !== 1 ) return;


		// Import new sublistings.
		for( i = 0; i < toImport.length; i++ ) {
			this.createSubListing( toImport[i] );
		}

		// Update existing sublistings.
		for( i = 0; i < toUpdate.length; i++ ) {
			this.updateSubListing( toUpdate[i] );
		}
	}


} )();