"use strict";

// Global object to contain the utility methods.
var viewUtils = {};

( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;


	/**
	* Finds the default path for an object version.
	*
	* The default path is the one that will be used when resolving an object's url,
	* or using the "Browse in this window" option. Paths are always in the built in view "ID2",
	* i.e. MFBuiltInViewByID, ID:12.
	*
	* @param {MFiles.Vault} vault - The vault in which the object resided.
	* @param {MFiles.ObjVer} objVer - The object and version to which the path should point.
	* @param {MFiles.MFLatestSpecificBehavior } latestBehavior - The latest/specific version behavior of the view.
	* @returns {string} - The default path to the passed objVer.
	* @throws An exception is thrown if a path could not be found for the object version.
	*/
	viewUtils.getDefaultPathForObject = function( vault, objVer, latestBehavior ) {

		// Set the default, but don't overwrite a valid value of zero ("normal" behavior).
		if( !latestBehavior && latestBehavior !== 0 )
			latestBehavior = MFLatestSpecificBehaviorAutomatic;

		// MFLatestSpecificBehaviorAutomatic is used so that -1 can be used as the version
		var paths = vault.ObjectOperations.GetObjectLocationsInView( MFBuiltInViewByID, latestBehavior, objVer );

		// Ensure a path was found.
		if( !paths.Count )
			MFiles.ThrowError( l( "ERR_NO_OBJECT_PATH" ) );

		// Return the first path found.
		return paths.Item( 1 );
	}


	/**
	* Removes the M-Files root from an absolute M-Files path.
	*
	* This is useful for getting view path locations to pass to 
	* {MFiles.ShellFrame}.CreateAdditionalListingForPath() as it requires the
	* relative path and MFilesAPI only provides methods to get the absolute path.
	*
	* @param {MFilesAPI.Vault} vault - The vault into which the path points.
	* @param {string} path - The absolute path inside the vault.
	* @returns {string} The relative path from the vault root (drive and vault name removed).
	*/
	viewUtils.getRelativeMFilesPath = function( vault, path ) {

		// Resolve the name of the vault.
		var vaultName = vault.Name;

		// Find the first occurence of the vault in the path.
		var index = path.indexOf( vaultName );

		// Remove everything up to the end of the vault name in the path, 
		// and an additional character to remove the path separator.
		var relPath = path.substr( index + vaultName.length + 1 );

		// Return the relatie path.
		return relPath;
	}

	/**
	* Creates a TypedValues collection to be used as a PropertyFolders
	* argument. A single TypedValue (PropertyFolder) is defined with a lookup
	* pointing to the passed object.
	*
	* @param {MFiles.Vault} vault - The vault in which the folder path resides.
	* @param {string} folderPath - The folder path info is being resolved for.
	* @param {MFiles.FolderDefs} folderDefs - The folderDefs that describe the path.
	* @returns {FolderPathInfo} - The path information.
	*/
	viewUtils.getFolderPathInfo = function( vault, folderPath, folderDefs ) {


		/*
		 * @typedef FolderPathInfo
		 * @type Object
		 * @property {string} folderPath - The folder path which the information describes.
		 * @property {MFiles.FolderDefs} folderDefs - The folderDefs which the information describes.
		 * @property {MFiles.View} view - The definition of the lowest level view in the path.
		 * @property {int} depth - The number of grouping levels below the lowset level view the path is.
		 * @property {bool} base - Indicates whether the path is at the lowest level defined for the view.
		 *			Always true if the view defines no grouping levels.
		 * @property {ViewGroupingLevelInfo} propertyFolders
		 * @property {bool} groupedByRealObject
		 */

		// Return the {FolderPathInfo}.
		var info = {
			folderPath: folderPath,
			folderDefs: folderDefs,
			view: null,
			search: null,
			depth: 0,
			base: true,
			folderInfo: [],
			realObjLevels: 0,
			groupLevels: []
		};


		// Resolve the lowest level view and the number of folders open underneath it.
		// Loop backwards through the folderDefs until we find a view and count the 
		// property folders along the way.
		for( var i = 1; i <= folderDefs.length; i++ ) {

			var folderDef = folderDefs.Item( i );

			// Branch by folderdef type.
			if( folderDef.FolderDefType == MFFolderDefTypeViewFolder ) {

				// View.

				// Reset the level
				info.depth = 0;

				// Lookup the view.
				if( folderDef.View ) {
					info.view = vault.ViewOperations.GetView( folderDef.View );
					info.folderInfo.push( info.view );
				}

			} else if( folderDef.FolderDefType == MFFolderDefTypeSearchFolder ) {

				// Search View.

				// Reset the level
				info.depth = 0;
				info.search = folderDef.SearchDef;

			} else if( folderDef.FolderDefType == MFFolderDefTypePropertyFolder ) {

				// Property Folder.

				// Keep track of the current level
				info.depth++;				

				// Resolve information about the level and add it to the folderInfo array.
				var levelContainer = ( info.search ) ? info.search : info.view;
				var propFolderInfo = viewUtils.resolveViewGroupingLevelInfo(
							vault, levelContainer.Levels.Item( info.depth ), folderDef );

				// Add the level info to the info object.
				info.folderInfo.push( propFolderInfo );
				info.groupLevels.push( propFolderInfo );

				// Determine if this level is based on a real object.
				if( propFolderInfo.valueList !== undefined && propFolderInfo.objType.RealObjectType )
					info.realObjLevels++;
			}
		}

		// Determine whether the lowest level we have is the lowset level in the lowest view.
		info.base = ( info.view.Levels.Count == info.depth );

		return info;

	}



	/**
	* Resolves the objectType that a view's grouping level is based on.
	*
	* @param {MFiles.Vault} vault - The vault where the view exists.
	* @param {MFiles.ExpressionEx} level - The view's level definition.
	* @param {MFiles.FolderDef} - The folder def that references the grouping level.
	* @returns {ViewGroupingLevelInfo} The information resolved about the grouping level. 
	*/
	viewUtils.resolveViewGroupingLevelInfo = function( vault, level, folderDef ) {

		/** 
		* @typedef ViewGroupingLevelInfo
		* @type {Object}
		*
		* @property {MFiles.FolderDef} - The folder def that references the grouping level.
		* @property {MFiles.ExpressionEx} level - A reference to the level on which the object's information is based.
		* @property {bool} basedOnValueList - True if the grouping level's value is based on a value list/object type.
		* @property {int} valueList - The value list / object type id (if the level is basedOnValueList).
		* @property {MFiles.PropertyDef} propertyDef - The propertyDef on which the level is based (if applicable)
		* @property {MFiles.ObjType} objType - The value list / object type definition (if the level is basedOnValueList).
		* @property {MFiles.MFDataType} dataType - The data type of the grouping level.
		* @property {string} displayValue - The display value for the level (i.e. the property def or value list name).
		* @property {bool} usesReverseReferences - Indicates whether the grouping level uses reverse reference to match objects.
		*     i.e. Objects appearing in the folder are referenced in the grouping item's own metadata. As opposed to referencing
		*     the grouping item in their own metadata.
		*/

		// Initialize {ViewGroupingLevelInfo} return value.
		var info = {
			folderDef: folderDef,
			level: level,
			basedOnValueList: false,
			valueList: undefined,
			propertyDef: undefined,
			objType: undefined,
			dataType: undefined,
			displayValue: "",
			usesReverseReferences: false
		}

		var levelExpr = level.Expression;

		// Handle based on view or expression type.
		if( Object.hasOwnProperty.call( level, "ReferFromViewLevel" ) && level.ReferFromViewLevel ) {

			// Reverse Reference.
			info.usesReverseReferences = true;
			info.dataType = MFDatatypeMultiSelectLookup;
			info.basedOnValueList = true;
			info.valueList = level.ObjectTypeToGroupBy;
			info.objType = vault.ValueListOperations.GetValueList( info.valueList );
			info.displayValue = info.objType.NameSingular;

		} else if( levelExpr.Type == MFExpressionTypePropertyValue ) {

			// Level is based on a property value expression.

			// Lookup the property definition and set the name and data type.
			var pd = info.propertyDef = vault.PropertyDefOperations.GetPropertyDef( levelExpr.DataPropertyValuePropertyDef );
			info.dataType = pd.DataType;
			info.displayValue = pd.Name;

			// If the property defintion is based on a value list, set the relevant info.
			if( pd.BasedOnValueList ) {
				info.basedOnValueList = true;
				info.valueList = pd.ValueList;
				info.objType = vault.ValueListOperations.GetValueList( pd.ValueList );
			}

		} else if( levelExpr.Type == MFExpressionTypeTypedValue ) {

			// Level is based on a TypedValue expression.

			// Set the data type.
			var dataType = info.dataType = levelExpr.DataTypedValueDatatype;

			// If data type is based on a lookup property, set the additional value list information.
			if( dataType == MFDatatypeLookup || dataType == MFDatatypeMultiSelectLookup ) {
				info.basedOnValueList = true;
				info.valueList = levelExpr.DataTypedValueValueList
				info.objType = vault.ValueListOperations.GetValueList( info.valueList );
				info.displayValue = info.objType.NameSingular;
			}
		}

		// If we couldn't resolve an object type return undefined.
		return info;

	}


	/**
	* Gets the best default option for grouping objects of a certain type in a view.
	* @param {MFiles.Vault} vault - The vault in which the objType resides.
	* @param {MFiles.ObjType} objType - The objType to group.
	* @returns {MFiles.PropertyDefOrObjectType|null} The definition to group by, if any.
	*/
	viewUtils.getBestObjTypeGroupingDef = function( vault, objType ) {

		var groupingDef = new MFiles.PropertyDefOrObjectType();

		// Find the best wasy to group the objType.
		if( objType.AllowedAsGroupingLevel ) {

			// ObjType is groupable.
			groupingDef.ID = objType.ID;
			groupingDef.PropertyDef = false;

		} else if( vault.PropertyDefOperations.GetPropertyDef( objType.DefaultPropertyDef ).AllowedAsGroupingLevel ) {

			// Default property def is groupable.
			groupingDef.ID = objType.DefaultPropertyDef;
			groupingDef.PropertyDef = true;

		} else {

			// Loop over all properties, search for those that reference this object type and are groupable.
			var propDefs = vault.PropertyDefOperations.GetPropertyDefs();
			for( var i = 1; i <= propDefs.Count; i++ ) {

				// Check if this is a propDef of interest.
				var propDef = propDefs.Item( i );
				if( propDef.BasedOnValueList && propDef.ValueList === objType.ID && propDef.AllowedAsGroupingLevel ) {

					// A custom property def is groupable, use it.
					groupingDef.ID = propDef.ID;
					groupingDef.PropertyDef = true;
					break;
				}
			}

			// If no property def was found, give up.
			if( !groupingDef.PropertyDef )
				return null;
		}

		// Return the grouping def.
		return groupingDef;

	}

	/**
	* Determines if an objType can be grouped in a view, either by the objType itself or one of it's propertyDefs.
	* @param {MFiles.Vault} vault - The vault in which the objType resides.
	* @param {MFiles.ObjType} objType - The objType to group.
	* @returns {boolean}
	*/
	viewUtils.objTypeCanBeGrouped = function( vault, objType ) {

		// Check if the object type itself is groupable.
		if( objType.AllowedAsGroupingLevel )
			return true;

		// Check if the default property def is groupable.
		if( vault.PropertyDefOperations.GetPropertyDef( objType.DefaultPropertyDef ).AllowedAsGroupingLevel )
			return true;

		// Loop over all property defs.
		var propDefs = vault.PropertyDefOperations.GetPropertyDefs();
		for( var i = 1; i <= propDefs.Count; i++ ) {

			// Check if this is a relevant propdef and groupable.
			var propDef = propDefs.Item( i );
			if( propDef.BasedOnValueList && propDef.ValueList === objType.ID && propDef.AllowedAsGroupingLevel )
				return true;
		}

		return false;
	}

	/**
	* Determines if the expression is based on an integer segment, and if so,
	* resolves it's size.  There are 4 different ways in which segments can be
	* specified in expressions, and when data functions are used, the segment sizes
	* are not exposed via the API, so we need to hack it.
	*
	* @param {MFiles.Vault} vault - The vault for which the expression is defined.
	* @param {MFiles.Expression} expr - The expression.
	* @returns {number} The segment size, or 0 (zero) if expression is not based on segment sizes.
	*/
	viewUtils.getExpressionSegmentSize = function( vault, expr ) {
		
		if( expr.Type === MFExpressionTypeObjectIDSegment ) {

			// Expression is explicitly an ID segment level, return the defined size.
			return expr.DataObjectIDSegmentSegmentSize;

		} else if( expr.Type === MFExpressionTypeStatusValue &&
			expr.DataStatusValueType === MFStatusTypeOriginalObjectIDSegment ) {

			// Value is the default ID segment used in default ID views.
			// Segment size is always 1000.
			return 1000;

		} else if( expr.Type === MFExpressionTypePropertyValue &&
			expr.DataPropertyValueDataFunction === MFDataFunctionIntegerSegment ) {

			// Property value defines integer segment data function.
			return this.resolveExpressionDataFunctionSegmentSize( vault, expr );

		} else if( expr.Type === MFExpressionTypeTypedValue &&
			expr.DataTypedValueDataFunction === MFDataFunctionIntegerSegment ) {

			// Typed value defines integer segment data function.
			return this.resolveExpressionDataFunctionSegmentSize( vault, expr );
		}

		return 0;
	}

	/**
	* Extracts the segment size from a data function defined for an expression.
	* HACK: This is a hack that extracts the value from the expression's formated display value.
	* The M-Files API currently has no way to report any values defined for a DataFunction.
	*
	* @param {MFiles.Vault} vault - The vault for which the expression is defined.
	* @param {MFiles.Expression} expr - The expression.
	* @returns {number} The segment size, or 0 (zero) if expression is not based on segment sizes.
	*/
	viewUtils.resolveExpressionDataFunctionSegmentSize = function( vault, expr ) {
		
		// Create a clone of the expression with a known segment size.
		var clone = expr.Clone();
		var knownSegmentSize = 555;
		var segmentFunc = new MFiles.DataFunctionCall();
		segmentFunc.SetDataIntegerSegment( knownSegmentSize );

		// Setup the clone's datafunction based on the expression type.
		if( expr.Type === MFExpressionTypePropertyValue ) {
			
			// Property Value expression.

			// Sanity check.
			if( expr.DataPropertyValueDataFunction !==  MFDataFunctionIntegerSegment )
				throw Error();

			// Setup clone.
			clone.SetPropertyValueExpression(
					expr.DataPropertyValuePropertyDef,
					expr.DataPropertyValueParentChildBehavior,
					segmentFunc );

		} else if (expr.Type === MFExpressionTypeTypedValue ){
		
			// Typed Value expression.

			// Sanity check.
			if( expr.DataTypedValueDataFunction !== MFDataFunctionIntegerSegment )
				throw Error();

			// Setup clone.
			clone.SetTypedValueExpression(
					expr.DataTypedValueDatatype,
					expr.DataTypedValueValueList,
					expr.DataTypedValueParentChildBehavior,
					segmentFunc );

		} else if( expr.Type === MFExpressionTypeStatusValue ) {

			// Status Value expression.

			// Sanity check.
			if( expr.MFExpressionTypeStatusValue !== MFDataFunctionIntegerSegment )
				throw Error();

			// Setup clone.
			clone.SetStatusValueExpression(
					expr.DataStatusValueType,
					segmentFunc );

		} else {

			// Incompatible expression type.
			throw Error();
		}

		// Resolve the text value of both the original and cloned expression.
		var propDefs = vault.PropertyDefOperations.GetPropertyDefs();
		var objTypes = vault.ObjectTypeOperations.GetObjectTypes();
		var origText = expr.GetExpressionAsText( propDefs, objTypes );
		var cloneText = clone.GetExpressionAsText( propDefs, objTypes );

		// We assume the formatted text's are identifical except for the segment sizes.
		// So we determine where our known segment size is in the cloned text, and 
		// extract segment size from the origText at the same location. Absolute lengths
		// of the texts are used to determine how many digits are in the orig segment size.
		var segmentStartIndex = cloneText.indexOf( knownSegmentSize );
		var segmentLength = knownSegmentSize.toString().length + ( origText.length - cloneText.length );
		var segmentSize = origText.substr( segmentStartIndex, segmentLength );
		return parseInt( segmentSize );

	}

	/**
	* Calculates the display name of an id segment shown as a range.
	* @param {number} segment - The index of the segment.
	* @param {number} [size] - Optional. The size of the segments. Default = 1000.
	* @returns {string}
	*/
	viewUtils.getIDSegmentDisplayName = function( segment, size ) {

		// Set default size, if it wasn't specified.
		size = size || 1000;

		var min = segment * size;
		var max = min + size - 1;
		return min + "-" + max;
	}

} )();

