"use strict";

// Export.
var ThemeUtil;


( function( 
		jss,
		undefined ) {

	/**
	 * Creates a new {ThemeUtil} object for the current shellUI and dashboard.
	 * @constructor
	 * @param {MFilesUI.ShellUI} shellUI - The shellUI the dashboard is running in.
	 * @param {MFilesUI.Dashboard} dashboard - The dashboard to theme.
	 * @param {ThemeSettings} settings - The theme settings for the dashboard.
	 *
	 * @class
	 * @classdesc - Provides utility functions to easily add custom theme support to UIExt app dashboards.
	 */
	ThemeUtil = function( 
			shellUI, 
			dashboard, 
			settings ) {
		
		// Sanity checks.
		if( !shellUI ) {
			throw new ReferenceError( "shellUI" );
		} else if( !dashboard ) {
			throw new ReferenceError( "dashboard" );
		}

		/**
		 * The shellUI in which this object is running.
		 * @member {MFilesUI.ShellUI}
		 */
		this.shellUI = shellUI;

		/**
		 * The dashboard for which the settings are defined.
		 * @member {MFilesUI.Dashboard}
		 */
		this.dashboard = dashboard;

		/**
		 * The theme settings for this dashboard.
		 * @member {ThemeSettings}
		 */
		this.settings = settings;

		/**
		 * The handle of the registered shellUI listener.
		 * @private {number}
		 */
		this.listenHandle = 0;

		/**
		 * The stylesheet that contains the custom theme's style.
		 * @private {StyleSheet}
		 */
		this.styleSheet = null;

		/**
		 * The canvas element used to modify images.
		 * @private {HTMLCanvasElement}
		 */
		this.canvas = null;

	}

	/**
	 * Message id to broadcast to theme providers.
	 * @const {string}
	 */
	ThemeUtil.MSGID_THEME_LISTENING = "M-Files.Theme.Listening";

	/**
	 * The id of the message that providers send when notifying that the theme has changed.
	 * @const {string}
	 */
	ThemeUtil.MSGID_THEME_CHANGED = "M-Files.Theme.ThemeChanged";

	/**
	 * Message id to notify theme providers with to get the DataURI for a resource reference.
	 * @const {string}
	 */
	ThemeUtil.MSGID_GET_DATA_URI = "M-Files.Theme.GetDataURIForResource";

	/**
	 * Helper method to add theme support to a dashboard.
	 * Initializes a {ThemeUtil}, applies the current custom theme, and
	 * listens for future changes, applying them when they occur.
	 * @static
	 * @returns {ThemeUtil}
	 */
	ThemeUtil.addThemeSupport = function( 
			shellUI, 
			dashboard, 
			settings ) {

		var util = new ThemeUtil( shellUI, dashboard, settings );
		util.startListening();

		return util;
	}

	/**
	 * Indicates if the MFiles client supports themeing.
	 * @returns {boolean}
	 */
	ThemeUtil.prototype.clientSupportsThemes = function() {
		return !!MFiles.Event.CrossApplicationNotification;
	}

	/**
	 * Indicates whether the object is actively listening for theme changes.
	 * @returns {boolean}
	 */
	ThemeUtil.prototype.isListening = function() {
		return !!this.listenHandle;
	}

	/**
	 * Starts listening for theme change notifications, and applies them when received.
	 * If the object is already listening, this method has no effect.
	 * If the client doesn't support themes, this method has no effect.
	 * @returns {boolean} - Returns true if the object is actively listening.
	 */
	ThemeUtil.prototype.startListening = function() {

		// Nothing to bind to if we can't recieve notifications for other applications.
		if( !this.clientSupportsThemes() ) {
			return false;
		}

		// Don't do anything if we're already listening.
		if( this.isListening() ) {
			return true;
		}

		// Get a reference to ourselves for use in closures.
		var self = this;

		// Setup theme a theme change listener.
		// This needs to be done before the broadcast.
		this.listenHandle = this.shellUI.Events.Register(
				MFiles.Event.CrossApplicationNotification,
				function( appGuid, msgId, data ) {
					self._onNotification( appGuid, msgId, data );
				} );

		// We need to unregister the event when the dashboard closes as MFiles doesn't
		// clean-up the listeners on the shellUI when just the shellFrame changes.
		this.dashboard.Events.Register( 
				MFiles.Event.Stop,
				function() {
					self.stopListening();
				} );

		// Broadcast to theme providers to let them know that we are listening.
		// Also inform them of the custom theme this application defines (if specified).
		var customTheme = "";
		if( this.settings && this.settings.customTheme ) {
			customTheme = JSON.stringify( this.settings.customTheme, null, 4 );
		}
		this.shellUI.BroadcastMessage( ThemeUtil.MSGID_THEME_LISTENING, customTheme );

		return true;
	}

	/**
	 * Stops listening for theme change events.
	 * Has no effect if the object isn't currently listening.
	 */
	ThemeUtil.prototype.stopListening = function() {

		// Check if there is a handle to unregister.
		if( this.listenHandle ) {

			// Unregister the listener.
			this.shellUI.Events.Unregister( this.listenHandle );

			// Reset our internal handle.
			this.listenHandle = 0;
		}
	}

	/**
	 * The handler for when notifications are recieved from other applications.
	 * @private
	 * @param {string} appGuid - The guid of the app that sent the notification.
	 * @param {string} msgId - The id of the message that was sent.
	 * @param {*} data - The payload of the notification.
	 */
	ThemeUtil.prototype._onNotification = function( 
			appGuid, 
			msgId, 
			data ) {

		// Check if this is a theme changed notification.
		if( msgId === ThemeUtil.MSGID_THEME_CHANGED ) {

			// Theme changed notification.

			// Parse the theme string into an object, 
			// and handle the theme change.
			var theme = JSON.parse( data );
			this._onThemeChanged( theme );
		}
	}

	/**
	 * The handler for when a theme has changed.
	 * @private
	 * @param {ThemeDefinition} theme - The updated theme definition.
	 */
	ThemeUtil.prototype._onThemeChanged = function( theme ) {

		// Clear any previously set custom stylesheet.
		if( jss.defaultSheet ) {
			jss._removeSheet( jss.defaultSheet );
			jss.defaultSheet = undefined;
		}		
		
		// Nothing to do if there are no mappings defined.
		if( !this.settings || !this.settings.mappings ) {
			return;
		}

		// Resolve all image resources as data URIs.
		// (UI Ext apps seem to have a limitation preventing referencing 
		// images outside of the application directory.)
		if( this.settings.mappings.resources ) {
			this._replaceResourceReferencesWithDataURIs( theme, this.settings.mappings.resources );
		}

		// Map all theme properties to css style rules.
		if( this.settings.mappings.style ) {
			this._createBasicStyleRules( theme, this.settings.mappings.style );
		}

		// Generate new icons based on theme color properties and
		// add them to applicable css style rules.
		if( this.settings.mappings.images ) {
			this._createCustomImagesAndRules( theme, this.settings.mappings.images );
		}
	}

	/**
	 * Replaces all theme properties that are set and defined as resources with a css url value containing a dataURI.
	 * This is a work-around for the fact that UIExt apps cannot reference images defined
	 * outside of the working directory (and also means we don't have to worry about paths).
	 * The theme provider should set a name or path of a resource into the theme, then, for
	 * the specific resources that we want to use, we ask the proivder to give us the dataURIs
	 * of those resources which we can place directly in the css style rules.
	 * @private
	 * @param {ThemeDefinition} theme - The theme to replace resource references in.
	 * @param {ThemeResourceMappings} resources - The resource references to replace.
	 */
	ThemeUtil.prototype._replaceResourceReferencesWithDataURIs = function( 
			theme, 
			resources ) {

		// Resolve the path to the resources.
		// Typically they are specified in the theme relative to this.
		var resourcePath = "";
		if( theme.meta && theme.meta.resourcePath ) {
			resourcePath = theme.meta.resourcePath + "\\";
		}

		// Loop over each resource defined in the mapping.
		// If the theme defines the resource, replace it's reference
		// with a dataURI of the resource.
		for( var key in resources ) {

			// Check if the theme defines this resource.
			var r = resources[ key ];
			var fileName = theme[ r.section ] && theme[ r.section ][ r.key ];
			if( fileName && fileName !== "default" ) {

				// Resource is defined in the theme.

				// Replace the reference in the theme, with a
				// dataURI containing the resource data.
				var filePath = resourcePath + theme[ r.section ][ r.key ];
				var dataURI = this.getImageFileAsDataURI( filePath )
				theme[ r.section ][ r.key ] = "url( '" + dataURI + "' )";
			}
		}
	}

	/**
	 * Maps custom theme properties to css rules in the current dashboard.
	 * @private
	 * @param {ThemeDefinition} theme - The theme to use to set the values of the css rule properties.
	 * @param {ThemeStyleMappings} styleMappings - The style mapping definition.
	 */
	ThemeUtil.prototype._createBasicStyleRules = function( 
			theme, 
			styleMappings ) {

		// Loop over the theme sections defined in the style mappings.
		for( var section in styleMappings ) {

			// Loop over each rule defined in the section.
			for( var i in styleMappings[ section ] ) {

				// Get a handle on the current rule.
				var mapRule = styleMappings[ section ][ i ];
				var cssRule = null;

				// Loop over the properties defined for this rule.
				for( var cssPropName in mapRule.properties ) {

					// Resolve the theme key defined as the rule property's value.
					var key = mapRule.properties[ cssPropName ];

					// Resolve the value from the theme with the theme key.
					var value = theme[ section ] && theme[ section ][ key ];

					// Check if the theme defines a value for this mapping.
					if( value && value != "default" ) {

						// Theme defines a custom value for this property.

						// Add the theme valueto the css rule
						// (creating the cssRule if needed).
						cssRule = cssRule || {};
						cssRule[ cssPropName ] = value;
					}
				}

				// If a css rule was created add it to the stylesheet.
				if( cssRule ) {
					jss.set( mapRule.selector, cssRule );
				}
			}
		}
	}

	/**
	 * Converts the colors of certain images and places the generated images
	 * as the properties of specified css style rule properties.
	 * @private
	 * @param {ThemeDefintiion} theme - The theme that defines colors to update images with.
	 * @param {ThemeImageMappings} imageMappings - Defines which images should be regenerated
	 *    with which theme definied color, and which css rule/property to write the new image to.
	 */
	ThemeUtil.prototype._createCustomImagesAndRules = function( 
			theme,
			imageMappings ) {

		// Get a reference to ourselves for use in closures.
		var self = this;

		// Loop over each image mapping section.
		for( var section in imageMappings ) {

			// Loop over each set of theme color/set of images.
			for( var mapName in imageMappings[ section ] ) {

				// Get a handle on the image set that should be
				// processed the same way 
				// (same size, same color mapped to same theme property ).
				var map = imageMappings[ section ][ mapName ];

				// Resolve the custom color from the theme.
				var color = theme[ section ] && theme[ section ][ map.color ];

				// If no custom color is defined, do not process these images.
				if( !color || color === "default" ) {
					continue;
				}

				// Convert the color string into an rgb object.
				var rgb = this._colorStringToRgbObj( color );

				// Loop over each image defined in the map.
				for( var i = 0; i < map.images.length; i++ ) {

					// Recolor the image and assign the result to a css rule.
					var img = map.images[ i ];
					this._generateImageAndUpdateRule( img.src, rgb, img.selector, img.property );
				}
			}
		}
	}

	/**
	 * Resolves the canvas element to use for image processing (creating one if needed).
	 * @private
	 * @returns {HTMLCanvasElement} - The canvas element.
	 */
	ThemeUtil.prototype._getCanvas = function() {

		// Create the canvas if it doesn't yet exist.
		if( !this.canvas ) {

			// Create the canvas element, positioning it far off screen.
			this.canvas = document.createElement( "canvas" );
			this.canvas.setAttribute( "style", "position: absolute; top:-10000px;" );
			this.canvas.width = 16;
			this.canvas.height = 16;
			document.body.appendChild( this.canvas );
		}

		return this.canvas;
	}

	/**
	 * Converts a 7 character hex color string (e.g. #F5F5F5) to an
	 * object of numeric (integers between 0 and 255) r-g-b color values.
	 * @private
	 * @returns {Object} - RGB color object.
	 */
	ThemeUtil.prototype._colorStringToRgbObj = function( colorStr ) {

		// Parse the string into separate integer values and return.
		return {
			r: parseInt( colorStr.substring( 1,3 ), 16 ),
			g: parseInt( colorStr.substring( 3,5 ), 16 ),
			b: parseInt( colorStr.substring( 5 ), 16 )
		};
	}

	/**
	 * Recolors an image and sets the recolored image as a css rule property value.
	 * Sets all non transparent pixels to the specified color (does not alter alpha values).
	 * @private
	 * @param {string} src - The source of the image to recolor.
	 * @param {object} rgb - The rgb color object.
	 * @param {string} selector - The css selector of the rule to add the image to.
	 * @param {string} [prop] - The property to assign the image as a value of in the css rule.
	 *    Optional. Default = "background-image".
	 */
	ThemeUtil.prototype._generateImageAndUpdateRule = function( 
			src,
			rgb,
			selector,
			prop ) {

		// Get a reference to ourselves for use in closures.
		var self = this;

		// Load the image and then process it.
		var img = new Image();
		img.src = src;
		img.onload = function() {

			// Image loaded.

			// Generate a new image with the specified color.
			var dataURI = self._recolorCssImage( img, rgb );

			// Add the new image to the correct css rule/prop.
			prop = prop || "background-image";
			var cssRule = {};
			cssRule[ prop ] = "url( '" + dataURI + "' )";
			jss.set( selector, cssRule );

		};
	}

	/**
	 * Recolors an image and returns a dataURI of the generated image.
	 * Sets all pixels that aren't fully transparent to the specified color (does not alter alpha values).
	 * @private
	 * @param {Image} img - The loaded image to recolor.
	 * @param {object} rgb - The rgb color object.
	 */
	ThemeUtil.prototype._recolorCssImage = function( 
			img, 
			rgb ) {

		// Resolve the dimensions of the image.
		var width = img.naturalWidth;
		var height = img.naturalHeight;

		// Setup the canvas for this image,
		// updating the size and clearing any previous content.
		var canvas = this._getCanvas();
		canvas.width = width;
		canvas.height = height;
		var ctx = canvas.getContext( "2d" );
		ctx.clearRect( 0, 0, width, height );

		// Load the image onto the canvas and
		// get the underlying byte array.
		ctx.drawImage( img, 0, 0 );
		var imgData = ctx.getImageData( 0, 0, width, height );
		var byteArr = imgData.data;

		// Loop over each pixel (represented by every set of four bytes) in the image.
		// If any pixel isn't fully transparent, we convert it to the target color.
		for( var i = 0; i < byteArr.length; i += 4 ) {

			// Check if the pixel is transparent.
			if( byteArr[ i + 3 ] > 0 ) {

				// This pixel isn't transparent.

				// Change it to the custom color.
				byteArr[ i + 0 ] = rgb.r;
				byteArr[ i + 1 ] = rgb.g;
				byteArr[ i + 2 ] = rgb.b;
			}
		}

		// Draw the updated image to the canvas
		// and then export it as a data URL.
		ctx.putImageData( imgData, 0, 0 );
		return canvas.toDataURL();
	}

	/**
	 * Gets a file's content as a dataURI.
	 * @param {string} filePath - The absolute path to the file.
	 * @returns {string} The dataURI.
	 */
	ThemeUtil.prototype.getImageFileAsDataURI = function( filePath ) {

		try {

			// Read the file into a base64 encoded string.
			var stream = new ActiveXObject( "ADODB.Stream" );
			stream.Open();
			stream.Type = 1; // binary
			stream.LoadFromFile( filePath );
			var bytes = stream.Read();
			var base64Str = this.base64Encode( bytes.toArray() );

			// Create the dataURI.
			var fileExt = filePath.split( "." ).pop();
			var dataURI = "data:image/" + fileExt + ";base64," + base64Str;

			return dataURI;

		} finally {

			// Clean-up.
			if( stream ) {
				stream.Close();
			}
		}
	}

	/**
	 * Turns a byte array into a base64 encoded string.
	 * Adapted from: https://stackoverflow.com/a/7372816
	 * @param {Number[]} An array of bytes represented as integers between 0 and 255.
	 * @returns {string} - Base64 encoded string.
	 */
	ThemeUtil.prototype.base64Encode = function( byteArr ) {
	
		var CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
		var out = "", i = 0, len = byteArr.length, c1, c2, c3;
		while( i < len ) {
			c1 = byteArr[ i++ ];
			if( i == len ) {
				out += CHARS.charAt( c1 >> 2 );
				out += CHARS.charAt( ( c1 & 0x3 ) << 4 );
				out += "==";
				break;
			}
			c2 = byteArr[ i++ ];
			if( i == len ) {
				out += CHARS.charAt( c1 >> 2 );
				out += CHARS.charAt( ( ( c1 & 0x3 ) << 4 ) | ( ( c2 & 0xF0 ) >> 4 ) );
				out += CHARS.charAt( ( c2 & 0xF ) << 2 );
				out += "=";
				break;
			}
			c3 = byteArr[ i++ ];
			out += CHARS.charAt( c1 >> 2 );
			out += CHARS.charAt( ( (c1 & 0x3) << 4 ) | ( ( c2 & 0xF0 ) >> 4 ) );
			out += CHARS.charAt( ( ( c2 & 0xF ) << 2 ) | ( ( c3 & 0xC0 ) >> 6 ) );
			out += CHARS.charAt( c3 & 0x3F );
		}
		return out;
	}

})( jss );