"use strict"

define("mf/ui-ext/events", [], function() {

	var eventList = {

		ShellUI: { // ShellFrame
			name: "ShellUI",
			events: {
				OnStarted : null,
				OnStop : null,
				OnNewShellFrame: null, // This will get triggered automatically by the extended types.
				OnNewNormalShellFrame: "NormalShellFrame",
				OnNewCommonDialogShellFrame: "CommonDialogShellFrame"
				//OnNewEmbeddedShellFrame: "ShellFrame"			
			}
		},

		ShellFrame: {
			name: "ShellFrame",
			events: {
				OnStarted : null,
				OnStop : null,
				OnNewTaskPane : "TaskPane",
				OnNewBottomPane : "BottomPane",
				OnNewRightPane : "RightPane",
				OnNewSearchPane : "SearchPane",
				OnNewShellListing : "Listing",
				OnNewCommands : "Commands",
				OnShowMainMenu : null,
				OnActiveListingChanged: null		
			}
		},

		NormalShellFrame: {
			name: "NormalShellFrame",
			base: "ShellFrame",
			events: {

			}
		},

		CommonDialogShellFrame: {
			name: "CommonDialogShellFrame",
			base: "ShellFrame",
			events: {

			}
		},

		Pane: {
			name: "Pane",
			events: {
				OnStarted : null,
				OnStop : null,
				OnShowPane : null,
				OnHidePane : null	
			}	
		},

		SearchPane: {
			name: "SearchPane",
			base: "Pane",
			events: {

			}
		},

		TaskPane: {
			name: "TaskPane",
			base: "Pane",
			events: {
			}
		},

		RightPane: {
			name: "RightPane",
			base: "Pane",
			events: {
				PaneResized: null,
				MinimizedStateChanged: null,
				OnNewTab : null,
				// OnHideTab : null,
				OnTabSelected : null,
				OnTabUnselected : null
			}
		},

		BottomPane: {
			name: "BottomPane",	
			base: "Pane",
			events: {
				PaneResized: null,
				MinimizedStateChanged: null
				// OnShowTab : null,
				// OnHideTab : null,
				// OnTabSelected : null,
				// OnTabUnselected : null
			}
		},

		Listing: {
			name: "Listing",
			events: {
				OnStarted: null,
				OnStop: null,
				OnSelectionChanged: null,
				OnSelectedItemsChanged: null,
				OnContentChanged: null,
				OnShowContextMenu: null,
				OnShowMainMenu: null,
				OnListingActivated: null,
				OnListingDeactivated: null
			}
		},

		Commands: {
			name: "Commands",
			events: {
				OnCustomCommand: null
			}
		}
	}

	var priv = {
		shellUI: null,
		shellFrames: {},
		shellListings: {},
		callbacks: {}
	};


	function argsToArr(args) {
		return Array.prototype.slice.call(arguments)
	}

	function arrCopy(arr) {
		return arr.slice();
	}

	function copy(obj, into) {

		var newObj = into || {};

		for(var p in obj) {
			newObj[p] = obj[p];
		}

		return newObj;

	}


	/*
	 * Internal method to bind to all of an object's events.
	 *
	 */
	function bind( type, obj, context ) {	
	
		// Ensure we have the type defintion and not just the name
		if( typeof type === "string" )
			type = eventList[type];

		// Update the event context with the object.
		context = context || { hierarchy: [] };
		context[type.name] = obj;
		context.hierarchy = context.hierarchy.concat([type.name]);

		// Get a copy of the events applicable to our type.
		var events = copy(type.events);


		// Include the base types events in our applicable events if one was found.
		// Also reference the current object in the context by the base type
		if( type.base ) {			
			var baseType = eventList[type.base];
			//console.dump("Base Type Events: " + type.base, baseType.events);
			copy( baseType.events, events);
			context[baseType.name] = obj;
		}

		//console.dump("All Events: " + type.name, events);

		for( evt in events ) {

			// Resolve the event name and internal id.
			var evtName = evt.replace("On", "");
			var evtID = MFiles.Event[evtName];
			//console.log( "Bind: " + type.name + "." + evtName + " (" + evtID + ")" );

			// Resolve the base for this event (if there is one).
			var base = (baseType && baseType.events && baseType.events[evt] != undefined ) ? baseType : null;

			// Bind to the event with a custom generated handler.
			obj.Events.Register(evtID, getHandler(evtName, type, context, events[evt], base));
		}

	}


	/*
	 * Internal method to generate a custom handler for a specific object's event.
	 *
	 */
	function getHandler(evtName, type, context, returnType, base) {

		return function() {

			// If a return type is specified, bind to it's events.
			// Assume the return is the first argument to the callback.
			if( returnType ) {
				bind(returnType, arguments[0], copy(context));
			}


			// Get the passed arguments as an array.
			var params = Array.prototype.slice.call(arguments);

			context = copy(context)
			context.event = type.name + "." + evtName;
			context.params = params;
			
			// Add our event context as the last argument to the callbacks.
			params.push(context);

			// Trigger the events by each possible name in order of specificity
			// Least specific first.

			// Trigger the event with just the event name, i.e. "Started", "Stoped".
			trigger(evtName, params, false);

			// Trigger the event prefixed with the base object, if there is one.
			if( base )
				trigger(base.name + "." + evtName, params, false);

			// Trigger the event prefixed with the bound object. (Fully qualified event name in context).
			trigger(context.event, params, false);
		


		}

	}


	/*
     * Internal method for triggering events, and executing their bound callbacks.
	 *
	 * If an event is cancelable and a callback returns false, the method will immediately 
	 * return false. No further callbacks will be triggered.
	 *
	 * @param {string} event - The name of the event.
	 * @param {Array} [params] - The params to pass to the bound callbacks when triggered.
	 * @param {bool} [cancelable] - Indicates whether the event is cancelable. Default: False.
	 * @returns {bool} Returns false if the event is canceled, true otherwise.
     */
	function trigger( event, params, cancelable ) {

		//console.log( "Trigger: " + event + ", params:" + params.length + ", cancelable:" + !!cancelable );

		// If no handlers exist for the event, nothing to do.
		if( !priv.callbacks || !priv.callbacks[event] )
			return true;

		var callbacks = priv.callbacks[event];
		 
		// Loop over the callbacks for the specified event.
		for( var i = 0; i < callbacks.length; i++ ) {

			var callback = callbacks[i].callback;
			var context = callbacks[i].context || null;

			if( !cancelable ) {

				// Run non-cancelable callbacks seperately so
				// an error doesn't impede the others.  Also
				// prevents issues with errors stoping the main event
				// thread. We delegate to a separate function so the 
				// loop iterations values are locked in each timeout.
				runCallbackTimeout(callback, context, params);

			} else {

				// Trigger the cancellable callbacks synchronously.
				var retVal = callback.apply( context, params );

				// If the event is cancelable, and the handler returned "false"
				// cancel the event.
				if( retVal === false )
					return false;

			}
		}

		// Return true if we successfully called all handlers and there were no cancellations.
		return true;
	}

	function runCallbackTimeout(callback, context, params) {
		MFiles.setTimer(1, function() {
			callback.apply( context, params );
		});
	}


	var ShellUI = function() {};


	ShellUI.prototype.init = function ( shellUI ) {
		this.shellUI = shellUI;
		bind("ShellUI", shellUI);
	}


	/*
     * Binds a callback to an event.
	 *
	 * @param {string} event - The name of the event to listen to.
	 * @param {function} callback - The callback to run when the event is triggered.
	 * @param {object} [context] - The object for which the callback is run ("this" inside the callback).
     */
	ShellUI.prototype.on = function (event, callback, context) {

		// Add the new handler to the events structure.
		priv.callbacks = priv.callbacks || {};
		priv.callbacks[event] = priv.callbacks[event] || [];
		priv.callbacks[event].push( {
			callback: callback,
			context: context || null
		} );
	}


	/*
     * Unbinds a callback from an event (stops listening). 
	 *
	 * @param {string} [event] - The name of the event the handler is currently bound to.
	 *		If not specified, the handler is removed from all events.
	 * @param {function} [callback] - The callback handler to unbind.
	 *		If not specified, all handlers will be unbound from the specified event, or
	 *		if no event is specified, all handlers will be unbound from all events.
     */
	ShellUI.prototype.off = function( event, callback ) {

		// Check specified parameters.
		if( !event && !callback ) {

			// No event or handler specified, unbind all events.
			priv.callbacks = {}

		}
		else if( !callback ) {

			// No handler specified, delete all handlers for the specified event.
			priv.callbacks[event];

		} else {

			// Look for the index of the specified callback, if it can be found.
			var callbacks = priv.callbacks[event];
			var index = -1;
			for( var i = 0; i < callbacks.length; i++ ) {
				if( callbacks[i].callback == callback ) {
					index = i;
					break;
				}
			}

			// If the callback was found, remove it.
			if( index !== -1 ) {
				callbacks.splice( index, 1 );
			}
		}
	}


	return new ShellUI();
});



// Declare debugging globals.
var alert = function(msg) {}
var console = {
	log: function(msg) {}
};


// Called when a new shell UI is started.
function OnNewShellUI( shellUI ) {

	if( MFiles.CurrentApplicationPlatform === MFExtApplicationPlatformWeb ) return {};

	// Implement global alert function
	alert = function(msg) {
		shellUI.ShowMessage({message:msg.toString()});
	}

	// Implement global debugging console.
	//console = new DebuggingConsole(shellUI);	

	// Initialize events module.
	require(["mf/ui-ext/events"], function(events) {
		events.init(shellUI);


		events.on("ShellUI.Started", function() {
			//console.setEnabledState(true);
			//console.log("ShellUI Started");		
		});

	});	
}