"use strict";

// Declare global where our object will be exported.
var ViewExtensions;

( function() {

	// Private, local access to string resources.
	var l = ViewExtensionLocalization;

	/*
	 * Top Level View Extension Application.
	 * 
	 * Expects all new shellUI's to be registered with registerNewShellUI().
	 *
	 * @class
	 */
	ViewExtensions = function() {

		/* @member {MFiles.Vault}  */
		this.vault = null;

		/* @member {MFiles.ShellUI} */
		this.shellUI = null;

		/* @member {FolderPathSettingsManager} */
		this.folderPathSettingsManager = null;

		/* @member {ViewExtensionSettingsManager} */
		this.appSettingsManager = null;

		/* @member {SubListingManager} */
		this.subListingManager = null;

		/**
		 * Id of the task pane group created to store commands for the current shellFrame.
		 * -1 indicates no group currently exists for the shellFrame.
		 * @member {Number}
		 * */
		this.taskPaneGroup = -1;

		/**
		 * Ids of the custom commands that exist in the current shellFrame.
		 * @member {Number[]}
		 * @readonly
		 * */
		this.customCommands = [];

		// Initialize the localization language.
		ViewExtensionLocalization.init();
	}

	/*
	* Binds to a new ShellUI's events.
	*
	* @param {MFiles.ShellUI} shellUI - The ShellUI to bind events to.
	*/
	ViewExtensions.prototype.registerNewShellUI = function( shellUI ) {

		// Store the reference to the shellui.
		this.shellUI = shellUI;

		// Make self reference available in callbacks.
		var that = this;

		// Bind to the shellUI start event so we can initialize the settings object
		// once a valid handle on the Vault object is available
		shellUI.Events.Register( MFiles.Event.Started, function() {

			// Initialize vault and settings members.
			that.vault = shellUI.Vault;
			that.folderPathSettingsManager = new FolderPathSettingsManager( shellUI );
			that.appSettingsManager = new ViewExtensionSettingsManager( shellUI );
			that.subListingManager = new SubListingManager( shellUI, that.appSettingsManager );

		} );

		// Listen for all starting of normal shellFrames.
		shellUI.Events.Register( MFiles.Event.NewNormalShellFrame, function( shellFrame ) {
			shellFrame.Events.Register( MFiles.Event.Started, function() {
				that.normalShellFrameStarted( shellFrame );
			} );
		} );

	}

	/*
	* Triggered when a new normal ShellFrame is started.
	*
	* @param {MFiles.ShellFrame} shellFrame - The shellFrame that just started.
	*/
	ViewExtensions.prototype.normalShellFrameStarted = function( shellFrame ) {

		// Listen for location changes if the event is supported.
		var self = this;
		if( MFiles.Event.ViewLocationChanged ) {
			shellFrame.Events.Register( MFiles.Event.ViewLocationChanged, function() {

				// Clean-up the commands and taskpane groups used on the previous location.
				// This simulates legacy behavior so we don't need to significantly overhaul 
				// how we manage commands.
				if( self.taskPaneGroup != -1 ) {
					shellFrame.TaskPane.DeleteGroup( self.taskPaneGroup );
				}
				self.taskPaneGroup = -1;
				while( self.customCommands.length ) {
					shellFrame.Commands.DeleteCustomCommand( self.customCommands.pop() );					
				}

				// Indicate the location has changed.
				self.viewLocationChanged( shellFrame );
			} );
		}

		// Trigger the location change handler to update to
		// the new shellFrame's location.
		this.viewLocationChanged( shellFrame );
	}

	/*
	* Triggered when the current ShellFrame's location has changed.
	*
	* @param {MFiles.ShellFrame} shellFrame - The current shellFrame.
	*/
	ViewExtensions.prototype.viewLocationChanged = function( shellFrame ) {

		// Replace the default listing with our custom rendering if we are in a folder location,
		// and the vault is online.
		if( this.vault.ClientOperations.IsOnline() &&
			shellFrame.IsFolderLocation() ) {
			
			// Check if we are in the root folder. If so, check the software language
			// (resetting the language takes you back to root), but don't show any 
			// extensions here.
			if( !shellFrame.CurrentFolder || !shellFrame.CurrentFolder.Count ) {

				// We are in the root folder.

				// Update the localization.
				ViewExtensionLocalization.init();

				// Hide the detail listing folders if we haven't yet.
				this.hideDetailListingViews( shellFrame );

				// Don't load any view extension behavior for the root view.
				console.log( "View Extensions disabled in root view." );
				return;
			}

			// Only allow view extensions to be used in known and supported folder types.
			// We don't allow it to be used in TraditionalFolders or External view folders.
			// This check is done with a white-list for future-proofing. If we black-listed
			// the known types we don't support, view extensions would try to work with new
			// types added later that view extensions will probably not be able to support
			// implicitly.
			var currentFolder = shellFrame.CurrentFolder[0];
			if( currentFolder.FolderDefType !== MFFolderDefTypeViewFolder &&
				currentFolder.FolderDefType !== MFFolderDefTypePropertyFolder &&
				currentFolder.FolderDefType !== MFFolderDefTypeSearchFolder ) {

				// We are in an unsupported or unknown folder type.

				// Don't load any view extension behavior for this folder.
				console.log( "View Extensions disabled in unsupported view type %s.", currentFolder.FolderDefType );
				return;
			}

			// Make self reference available in callbacks.
			var that = this;

			// Create a data object for this shellFrame.
			var data = {
				appSettingsManager: this.appSettingsManager,
				folderPathSettingsManager: this.folderPathSettingsManager,
				subListingManager: this.subListingManager,
				pathInfo: viewUtils.getFolderPathInfo( this.vault, shellFrame.CurrentPath, shellFrame.CurrentFolder ),
				taskPaneGroup: shellFrame.TaskPane.CreateGroup( l( "TASKPANEGROUP_VIEW_EXTENSIONS" ), 0 )
			};
			this.taskPaneGroup = data.taskPaneGroup;

			// Resolve folder settings.
			var folderSettings = data.folderPathSettingsManager.getUserFolderPathSettings( data.pathInfo );

			// Determine if any panels are visible.
			var showListings = folderSettings.subListingPanelVisible;
			var showGroupLevels = data.pathInfo.groupLevels.length &&
								( folderSettings.groupingPanelVisible === true ||
								folderSettings.groupingPanelVisible === -1 && data.pathInfo.realObjLevels );

			// Check if the dashboard should be launched.
			if( showGroupLevels || showListings ) {

				// At least one of the panels should be visible.

				// Show the view extension dashboard.
				shellFrame.ShowDashboard( "view_extensions", data );

			} else {

				// No panels are visible. The dashboard isn't loaded.

				// Check if grouping levels are applicable to this location.
				if( data.pathInfo.groupLevels.length ) {

					// Add a button to show the grouping level listing panel.
					var showGroupingLevelsCmd = shellFrame.Commands.CreateCustomCommand( l( "CMD_SHOW_GROUP_LEVEL_PANEL" ) );
					shellFrame.Commands.SetIconFromPath( showGroupingLevelsCmd, "images\\grouping_levels.ico" );
					shellFrame.Commands.SetCommandState( showGroupingLevelsCmd, CommandLocation_All, CommandState_Active );
					shellFrame.TaskPane.AddCustomCommandToGroup( showGroupingLevelsCmd, data.taskPaneGroup, 1 );
					this.customCommands.push( showGroupingLevelsCmd );
				}

				// Add a button to show the sublisting panel.
				var showSubListingsCmd = shellFrame.Commands.CreateCustomCommand( l( "CMD_SHOW_SUBLISTING_PANEL" ) );
				shellFrame.Commands.SetIconFromPath( showSubListingsCmd, "images\\details_listing.ico" );
				shellFrame.Commands.SetCommandState( showSubListingsCmd, CommandLocation_All, CommandState_Active );
				shellFrame.TaskPane.AddCustomCommandToGroup( showSubListingsCmd, data.taskPaneGroup, 2 );
				this.customCommands.push( showSubListingsCmd );

				// Add a button to show the manage sublisting dialog.
				var settingsCmd = shellFrame.Commands.CreateCustomCommand( l( "CMD_MANAGE_SUBLISTINGS" ) );
				shellFrame.Commands.SetIconFromPath( settingsCmd, "images\\process.ico" );
				shellFrame.Commands.SetCommandState( settingsCmd, CommandLocation_All, CommandState_Active );
				shellFrame.TaskPane.AddCustomCommandToGroup( settingsCmd, data.taskPaneGroup, 3 );
				this.customCommands.push( settingsCmd );
			}

			// Listen for custom commands
			shellFrame.Commands.Events.Register( MFiles.Event.CustomCommand, function( cmd ) {

				var launchDashboard = false;

				if( showSubListingsCmd && cmd === showSubListingsCmd ) {

					// User has chosen to show the sublisting panel.
					// Change the settings and launch the dashboard.
					folderSettings.subListingPanelVisible = true;
					launchDashboard = true;

				} else if( showGroupingLevelsCmd && cmd === showGroupingLevelsCmd ) {

					// User has chosen to show the grouping level listing.
					// Change the settings and launch the dashboard.
					folderSettings.groupingPanelVisible = true;
					launchDashboard = true;

				} else if( settingsCmd && cmd === settingsCmd ) {

					// User has chosen to show the manage sublisting dialog.
					// We need to launch the dashboard so the view properties button works
					// and then indicate that it should be launched, once the dashboard is loaded.
					data.launchSettingsDialog = true;
					launchDashboard = true;
				}

				// Check if the dashboard should be launched.
				if( launchDashboard ) {

					// Save the settings that will make the dashboard visible.
					data.folderPathSettingsManager.setUserFolderPathSettings( data.pathInfo, folderSettings );

					// Remove all custom commands as the dashboard will manage those now.
					// For some reason this throws errors if we run it after the dashboard is shown.
					while( that.customCommands.length ) {
						shellFrame.Commands.DeleteCustomCommand( that.customCommands.pop() );
					}

					// Show dashboard.
					shellFrame.ShowDashboard( "view_extensions", data );
				}

			} );

		}
	}

	/**
	 * Tries to hide the detail listing views in the root view if they haven't 
	 * already been hidden for the user.
	 *
	 * The API only allows us to hide a view via a built in command when the
	 * view is selected in a listing, so we need to always check the root view
	 * to make sure the detail listing views are always hidden initially for
	 * each user. Once a detail listing view has been hidden, we never try to 
	 * alter its state again. There are user-settings that keep track of that.
	 *
	 * @param {MFilesUI.ShellFrame} shellFrame - The current shellFrame object 
	 *      that points to the root view.
	 */
	ViewExtensions.prototype.hideDetailListingViews = function( shellFrame ) {
		
		// Sanity check.
		if( !shellFrame.IsFolderLocation() || ( shellFrame.CurrentFolder && shellFrame.CurrentFolder.Count ) ) {
			throw new Error( "Invalid argument. ShellFrame must be for root view." );
		}

		// Resolve the view visiblity settings.
		// View visibility is stored only as a user setting.
		var visibilitySettingsKey = "SubListingParentViewVisibility"
		var visibilitySettings = this.appSettingsManager.value( visibilitySettingsKey, false ) || {};

		// Try to hide both the common and personal detail listings view for this user.
		var changed = this.tryHidingDetailListingParentView( shellFrame, true, visibilitySettings );
		changed = this.tryHidingDetailListingParentView( shellFrame, false, visibilitySettings ) || changed;

		// Save the settings if something changed.
		if( changed ) {
			this.appSettingsManager.save( visibilitySettingsKey, visibilitySettings, false );
		}
	}

	/**
	 * Tries to hide a detail listing view in the root folder if it hasn't been hidden before.
	 * 
	 * @param {MFilesUI.ShellFrame} shellFrame - The current shellFrame object that points to the root view.
	 * @param {boolean} common - Indicates whether the common or personal detail listings view should be hidden.
	 * @param {Object} visibilitySettings - The object that contains the settings indicating wether the common
	 *     and personal detail listings views have been hidden for the current user.
	 * @returns {boolean} - Returns true if the view has been/is hidden and the visibilitySettings have changed,
	 *     false otherwise.
	 */
	ViewExtensions.prototype.tryHidingDetailListingParentView = function( 
			shellFrame,
			common, 
			visibilitySettings ) {

		// Sanity checks.
		if( !visibilitySettings ) {
			throw new Error( "Argument null: visibilitySettings" );
		}
		if( !shellFrame.IsFolderLocation() || ( shellFrame.CurrentFolder && shellFrame.CurrentFolder.Count ) ) {
			throw new Error( "Invalid argument. ShellFrame must be for root view." );
		}

		// Resolve the applicable visibility key to use.
		var visibilityKey = ( common ) ? "commonViewHasBeenHidden" : "userViewHasBeenHidden";
		
		// If this view has been hidden before, we will not try to hide it again.
		if( visibilitySettings[ visibilityKey ] === true ) {
			return false;
		}

		// Resolve the detail listing view id, and be done if there isn't one.
		var viewInfo = this.appSettingsManager.value( "SubListingParentView", common );

		// No detail view listing has been created/defined yet.
		if( !viewInfo || !viewInfo.id ) {
			return false;
		}

		// Try to load the view.
		try {
			viewDef = this.vault.ViewOperations.GetView( viewInfo.id );
		} catch( e ) {
			return false;
		}

		// If the view is already hidden, or is not located in the root view,
		// mark it as having been hidden, so we don't try this every time the
		// user enters the root view.
		if( !viewDef.Visible || viewDef.HasParent ) {

			// Update the setting, and return true indicating it was changed.
			visibilitySettings[ visibilityKey ] = true;
			return true;
		}

		// Try to hide the view.
		try {

			// Select the folder in the listing.
			var folderDef = new MFiles.FolderDef();
			folderDef.SetView( viewInfo.id );
			shellFrame.Listing.SelectFolder( folderDef );

			// Run the "hide view" command.
			shellFrame.Commands.ExecuteCommand( MFiles.BuiltInCommand.HideView, null );

			// Update the setting, and return true indicating it was changed.
			visibilitySettings[ visibilityKey ] = true;
			return true;

		} catch( e ) {

			// Something went wrong. Nothing changed.
			return false;
		}

	}
} )();